/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Helene Joanin
 * Contributor(s): ______________________________________.
 *
 * 01/05/30 Christophe Ney - Lutris Technologies
 *          Fix to handle arguments containing white spaces.
 * 00/13/03 Vincent Sheffer
 *          Exit properly on failure of the external command.
 * 00/02/05 Nozomi Matsumura
 *          GenIC i18n problem.
 *          GenIC invoke javac and javac's output has non asc-ii characters,
 *          javac's output did not show collectly.
 * 04/16/01 Guillaume Tulloup
 *          Invoke directly the com.sun.tools.javac.Main.compile() method
 *          to work around a limitation on Windows.
 *          (See bug #100587)
 *
 * --------------------------------------------------------------------------
 * $Id: Cmd.java,v 1.6 2004/01/16 12:48:01 joaninh Exp $
 * --------------------------------------------------------------------------
 */


package org.objectweb.common;


/**
 * This class allows to run a command in sub-process.
 * @author Helene Joanin
 */

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.PrintStream;
import java.lang.Process;
import java.lang.String;
import java.util.*;

import org.objectweb.util.monolog.api.BasicLevel;

public class Cmd {

    private static final String COMPILER_CLASS_NAME = "com.sun.tools.javac.Main";

    private Vector mCmd = new Vector();
    private String[] mEnv = null;
    private boolean tryToInvoke = false;

    /**
     * Construtor method of Cmd class.
     * @param cmd command
     * @param invoke try to invoke directly the method of the java class of the command
     */
    public Cmd(String cmd, boolean invoke) {
        mCmd.addElement(cmd);
        tryToInvoke = invoke;
    }

    /**
     * Construtor method of Cmd class.
     * Equivalent to Cmd(cmd, false).
     * @param cmd command
     */
    public Cmd(String cmd) {
        this(cmd, false);
    }

    /**
     * Construtor method of Cmd class.
     * @param cmd command
     * @param env environment of the command
     */
    public Cmd(String cmd, String[] env, boolean invoke) {
        mCmd.addElement(cmd);
        mEnv = env;
        tryToInvoke = invoke;
    }

    /**
     * Add a new argument to the command.
     * @param arg argument added to the tail of the command argument list
     */
    public void addArgument(String arg) {
        mCmd.addElement(arg);
    }
    
    /**
     * Add a new argument to the command.
     * @param arg argument added to the tail of the command argument list
     */
    public void addArguments(Vector args) {
        Enumeration e = args.elements();
        while (e.hasMoreElements()) {
            mCmd.addElement((String)e.nextElement());
        }
    }
    
    /**
     * Execute the command.
     * (In case of the exit value of the command is not 0,
     * the output and error streams of the command is traced.)
     * @return false if the command cannot be executed, or if its exit value is not 0.
     */
    public boolean run() {

        if ( tryToInvoke && (((String) mCmd.get(0)).endsWith("javac")) ) {
            // Bug #100587 fixed to work around a limitation on Windows
            return compile();
        }

        RunnableStreamListener out, err;
        Process proc = null;
        boolean val = true;
        try {
            String[] cmd = new String[mCmd.size()];
            mCmd.copyInto(cmd);
            proc = Runtime.getRuntime().exec(cmd, mEnv);
        } catch (IOException e) {
            TraceCore.logger.log(BasicLevel.ERROR, "exception", e);
            return(false);
        }
        out = new RunnableStreamListener(new BufferedReader(new InputStreamReader(proc.getInputStream())),
                                         System.out);
        new Thread(out, "stdout listener for "+this.toString()).start();
        err = new RunnableStreamListener(new BufferedReader(new InputStreamReader(proc.getErrorStream())),
                                         System.err);
        new Thread(err, "stderr listener for "+this.toString()).start();
        try {
            val = proc.waitFor() == 0;
        } catch (InterruptedException e) {
            TraceCore.logger.log(BasicLevel.ERROR, "exception", e);
            val = false;
        }

        return(val);
    }

    private boolean compile() {

        // first arg should not be added because it's the name of the command
        String[] args = new String[mCmd.size() - 1];
        for (int i=0; i<mCmd.size() - 1; i++) {
            args[i] = (String) mCmd.get(i + 1);
        }

        try {
            // Use reflection
            Class c = Class.forName(COMPILER_CLASS_NAME);
            Object compiler = c.newInstance();
            java.lang.reflect.Method compile = c.getMethod("compile",
                                                           new Class [] {(new String [] {}).getClass()});
            int result = ((Integer) compile.invoke(compiler,
                                                   new Object[] {args})).intValue();
            return (result == 0);
        } catch (Exception e) {
            TraceCore.logger.log(BasicLevel.ERROR, "exception", e);
            return false;
        } 
    }
    
    /*
     * String representation of the object for UI purpose
     */
    public String toString() {
        StringBuffer buf = new StringBuffer();
        Enumeration e = mCmd.elements();
        while (e.hasMoreElements()) {
            String arg = (String)e.nextElement();
            if (arg == null) {
                arg = "null";
            }
            for (int i=0; i<arg.length();i++) {
                if (Character.isWhitespace(arg.charAt(i))) {
                    arg="\""+arg+"\"";
                    break;
                }
            }
            buf.append(arg);
            buf.append(' ');
        }
        return buf.toString().trim();
    }
    
}


//  this runs happily, swallowing the stdout from the process
class RunnableStreamListener implements Runnable {
    BufferedReader in_;
    PrintStream stream_;
    RunnableStreamListener(BufferedReader in, PrintStream stream ) {
        in_ = in;
        stream_ = stream;
    }
    public void run() {
        String line;
        try {
            while ((line = in_.readLine()) != null)
                stream_.println(line);
        } catch (IOException e ) {
            stream_.println(e.toString());
        }
    }
}

