/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 * --------------------------------------------------------------------------
 * $Id: AbsJmxServiceImpl.java,v 1.11 2004/03/25 09:19:01 danesa Exp $
 * --------------------------------------------------------------------------
 */
package org.objectweb.jonas.jmx;

// JNDI includes
import java.net.MalformedURLException;
import java.util.StringTokenizer;

import javax.management.MBeanServer;
import javax.management.MBeanServerFactory;
import javax.management.remote.JMXServiceURL;
import javax.naming.Context;
import javax.naming.NamingException;

import org.objectweb.carol.util.configuration.CarolConfiguration;
import org.objectweb.jonas.common.JProp;
import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.service.AbsServiceImpl;
import org.objectweb.jonas.service.ServiceException;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * JMX Service implementation
 * This singleton class must exist in each jonas server that is to be administered via JMX
 * @author Guillaume Riviere
 * Contributor(s): Michel Bruno, Adriana Danes
 */

public abstract class AbsJmxServiceImpl extends AbsServiceImpl implements JmxService {

    // Trace logger
     static protected Logger logger = null;

    // the name of the JOnAS server that is administerd by the MBeanServer
    protected String jonasName = null;

    // the name of the management domain
    protected String domainName = null;

    // the name of the RMI connector allowing to access the MBeanServer remotelly
    protected String rmiConnectorName = null;

    // JMX server (MBean server) instance
    protected MBeanServer jmxServer = null;
    // JMX server address
    protected JMXServiceURL[] jmxServiceUrl = null;

    // Jmx service configuration parameters
    public static final String CLASS = "jonas.service.jmx.class";

    private static final String JRMP = "jrmp";
    private static final String IIOP = "iiop";
    private static final String JEREMIE = "jeremie";
    
    private boolean useJrmp = false;
    private boolean useIiop = false;
    private boolean useJeremie = false;
    /**
     * Init the Service.
     * Create the MBeanServer.
     * Configuration information is passed through a Context object.
     */
    public void doInit(Context ctx) throws ServiceException {
        logger = Log.getLogger(Log.JONAS_SERVER_PREFIX);
        super.initLogger(Log.getLogger(Log.JONAS_MANAGEMENT_PREFIX));

        // Get the current server name
        try {
            jonasName = (String) ctx.lookup("jonas.name");
        } catch (NamingException ne) {
            logger.log(BasicLevel.DEBUG, "Cannot initialize JMX service "+ne);
            throw new ServiceException("Cannot initialize JMX service", ne);
        }

        // Create a MBeanServer
        // Get the current domain name
        try {
            domainName = (String) ctx.lookup(JProp.DOMAIN_NAME);
        } catch (NamingException ne) {
            // This can't happen as JProp set it in global environment 
            // and the ServiceManager passed it to this service
        }
        jmxServer = MBeanServerFactory.createMBeanServer(domainName);

        // Determine protocols used by Carol
        String s = CarolConfiguration.getProtocols();
        StringTokenizer st = new StringTokenizer(s, ",", false);
        int nbProtocols = 0;
        while (st.hasMoreElements()) {
            String protocol = st.nextToken().trim();
            if (protocol.equals(JRMP) == true) {
            	if (useJrmp == false) {
            		useJrmp = true;
            		nbProtocols = nbProtocols + 1;
            	}
            } else if (protocol.equals(IIOP) == true) {
            	if (useIiop == false) {
            		useIiop = true;
            		nbProtocols = nbProtocols + 1;
            	}
            }  else if (protocol.equals(JEREMIE) == true) {
            	if (useJeremie == false) {
            		useJeremie = true;
            		nbProtocols = nbProtocols + 1;
            	}
            }
        }
        // Create addresses for the JMX Connector Servers   
        jmxServiceUrl = new JMXServiceURL[nbProtocols];
        int index = 0;
        try {
        	if (useJrmp) {
        		jmxServiceUrl[index++] = new JMXServiceURL("rmi", "localhost", 0, "/jndi/jrmpconnector");
        	}
        	if (useIiop) {
        		jmxServiceUrl[index++] = new JMXServiceURL("iiop", "localhost", 0, "/jndi/iiopconnector");
        	}
        	if (useJeremie) {
        		jmxServiceUrl[index++] = new JMXServiceURL("rmi", "localhost", 0, "/jndi/jeremieconnector");
        	}
        } catch (MalformedURLException e) {
            throw new ServiceException("Cannot initialize JMX service", e);
        }
        logger.log(BasicLevel.DEBUG, "JMX Service initialized");
    }

    /**
     * Start the Service
     * Initialization of the service is already done.
     */
    public abstract void doStart() throws ServiceException ;

    /**
     * Stop this service
     */
    public abstract void doStop() throws ServiceException ;

    /**
     * @return The reference of the MBean server
     */
    public MBeanServer getJmxServer() {
        return this.jmxServer;
    }

    /**
     * @return The name of the MBean server's rmi connector
     */
    public String getRmiConnectorName() {
        return this.rmiConnectorName;
    }
   /**
     * @return The name of the JOnAS server
     */
    public String getServerName() {
        return this.jonasName;
    }

    /**
     * @return The name of the management domain
     */    
    public String getDomainName() {
        return this.domainName;
    }

    /**
     * Remove internal references to the MBeanServer. 
     */
    protected void releaseJmxServer() {
	try {
	    MBeanServerFactory.releaseMBeanServer(jmxServer); 
	} catch(java.lang.IllegalArgumentException ex) {
	    // The mbeanServer is not found.
	    logger.log(BasicLevel.ERROR, "Cannot find the MBeanServer" + ex);
	}
    }

}
