/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: JmxServiceImpl.java,v 1.13 2004/03/23 15:23:34 danesa Exp $
 * --------------------------------------------------------------------------
 */
package org.objectweb.jonas.jmx;

import java.io.IOException;

import javax.management.ObjectName;
import javax.management.JMException;
import javax.management.remote.JMXConnectorServer;
import javax.management.remote.JMXConnectorServerFactory;
import javax.management.remote.JMXServiceURL;
import javax.naming.InitialContext;

import org.objectweb.jonas.jmx.sunri.RMIConnector;
import org.objectweb.jonas.jmx.sunri.RMIConnectorImpl;
import org.objectweb.jonas.service.JonasAlreadyStartedException;
import org.objectweb.jonas.service.ServiceException;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * JMX Service implementation for SUN RI
 * Provides specific doStart() and doStop () methods.

 * @author Adriana Danes.
 *
 */

public class JmxServiceImpl extends AbsJmxServiceImpl {

    /**
     * Start the Service. 
     * Only need to create a RMI connector
     */
    public void doStart() throws ServiceException {       
    	
    	try {
    		// Creat a RMI Connector for the JMX agent using the JOnAS RMIConnector class
    		// --------------------------------------------------------------------------
    		RMIConnector rmiConnector = new RMIConnectorImpl(jmxServer);
    		// Register the connector in JNDI
    		InitialContext ictx = new InitialContext();
    		rmiConnectorName = "RMIConnector_"+jonasName;
    		ictx.bind(rmiConnectorName, rmiConnector); 
    		ictx.close();
    		
    		// Use JMX Remote API: create one or more connector servers
    		// ---------------------------------------------
    		int nbServers = jmxServiceUrl.length;
    		if (nbServers == 1) {
        		// The environment map, null in this case
        		java.util.Map environment = null; 
        		startConnector(jmxServiceUrl[0], environment);			
    		} else {
    			for (int i = 0 ; i < nbServers ; i++) {
    				// The environment map, null in this case
    				java.util.Map environment = null; 
    				try {
    					startConnector(jmxServiceUrl[i], environment);
    				} catch (IllegalArgumentException e) {
    					// Catch Carol thrown exception when trying to register 
    					// a stub whith wrong type in a given registry
    				}
    			}
    		}
    	}catch (javax.naming.NameAlreadyBoundException ne) {
    		logger.log(BasicLevel.DEBUG, "Cannot start JMX service " + ne);
    		throw new JonasAlreadyStartedException();
    	} catch(IOException ioe) {
    		logger.log(BasicLevel.DEBUG, "Cannot start JMX service " + ioe);
    		throw new ServiceException("Cannot start JMX service", ioe);
    	} catch (Exception e) {
    		logger.log(BasicLevel.DEBUG, "Cannot start JMX service " + e);
    		throw new ServiceException("Cannot start JMX service", e);       
    	}
    }
    /**
     * Stop this service
     */
    public void doStop() {
    	try {
    		InitialContext ictx = new InitialContext();
    		ictx.unbind("RMIConnector_"+jonasName);
    		ictx.close();
    		
    		// Deactivates the connector server, that is, stops listening for client connections. 
    		// Calling this method will also close all client connections that were made by this server.
    		// Being a potentialy slow operation, we keep it commented for the moment
    		// cntorServer.stop();
    	} catch (Exception e) {
    		logger.log(BasicLevel.ERROR, "Cannot Unbind Jmx RMI Connector" + e);
    	}
    	
    	// Remove internal references to the MBeanServer. 
    	releaseJmxServer();
    	
    	logger.log(BasicLevel.DEBUG, "JMX Service stopped");
    }
    
    void startConnector(JMXServiceURL serviceURL, java.util.Map environment) 
	throws IOException, JMException {
    	JMXConnectorServer connectorServer = JMXConnectorServerFactory.newJMXConnectorServer(serviceURL, environment, null);
		// Register the JMXConnectorServer in the MBeanServer
    	String protocol = serviceURL.getProtocol();
    	String connectorObjectName = "connector_" + protocol;
		ObjectName connectorServerName = JonasObjectName.jmxConnectorServer(protocol, connectorObjectName);
		jmxServer.registerMBean(connectorServer, connectorServerName);
		// Start the JMXConnectorServer
		try {
			connectorServer.start();
		} catch (IllegalArgumentException e) {
			throw e;
		}
    }
}
