/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999-2004 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: Bootstrap.java,v 1.44 2004/03/25 17:15:24 benoitf Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.server;

import java.io.File;
import java.lang.reflect.Method;
import java.net.URL;
import java.util.StringTokenizer;

import org.objectweb.jonas.common.JProp;


/**
 * This class load all the jars needed to start JOnAS and after this, it
 * launch the JOnAS server (Server class).
 * @author Ludovic Bert   (initial developer)
 * @author Florent Benoit (initial developer)
 * @author Philippe Durieux (new jonas jar files architecture)
 */
public class Bootstrap {

    /**
     * JONAS_ROOT/lib directory
     */
    private static String jonasLib;

    /**
     * JONAS_ROOT/lib/common directory
     */
    private static String commonLib;

    /**
     * JONAS_BASE directory
     */
    private static String jonasBase;

    /**
     * jonas.properties
     */
    private static JProp props = null;

    /**
     * List of services in jonas.properties
     */
    private static String[] services;

    /**
     * Use of the 'web' service
     */
    private static boolean serviceWeb = false;

    /**
     * Return an array containing all the resources declared in the classpath
     * environment variable.
     * @param xtracp XTRA classpath
     * @return a list containing all the resources declared in the classpath
     * environment variable.
     * @throws Exception if a resource cannot be found.
     */
    private static JURLs getClassPath(boolean xtracp) throws Exception {
        String cpprop = xtracp ? "jonas.classpath" : "java.class.path";
        String cpValue = System.getProperty(cpprop);
        //System.out.println("classpath=" + cpValue); // DEBUG
        String sep = File.pathSeparator;
        JURLs jurls = new JURLs();
        StringTokenizer tokenizer = new StringTokenizer(cpValue, sep);
        while (tokenizer.hasMoreTokens()) {
            File file = new File(tokenizer.nextToken());
            if (file.isDirectory()) {
                jurls.addDir(file);
            } else {
                jurls.add(file);
            }
        }
        return jurls;
    }


    /**
     * Return an array containing all the resources used for the integration
     * of Catalina.
     * @param catalinaHomeDir the home directory of Catalina.
     * @return a list containing all the resources used for the integration
     * of Catalina.
     * @throws Exception if a resource cannot be found.
     */
    private static JURLs getCatalinaResources(File catalinaHomeDir) throws Exception {

        JURLs jurls = new JURLs();
        File jarf = new File(jonasLib, "web_catalina.jar");

        if (jarf.exists()) {
            jurls.add(jarf);
        } else {
            if (new File(commonLib, "ow_jonas_tomcat4.jar").exists()) {
                jurls.add(new File(commonLib, "ow_jonas_tomcat4.jar"));
            } else if (new File(commonLib, "ow_jonas_tomcat5.jar").exists()) {
                jurls.add(new File(commonLib, "ow_jonas_tomcat5.jar"));
            } else {
                System.out.println("WARN : No ow_jonas_tomcat*.jar found in lib/common directory.");
            }
            jurls.add(new File(catalinaHomeDir, "bin" + File.separator + "bootstrap.jar"));
            jurls.add(new File(catalinaHomeDir, "server" + File.separator + "lib"));
            jurls.add(new File(catalinaHomeDir, "common" + File.separator + "lib"));
        }
        return jurls;
    }

    /**
     * Return an array containing all the resources used for the integration
     * of Axis.
     * @return a list containing all the resources used for the integration
     * of Axis.
     * @throws Exception if a resource cannot be found.
     */
    private static JURLs getAxisResources() throws Exception {
        JURLs jurls = new JURLs();

        File jarf = new File(jonasLib, "webservices_axis.jar");

        // if ant install done
        if (jarf.exists()) {
            jurls.add(jarf);
        } else {
            // add manually all axis specific files
            jurls.add(new File(commonLib, "ws" + File.separator +  "axis"), ".jar");

            if (new File(commonLib, "ow_jonas_axis.jar").exists()) {
                jurls.add(new File(commonLib, "ow_jonas_axis.jar"));
            } else {
                System.out.println("WARN : No ow_jonas_axis.jar found in lib/common directory.");
            }
        }
        return jurls;
    }

    /**
     * Return an array containing all the resources used for WebServices.
     * @return a list containing all the resources used for WebServices.
     * @throws Exception if a resource cannot be found.
     */
    private static JURLs getWebServicesResources() throws Exception {
        JURLs jurls = new JURLs();

        jurls.add(new File(commonLib, "ws" + File.separator + "wsdl4j"), ".jar");
        jurls.add(new File(commonLib, "ws" + File.separator + "namespace"), ".jar");
        // add JAX-M Resources
        jurls.add(new File(commonLib, "ws" + File.separator + "jaxm"), ".jar");
        // add JAX-R Resources
        jurls.add(new File(commonLib, "ws" + File.separator + "jaxr"), ".jar");

        return jurls;
    }

    /**
     * Return an array containing all the resources used for the integration
     * of Jetty.
     * @param jettyHomeDir the home directory of Jetty.
     * @return a list containing all the resources used for the integration
     * of Jetty.
     * @throws Exception if a resource cannot be found.
     */
    private static JURLs getJettyResources(File jettyHomeDir) throws Exception {

        JURLs jurls = new JURLs();

        File jarf = new File(jonasLib, "web_jetty.jar");

        if (jarf.exists()) {
            jurls.add(jarf);
        } else {
            String jettyLib = jettyHomeDir + File.separator + "lib";
            jurls.add(new File(jettyLib, "org.mortbay.jetty.jar"));

            String jettyExt = jettyHomeDir + File.separator + "ext";
            jurls.add(new File(jettyExt, "ant.jar"));
            jurls.add(new File(jettyExt, "jasper-compiler.jar"));
            jurls.add(new File(jettyExt, "jasper-runtime.jar"));
            jurls.add(new File(jettyLib, "org.mortbay.jmx.jar"));
            jurls.add(new File(jettyLib, "javax.servlet.jar"));
            jurls.add(new File(commonLib, "ow_jonas_jetty.jar"));
        }
        return jurls;
    }

    /**
     * Return an array containing all the resources used for JOnAS
     * @return a list containing all the resources.
     * @throws Exception if a resource cannot be found.
     */
    private static JURLs getJonasResources() throws Exception {

        JURLs jurls = new JURLs();

        File jarf = new File(jonasLib, "jonas.jar");

        if (jarf.exists()) {
            jurls.add(jarf);
        } else {
            System.out.println("INFO : Please take the time to do the 'ant install'"
                               + " task in the JONAS_ROOT directory. This will "
                               + "improve the performances, remove some warnings and"
                               + " allow you to build the examples and use RMI/IIOP.");

            // add Orb Independant JOnAS resources
            jurls.add(new File(commonLib, "ow_jonas.jar"));

            // add external jars needed for JOnAS
            jurls.add(new File(commonLib, "j2ee"), ".jar");
            jurls.add(new File(commonLib, "acme"), ".jar");
            jurls.add(new File(commonLib, "joram"), ".jar");
            jurls.add(new File(commonLib, "jorm"), ".jar");
            jurls.add(new File(commonLib, "jotm"), ".jar");
            jurls.add(new File(commonLib, "log"), ".jar");
            jurls.add(new File(commonLib, "medor"), ".jar");
            jurls.add(new File(commonLib, "velocity"), ".jar");
            jurls.add(new File(commonLib, "xml"), ".jar");
            jurls.add(new File(commonLib, "carol"), ".jar");
            jurls.add(new File(commonLib, "jakarta"), ".jar");

            jurls.add(new File(commonLib, "jonathan"), ".jar");
            jurls.add(new File(commonLib, "p6spy"), ".jar");
            jurls.merge(getWebServicesResources());
        }

        // add jonas.base/lib/ext only if jonas.base != jonas.root
        // jonas.base/lib/ext is added before jonas.root/lib/ext
            
        if (!System.getProperty("install.root").toLowerCase().equals(jonasBase.toLowerCase())) {
            File jonasBaseLibExt = new File(jonasBase + File.separator + "lib" + File.separator + "ext");
            if (jonasBaseLibExt.exists()) {
                jurls.add(jonasBaseLibExt, ".jar");
            }
        }

        // add jonas.root/lib/ext
        jurls.add(new File(jonasLib, "ext"), ".jar");

        // Add resources for jmx, depending on the choosen implementation
        String myjmx = props.getValue("jonas.service.jmx.class");
        String jmxLib = commonLib + File.separator + "jmx";

        //if ((myjmx != null) && (myjmx.indexOf("sunri") == -1)) {
        //jurls.add(new File(jmxLib, "mx4j"), ".jar");
        //} else {
        //jurls.add(new File(jmxLib, "jmx"), ".jar");
        //}
        jurls.add(new File(jmxLib), ".jar");
        return jurls;
    }

    /**
     * Get the list of the resources to load when launching JOnAS.
     * @param defcl the default classloader
     * @return the list of the resources to load when launching JOnAS.
     * @throws Exception if a resource cannot be found.
     */
    public static URL[] getResources(boolean defcl) throws Exception {
        JURLs jurls = new JURLs();

        // Start with resources from the classpath
        jurls.merge(getClassPath(defcl));

        // Load now mail.jar because it must be loaded before activation.jar
        jurls.add(new File(commonLib, "j2ee" + File.separator + "mail"), ".jar");

        // activation.jar needed at least for GenIC
        jurls.add(new File(commonLib, "j2ee" + File.separator + "activation"), ".jar");

        // add JOnAS resources
        jurls.merge(getJonasResources());

        String catalinaHome = System.getProperty("catalina.home");
        String jettyHome = System.getProperty("jetty.home");

        // Set in all case catalina.home and base to $JONAS_BASE
        // if library is found. Useful if a user want to run in
        // a same environment JOnAS/Tomcat and Tomcat with different
        // versions
        File jarCatalina = new File(jonasLib, "web_catalina.jar");
        if (jarCatalina.exists()) {
            System.setProperty("catalina.home", jonasBase);
            System.setProperty("catalina.base", jonasBase);
            catalinaHome = jonasBase;
        }

        // Set jetty.home to $JONAS_BASE if library is found
        // and if jetty.home isn't set.
        if (jettyHome == null) {
            File jarJetty = new File(jonasLib, "web_jetty.jar");
            if (jarJetty.exists()) {
                System.setProperty("jetty.home", jonasBase);
                jettyHome = jonasBase;
            }
        }

        // Add the specific resource of the web_container (tomcat or jetty) service

        if (serviceWeb) {
            // add resources for catalina if required
            if (catalinaHome != null) {
                jurls.merge(getCatalinaResources(new File(catalinaHome)));
            }

            // add resources for jetty if required
            if (jettyHome != null) {
                jurls.merge(getJettyResources(new File(jettyHome)));
            }
            // else user has not set the webcontainer HOME var
        }


        // add resources for WebServices
        jurls.merge(getAxisResources());

        return jurls.toURLs();
    }


    /**
     * Server main routine (Load all the jars needed to start jonas)
     * @param args the list of the args to give to the bootstrap class.
     */
    public static void main(String[] args) {
        String classToRun = args[0];
        
        // JDK 1.4 is required
        String jdkInUse = System.getProperty("java.version");
        if (jdkInUse.indexOf("1.4") != 0) {
            System.err.println("A 1.4 JVM is required for JOnAS. Current JVM version is " + jdkInUse);
            System.exit(0);
        }
        
        try {
            // Get jonas properties
            props = JProp.getInstance();
            services = props.getValueAsArray("jonas.services");
            if (services != null) {
                for (int i = 0; i < services.length; i++) {
                    if (services[i].equals("web")) {
                        serviceWeb = true;
                    }
                }
            }

            // Get $JONAS_BASE
            jonasBase = JProp.getJonasBase();

            // Get the $JONAS_ROOT/lib directories
            String jonasRoot = System.getProperty("install.root");
            if (jonasRoot == null) {
                System.err.println("JONAS_ROOT is not set");
                System.exit(2);
            }
            jonasLib = jonasRoot + File.separator + "lib";
            commonLib = jonasLib + File.separator + "common";

            // Create a ClassLoader and set it for the current thread.
            boolean dlr = props.getValueAsBoolean("jonas.default.classloader", false);
            JClassLoader jonasLoader = null;
            if (dlr) {
                jonasLoader = new JClassLoader(getResources(dlr));
            } else {
                jonasLoader = new JClassLoader(getResources(dlr), null);
            }
            Thread.currentThread().setContextClassLoader(jonasLoader);

            //jonasLoader.printURLs(); // DEBUG

            // Launch the "class_to_run" by using our classloader.
            Class clazz = jonasLoader.loadClass(classToRun);
            Class[] argList = new Class[] {args.getClass() };
            Method meth = clazz.getMethod("main", argList);
            String[] newArgs = new String [args.length - 1];
            System.arraycopy(args, 1, newArgs, 0, newArgs.length);
            Object appli = meth.invoke(null, new Object[]{newArgs});
        } catch (Exception e) {
            System.err.println(classToRun + " error: " + e);
            e.printStackTrace();
            System.exit(2);
        }
    }
}
