/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 * --------------------------------------------------------------------------
 * $Id: J2EEServer.java,v 1.14 2004/03/30 23:41:57 ehardesty Exp $
 * --------------------------------------------------------------------------
 */
package org.objectweb.jonas.server;

// Java imports
import java.util.ArrayList;
import java.util.List;

import javax.management.Notification;
import javax.management.NotificationFilter;
import javax.management.NotificationListener;

import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.container.EJBService;
import org.objectweb.jonas.ear.EarService;
import org.objectweb.jonas.management.AttributeAddNotification;
import org.objectweb.jonas.management.AttributeRemoveNotification;
import org.objectweb.jonas.management.ListenerJavaBean;
import org.objectweb.jonas.management.j2eemanagement.J2EEManagedObject;
import org.objectweb.jonas.resource.ResourceService;
import org.objectweb.jonas.service.ServiceManager;
import org.objectweb.jonas_timer.TimerEvent;
import org.objectweb.jonas_timer.TimerEventListener;
import org.objectweb.jonas_timer.TimerManager;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * MBean class for Server management.
 * Two classes are used to provide all management services : J2EEServer and J2EEServerMBean.
 * This class provide the notification's process when add or remove resources.
 * @author Guillaume Riviere
 * @author Michel Bruno
 * <br>contributors Adriana Danes, Michel-Ange Anton
 *
 */
public class J2EEServer extends J2EEManagedObject implements TimerEventListener {
    /**
     * object to manage
     */
    private Server server = null;
    private String serverName = null;
    private String serverVersion = null;
    private String serverVendor = null;
    private String protocols = null;
    private ArrayList deployedObjects = null;
    private ArrayList resources = null;
    private ArrayList javaVMs = null;
    /**
     * monitoring objects
     */
    private boolean activated = true;

    private int sizeTableMeasures = 120;
    private int range = 10;
    private Long[] tableMeasures;
    private long maxtotalmemory;
    private TimerEvent mytimer = null;

    /**
     * Loggers
     */
    static private Logger mgtLogger = Log.getLogger(Log.JONAS_MANAGEMENT_PREFIX);
    static private Logger evtLogger = Log.getLogger(Log.JONAS_MANAGEMENT_EVENT_PREFIX);

    /**
     * Notifications variables
     */
    // sequence number needed when sending notifications
    private long sequenceNumber = 0;

    // Support JMX notification
    // -- temporary solution allowing to support one listener (the ManagementListener MBean)
    private ListenerJavaBean myListener = null;

    /**
     * Services used to deploy J2EE modules and applications
     */
    EJBService ejbService;
    EarService earService;
    ResourceService rarService;

    /**
     * MBean constructor
     * @param objectName object name of the managed object
     * @param server Server object
     * @param serverName server name
     * @param version JOnAS version
     * @param protocols protocols supported
     */
    public J2EEServer(String objectName, Server server, String serverName, String serverVersion
        , String serverVendor, String protocols) {
        super(objectName);
        this.server = server;
        this.serverName = serverName;
        this.serverVersion = serverVersion;
        this.serverVendor = serverVendor;
        this.protocols = protocols;
        initTable();
        deployedObjects = new ArrayList();
        resources = new ArrayList();
        javaVMs = new ArrayList();
        //Launch Timer
        mytimer = TimerManager.getInstance().addTimer(this, range, (Object)new Integer(1), true);
        try {
            ejbService = (EJBService) ServiceManager.getInstance().getEjbService();
            earService = (EarService) ServiceManager.getInstance().getEarService();
            rarService = (ResourceService) ServiceManager.getInstance().getRarService();
        } catch(Exception e) {
            // should never occurs
        }
    }

    /**
     * Server name
     * @return String JOnAS server name
     */
    public String getServerName() {
        return serverName;
    }

    /**
     * Server version
     * @return String JOnAS server version
     */
    public String getServerVersion() {
        return serverVersion;
    }

    /**
     * Server vendor
     * @return String JOnAS server version
     */
    public String getServerVendor() {
        return serverVendor;
    }

    /**
         * Return the list of MBean names corresponding to the deployed J2EEModules and J2EEApplications
     * @return java.util.ArrayList list of the deployed objects' MBean names
     */
    public List getDeployedObjects() {
        return deployedObjects;
    }

    /**
     * Add an object name to the <code>deployedObjects</code> list.
     * @param objectName Object name corresponding to a J2EEDeployedObject MBean
     */
    public void addDeployedObject(String objectName) {
        deployedObjects.add(objectName);
    }

    /**
     * Remove an object name to the <code>deployedObjects</code> list.
     * @param objectName Object name corresponding to a J2EEDeployedObject MBean
     */
    public void removeDeployedObject(String objectName) {
        deployedObjects.remove(objectName);
    }

    /**
     * Return the list of MBean names corresponding to the resources available on this server
     * @return java.util.ArrayList list of the resources' MBean names
     */
    public List getResources() {
        return resources;
    }

    /**
         * Return the list of MBean names corresponding to the JVMs on which this server has running threads
     * @return java.util.ArrayList list of the JVMs' MBean names
     */
    public List getJavaVMs() {
        return javaVMs;
    }

    /**
     * Add an object name to the <code>javaVMs</code> list.
     * @param objectName Object name correspondig to a JVM MBean
     */
    public void addJavaVM(String objectName) {
        javaVMs.add(objectName);
    }

    /**
     * Protocols supported by this Server
     * @return String protocols supported
     */
    public String getProtocols() {
        return protocols;
    }

    /**
     * Stop server (with stopping the JVM)
     */
    public void stop()
        throws Exception {
        server.stop();
    }

    /**
     * run the garbage collector
     */
    public void runGC() {
        Runtime.getRuntime().gc();
    }

    /**
     * get jvm used memory
     * @return Long - current used memory
     */
    public long getCurrentUsedMemory() {
        return usedMemory();
    }

    /**
     * get jvm total memory
     * @return Long - current total memory
     */
    public long getCurrentTotalMemory() {
        return totalMemory();
    }

    /**
     * set range
     * @param range range for free memory measurement
     */
    public void setRange(int range) {
        if (this.range != range) {
            synchronized (this) {
                if (range < 10) {
                    throw new IllegalArgumentException("range could not be < 10");
                }
                this.range = range;
                // reset table values
                initTable();
                /// change range
                mytimer.change(this.range, new Integer(1));
            }
        }
    }

    /**
     * get range
     * @return Integer - range
     */
    public int getRange() {
        return range;
    }

    /**
     * Set the size of the table of measures
     * @param Number of measures
     */
    public void setSizeTableMeasures(int sizeMeasuresTable) {
        if (sizeMeasuresTable != this.sizeTableMeasures) {
            synchronized (this) {
                if (sizeMeasuresTable <= 1) {
                    throw new IllegalArgumentException("number of measures could not be <= 1");
                }
                this.sizeTableMeasures = sizeMeasuresTable;
                initTable();
            }
        }
    }

    /**
     * Get the size of the table of measures
     * @return Number of measures
     */
    public int getSizeTableMeasures() {
        return sizeTableMeasures;
    }

    /**
     * get the table of value
     * @return Long[] measures table
     */
    public Long[] getTableMeasures() {
        return tableMeasures;
    }

    // ===============================================================
    // TimerEventListener implementation
    // ===============================================================
    /**
     * The measures timeout has expired
     * Do not synchronize this method to avoid deadlocks!
     * @param arg Object
     */
    public void timeoutExpired(Object arg) {
        int argvalue = ((Integer) arg).intValue();
        switch (argvalue) {
            case 1:

                //timeout expired
                long u_value = usedMemory();
                long tm_value = totalMemory();

                // put the max total memory
                if (tm_value > maxtotalmemory) {
                    maxtotalmemory = tm_value;
                }
                for (int i = 0; i < tableMeasures.length; i++) {
                    if (i != tableMeasures.length - 1) {
                        tableMeasures[i] = tableMeasures[i + 1];
                    }
                    else {
                        tableMeasures[i] = new Long(u_value);
                    }
                }
                break;
            default:

                //timeoutExpired do nothing;
                break;
        }
    }

    /**
     * set monitoring activation
     * @param act Boolean
     */
    public void setActivated(boolean p_Activated) {
        if (p_Activated) {
            if (!activated) {
                synchronized (this) {
                    activated = true;
                    initTable();
                }
                // launch timer
                mytimer = TimerManager.getInstance().addTimer(this, range, (Object)new Integer(1), true);
            }
        }
        else {
            if (activated) {
                synchronized (this) {
                    activated = false;
                    initTable();
                    // stop timer
                    mytimer.stop();
                    mytimer.unset();
                }
            }
        }
    }

    /**
     * get monitoring activation
     * @return Boolean is activated
     */

    public boolean isActivated() {
        return activated;
    }

    /**
     * private function for monitoring
     * @return long  used memory
     */

    private long usedMemory() {
        long result = -1;
        while (result < 0) {
            result = Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory();
        }

        return result / 1000;
    }

    /**
     * private function for monitoring
     * @return long  total memory
     */

    private long totalMemory() {
        return Runtime.getRuntime().totalMemory() / 1000;
    }

    /**
     * private init table
     */

    private void initTable() {
        tableMeasures = new Long[sizeTableMeasures];
        // init max totalmemory
        maxtotalmemory = totalMemory();
        for (int i = 0; i < tableMeasures.length; i++) {
            tableMeasures[i] = new Long(0);
        }
        // init the first value
        tableMeasures[tableMeasures.length - 1] = new Long(usedMemory());
    }

    /**
     * get the number of active threads in the JOnAS server's initial thread group
     * @return int the number of active threads
     */
    public int getActiveThreadsCount() {
        return server.getActiveThreadsCount();
    }

    /**
     * get the estimated number of active groups in the JOnAS server's initial thread group
     * @return int the estimated  number of active groups
     */
    public int getActiveGroupCount() {
        return server.getActiveGroupCount();
    }
    
    /**
     * Deploy a stand-alone J2EE module packaged in a JAR file
     * @param fileName the file name of the JAR to be be deployed.
     * @return The ObjectName of the MBean associated to the deployed module
     */
    public String deployJar(String fileName) throws Exception {
        return ejbService.createContainer(fileName);
    }
   
    /**
     * Uneploy a stand-alone J2EE module packaged in a JAR file
     * @param fileName the file name of the JAR to be be undeployed.
     */
    public void unDeployJar(String fileName) throws Exception {
        ejbService.removeContainer(fileName);
    }

    /**
     * Deploy a J2EE application packaged in a EAR file
     * @param fileName the file name of the EAR to be be deployed.
     * @return The ObjectName of the MBean associated to the deployed J2EE Application
     */
    public String deployEar(String fileName) throws Exception {
        return earService.deployEar(fileName);
    }

    /**
     * undeploy a J2EE application packaged in a EAR file
     * @param fileName the file name of the EAR to be be undeployed.
     */
    public void unDeployEar(String fileName) throws Exception {
        earService.unDeployEar(fileName);
    }

    /**
     * Deploy a J2EE application packaged in a RAR file
     * @param fileName the file name of the RAR to be be deployed.
     * @return The ObjectName of the MBean associated to the deployed J2EE Application
     */
    public String deployRar(String fileName) throws Exception {
        return rarService.deployRar(fileName);
    }

    /**
     * undeploy a J2EE application packaged in a RAR file
     * @param fileName the file name of the RAR to be be undeployed.
     */
    public void unDeployRar(String fileName) throws Exception {
        rarService.unDeployRar(fileName);
    }

    /**
     * Add an object name to the <code>resources</code> list.
     * @param p_ObjectName Object name correspondig to a J2EEResource MBean
     */
    public void addResource(String p_ObjectName) {
        synchronized (resources) {
            if (resources.contains(p_ObjectName) == true) {
                if (mgtLogger.isLoggable(BasicLevel.DEBUG)) {
                    mgtLogger.log(BasicLevel.DEBUG
                        , "The object name: " + p_ObjectName + " is already in the resources list");
                }
            }
            else {
                // Add
                resources.add(p_ObjectName);
                // Send the notification
                AttributeAddNotification notification = new AttributeAddNotification(getObjectName()
                    , sequenceNumber++, System.currentTimeMillis(), "", "resource"
                    , p_ObjectName.toString());
                sendNotification(notification);
                if (evtLogger.isLoggable(BasicLevel.DEBUG)) {
                    evtLogger.log(BasicLevel.DEBUG
                        , "AttributeAddNotification emitted: resource added to the J2EEServer: "
                        + p_ObjectName.toString());
                }
            }
        }
    }

    /**
     * Remove an object name from the <code>resources</code> list.
     * @param p_ObjectName Object name correspondig to a J2EEResource MBean
     * @return Object name to the removed J2EEResource MBean
     */
    public String removeResource(String p_ObjectName) {
        String sRet = null;
        synchronized (resources) {
            int index = resources.indexOf(p_ObjectName);
            if (index > -1) {
                // Remove
                sRet = (String) resources.remove(index);
                // Send the notification
                AttributeRemoveNotification notification = new AttributeRemoveNotification(
                    getObjectName(), sequenceNumber++, System.currentTimeMillis(), "", "resource"
                    , p_ObjectName.toString());
                sendNotification(notification);
                if (evtLogger.isLoggable(BasicLevel.DEBUG)) {
                    evtLogger.log(BasicLevel.DEBUG
                        , "AttributeRemoveNotification emitted: resource removed from the J2EEServer: "
                        + p_ObjectName.toString());
                }
            }
        }
        return sRet;
    }

    /**
     * Send a notification to the listener.
     * @param notification The notification to send
     */
    public void sendNotification(Notification notification) {
        // if registered listener(s)
        if (myListener != null) {
            NotificationListener listener = myListener.getListener();
            Object handback = myListener.getHandback();
            listener.handleNotification(notification, handback);
            if (evtLogger.isLoggable(BasicLevel.DEBUG)) {
                evtLogger.log(BasicLevel.DEBUG, "Notification sent");
            }
        }
    }

    /**
     * Add a new listener.
     * @param listner Listener to notify
     * @param filter Notification filter
     * @param handback ??
     * @throws java.lang.IllegalArgumentException
     */
    public void addNotificationListener(NotificationListener listner, NotificationFilter filter
        , java.lang.Object handback)
        throws java.lang.IllegalArgumentException {
        // Create new listener
        ListenerJavaBean lmb = new ListenerJavaBean(listner, filter, handback);
        myListener = lmb;
        if (evtLogger.isLoggable(BasicLevel.DEBUG)) {
            evtLogger.log(BasicLevel.DEBUG
                , "Listener added for J2EEServer MBean: " + getObjectName().toString());
            if (handback != null) {
                evtLogger.log(BasicLevel.DEBUG, " - with handback: " + handback.toString());
            }
        }
    }
}
