/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 * --------------------------------------------------------------------------
 * $Id: J2EEServerMBean.java,v 1.9 2004/03/30 23:41:57 ehardesty Exp $
 * --------------------------------------------------------------------------
 */
package org.objectweb.jonas.server;

// Java imports
import java.util.ArrayList;

import javax.management.MBeanException;
import javax.management.Notification;
import javax.management.NotificationFilter;
import javax.management.NotificationListener;

import org.apache.commons.modeler.BaseModelMBean;
import org.objectweb.jonas.common.Log;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * MBean class for Server management.
 * Two classes are used to provide all management services : J2EEServer and J2EEServerMBean.
 * This class provide the notification's process when add or remove resources.
 * @author Adriana Danes
 * @author Michel-Ange Anton
 */

public class J2EEServerMBean extends BaseModelMBean {

// ------------------------------------------------------------- Private variables

    static private Logger mgtLogger = Log.getLogger(Log.JONAS_MANAGEMENT_PREFIX);
    static private Logger evtLogger = Log.getLogger(Log.JONAS_MANAGEMENT_EVENT_PREFIX);

// ------------------------------------------------------------- Contructors

    public J2EEServerMBean()
        throws MBeanException {
        super();
    }

// ------------------------------------------------------------- Public methods

    /**
     * All The MBean names corresponding to the deployed J2EEModules (ear, jar, war, rar).
     * @return The String's array of deployed modules objects names
     */
    public String[] getDeployedObjects() {
        ArrayList al = (ArrayList) ((J2EEServer) (this.resource)).getDeployedObjects();
        return (String[]) al.toArray(new String[al.size()]);
    }

    /**
     * All The MBean names corresponding to the deployed J2EEResources.
     * @return The String's array of deployed resources objects names
     */
    public String[] getResources() {
        ArrayList al = (ArrayList) ((J2EEServer) (this.resource)).getResources();
        return (String[]) al.toArray(new String[al.size()]);
    }

    /**
     * All The MBean names corresponding to the used JVMs.
     * @return The String's array of used JVMs objects names
     */
    public String[] getJavaVMs() {
        ArrayList al = (ArrayList) ((J2EEServer) (this.resource)).getJavaVMs();
        return (String[]) al.toArray(new String[al.size()]);
    }

    /**
     * Send a notification to the listener.
     * @param notification The notification to send
     */
    public void sendNotification(Notification p_Notification) {
        ((J2EEServer) (this.resource)).sendNotification(p_Notification);
    }

    /**
     * Add a new listener.
     * @param listner Listener to notify
     * @param filter Notification filter
     * @param handback ??
     * @throws java.lang.IllegalArgumentException
     */
    public void addNotificationListener(NotificationListener p_Listner, NotificationFilter p_Filter
        , java.lang.Object p_Handback)
        throws java.lang.IllegalArgumentException {
        ((J2EEServer) (this.resource)).addNotificationListener(p_Listner, p_Filter, p_Handback);
    }

    /**
     * Deploy a stand-alone J2EE module packaged in a JAR file
     * @return The ObjectName of the MBean associated to the deployed module
     */
    public String deployJar(String fileName) throws Exception {
        return ((J2EEServer) (this.resource)).deployJar(fileName);
    }

    /**
     * Deploy a J2EE application packaged in a EAR file
     * @return The ObjectName of the MBean associated to the deployed J2EE Application
     */
    public String deployEar(String fileName)  throws Exception {
        return ((J2EEServer) (this.resource)).deployEar(fileName);
    }

    /**
     * Deploy a J2EE application packaged in a RAR file
     * @return The ObjectName of the MBean associated to the deployed J2EE Application
     */
    public String deployRar(String fileName)  throws Exception {
        return ((J2EEServer) (this.resource)).deployRar(fileName);
    }

    /**
     * Add a deployed object in the list of the J2EEServer.
     * @param p_DeployedObject The object name of deployed object
     */
    public static void addDeployedObject(String p_DeployedObject) {
        try {
            Server oServer = Server.getInstance();
            J2EEServer oJ2EEServer = oServer.getJ2EEServer();
            oJ2EEServer.addDeployedObject(p_DeployedObject);
        }
        catch (Exception e) {
            mgtLogger.log(BasicLevel.WARN
                , "Add deployed object error (object=" + p_DeployedObject + ") : "
                + e.getClass().getName() + " " + e.getMessage());
        }
    }

    /**
     * Remove a deployed object from the list of the J2EEServer.
     * @param p_DeployedObject The object name of deployed object
     */
    public static void removeDeployedObject(String p_DeployedObject) {
        try {
            Server oServer = Server.getInstance();
            J2EEServer oJ2EEServer = oServer.getJ2EEServer();
            oJ2EEServer.removeDeployedObject(p_DeployedObject);
        }
        catch (Exception e) {
            mgtLogger.log(BasicLevel.WARN
                , "Remove deployed object error (object=" + p_DeployedObject + ") : "
                + e.getClass().getName() + " " + e.getMessage());
        }
    }

    /**
     * Add an object name to the <code>resources</code> list.
     * @param p_ObjectName Object name correspondig to a J2EEResource MBean
     */
    public static void addResource(String p_ObjectName) {
        try {
            Server oServer = Server.getInstance();
            J2EEServer oJ2EEServer = oServer.getJ2EEServer();
            oJ2EEServer.addResource(p_ObjectName);
        }
        catch (Exception e) {
            mgtLogger.log(BasicLevel.WARN
                , "Add resource error (object=" + p_ObjectName + ") : "
                + e.getClass().getName() + " " + e.getMessage());
        }
    }

    /**
     * Remove an object name from the <code>resources</code> list.
     * @param p_ObjectName Object name correspondig to a J2EEResource MBean
     * @return Object name to the removed J2EEResource MBean
     */
    public static String removeResource(String p_ObjectName) {
        String sRet = null;
        try {
            Server oServer = Server.getInstance();
            J2EEServer oJ2EEServer = oServer.getJ2EEServer();
            sRet = oJ2EEServer.removeResource(p_ObjectName);
        }
        catch (Exception e) {
            mgtLogger.log(BasicLevel.WARN
                , "Remove resource object error (object=" + p_ObjectName + ") : "
                + e.getClass().getName() + " " + e.getMessage());
        }
        return sRet;
    }

}
