/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *
 * --------------------------------------------------------------------------
 * $Id: Server.java,v 1.80 2004/03/25 15:58:20 sauthieg Exp $
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.server;

import java.rmi.RMISecurityManager;
import java.util.StringTokenizer;

import javax.management.MBeanServer;
import javax.management.ObjectName;
import javax.management.modelmbean.ModelMBean;

import org.apache.commons.modeler.ManagedBean;
import org.apache.commons.modeler.Registry;
import org.objectweb.carol.util.configuration.CarolConfiguration;
import org.objectweb.jonas.adm.Adm;
import org.objectweb.jonas.common.JProp;
import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.common.LogManagement;
import org.objectweb.jonas.jmx.J2eeObjectName;
import org.objectweb.jonas.jmx.JmxService;
import org.objectweb.jonas.jmx.JonasObjectName;
import org.objectweb.jonas.management.JonasMBeanTools;
import org.objectweb.jonas.management.ReconfigManager;
import org.objectweb.jonas.management.j2eemanagement.J2EEDomain;
import org.objectweb.jonas.naming.JNDIResource;
import org.objectweb.jonas.naming.JNDIResourceMBean;
import org.objectweb.jonas.naming.NamingManager;
import org.objectweb.jonas.security.PolicyProvider;
import org.objectweb.jonas.service.JonasAlreadyStartedException;
import org.objectweb.jonas.service.Service;
import org.objectweb.jonas.service.ServiceException;
import org.objectweb.jonas.service.ServiceManager;
import org.objectweb.jonas_lib.I18n;
import org.objectweb.jonas_lib.version.Version;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * This class implements an JOnAS Server.
 * @author Philippe Coq, Philippe Durieux
 * 00/06/15: Jeff Mesnil
 * 01/06/15: Regis Le Brettevillois - Libelis / JOnAS team - evidian
 *           Service as org.objectweb.jonas.service.Service
 * 02/04/25: Sebastien Chassande-Barrioz sebastien.chassande@inrialpes.fr
 *           Monolog for traces
 *
 *           Adriana Danes
 * 03/01/20:  -Highlight configuration properties
 * 03/03/..:  -Force jmx service startup
 -Create and register the JonAS configuration MBean. This MBean manages JOnAS persistent re-configuration (see ReconfigManager  class)
*/
public class Server {

    private static Logger logger = null;
    private static String SERVERNAME = "JOnAS"; // J2EE Application Server Name
    private String serverName = null; // Current server name (the server instance name)

    // The Server singleton value
    private static Server unique = null;

    // The associated ServiceManager for JOnAS
    ServiceManager sm = null;

    // Associated Adm object
    Adm adm = null;

    // JOnAS configuration properties
    JProp props = null;

    // Number of active threads in the initial thread's group
    int threadCount = 0;

    // Estimated number of active groups in the initial thread's group
    int groupCount = 0;

    // String protocols
    String protocols = null;

    // Server configuration properties
    static final String ORB_PORT = "jonas.orb.port";
    static final String DEF_ORB_PORT = "0";
    static final String SECURITY_PROPAGATION = "jonas.security.propagation";
    static final String TRANSACTION_PROPAGATION = "jonas.transaction.propagation";
    static final String LOG_CONFIGFILE = "jonas.log.configfile";
    static final String DEF_LOG_CONFIGFILE = "trace";
    static final String CONFIGFILE_USECLASSPATH = "jonas.log.configfile.useclasspath";
    static final String DEF_CONFIGFILE_USECLASSPATH = "true";

    // Server configuration properties
    int orbPort;
    String logConfigFile;


    // J2EE Management
    String domainName = null;

    /**
     * Internationalization
     */
    private static I18n i18n = I18n.getInstance(Server.class);

    /**
     * J2EEServer associate
     */
    private J2EEServer m_J2EEServer = null;

    /**
     * Private constructor to force only 1 instance.
     */
    private Server() throws Exception {
        // read the JOnAS properties
        props = JProp.getInstance();

        // set configuration info into the Log System (monolog)
        logConfigFile = props.getValue(LOG_CONFIGFILE, DEF_LOG_CONFIGFILE);
        Log.configure(logConfigFile);

        // get a logger for server traces
        logger = Log.getLogger(Log.JONAS_SERVER_PREFIX);

        // create the service manager
        sm = ServiceManager.getInstance();

        PolicyProvider policyProvider = new PolicyProvider();
        
        // called only once for a JVM
        PolicyProvider.init();

    }

    /**
     * Get the unique instance.
     * Create it at first call.
     */
    public static Server getInstance() throws Exception {
        if (unique == null) {
            unique = new Server();
        }
        return unique;
    }

    /**
     * Start the server
     */
    public void start() throws Exception {

        // Server name
        serverName = props.getValue(JProp.JONAS_NAME, JProp.JONAS_DEF_NAME);

        // Initialize Context Interceptors and Carol
        String protocols = initCarol(props);

        // RMI Security Manager
        if (System.getSecurityManager() == null) {
            System.setSecurityManager(new RMISecurityManager());
        }

        // First of all, start registry service
        sm.startRegistry();

        // then, start JMX service (start an MBean server and create RMI connector for it registered in JNDI)
        sm.startJmx();

        // get the JMX server reference
        JmxService jmxService = null;
        MBeanServer mbeanServer = null;
        try {
            jmxService = (JmxService) ServiceManager.getInstance().getJmxService();
            mbeanServer = jmxService.getJmxServer();
        } catch (Exception se) {
            // this could mean that jmx service not available
            // this is not possible, as we started it even if it does not apper in the services list configuration parameter
            logger.log(BasicLevel.ERROR, "Could not access the jmx service");
            throw se;
        }

        // -----------------------------------------------------
        // J2EE Management
        // -----------------------------------------------------

        // Get the MBean register (and load mbeans-descriptors)
       Registry oRegistry = JonasMBeanTools.getRegistry();

        // Create and register a MBean for the Management Domain
        domainName = props.getValue(JProp.DOMAIN_NAME);

        ObjectName domain_on = null;
        boolean domainOK;
        try {
            // Create a JMX object name for the Domain MBean
            domain_on = J2eeObjectName.J2EEDomain(domainName);
            J2EEDomain domainMBean = new J2EEDomain(domain_on.toString(), false, false, false);
            mbeanServer.registerMBean(domainMBean, domain_on);
            domainOK = true;
        } catch (javax.management.InstanceAlreadyExistsException ae) {
            // should never occur
        } catch (Exception e) {
            logger.log(BasicLevel.WARN, "Could not register J2EEDomain MBean");
            domainOK = false;
        }

        // create and register necessary MBeans for the server management
        ObjectName server_on = null;
        try {
            // create and register a MBean allowing server reconfiguration management
            ReconfigManager reconfigManagerMBean = new ReconfigManager(props.getConfigFileEnv(), mbeanServer);
            mbeanServer.registerMBean(reconfigManagerMBean, JonasObjectName.serverConfig());

            // create and register a MBean allowing the JOnAS server administration
            server_on = J2eeObjectName.J2EEServer(domainName, serverName);
            m_J2EEServer = new J2EEServer(server_on.toString(), this, serverName,
                Version.NUMBER, Version.VENDOR, protocols);
            ManagedBean oManaged = oRegistry.findManagedBean("J2EEServer");
            ModelMBean oMBean = oManaged.createMBean(m_J2EEServer);
            mbeanServer.registerMBean(oMBean, server_on);

            // create and register a MBean for the logging service administration
            LogManagement logMBean = LogManagement.getInstance();
            mbeanServer.registerMBean(logMBean, JonasObjectName.logService(logConfigFile));
        } catch (Exception e) {
            // Cannot administrate this server through JMX
            jmxService.stop();
        }

        // Update J2EEDomain MBean
        // ------------------------
        try {
            if (domainOK = true) {
                String operationName = "addServer";
                Object[] params = { server_on.toString() };
                String[] signature = { "java.lang.String" };
                // use the domain_on ObjectName
                // this code is related to the temporary implementation as said before
                mbeanServer.invoke(domain_on, operationName, params, signature);
            }
        } catch (Exception e) {
            logger.log(BasicLevel.WARN, "Could not update J2EEDomain MBean with J2EE server object name " + server_on.toString());
        }

        // Create the MBean for the corresponding JVM
        ObjectName onJvm = null;
        try {
            onJvm = J2eeObjectName.JVM(domainName, serverName, serverName);
            ManagedBean oManaged = oRegistry.findManagedBean("JVM");
            JavaVm oJvm = new JavaVm(onJvm.toString(), props);
            ModelMBean oMBean = oManaged.createMBean(oJvm);
            mbeanServer.registerMBean(oMBean, onJvm);
            // Update J2EEServer MBean
            m_J2EEServer.addJavaVM(onJvm.toString());
        }
        catch (Exception e) {
            logger.log(BasicLevel.WARN, "Could not create JVM MBean");
        }

        // create and register some resources MBean
        ObjectName on = null;
        try {
            // JNDI resource creation + registration
            NamingManager naming = NamingManager.getInstance();
            JNDIResourceMBean resMBean = new JNDIResource(naming.getInitialContext());
            on = J2eeObjectName.JNDIResource(domainName, "jndi", serverName);
            mbeanServer.registerMBean(resMBean, on);
            // update the server's MBean
            J2EEServerMBean.addResource(on.toString());
        } catch (Exception e) {
            logger.log(BasicLevel.WARN, "Problem with J2EEResource MBean " + on.toString() + " creation or registration: " + e);
        }

        // Creates the Adm instance for remote administration
        try {
            adm = new Adm(props);
        } catch (Exception e) {
            e.printStackTrace();
            throw new Exception("Cannot create Adm : " + e);
        }

        // start all other services
        sm.startServices();

        // TM Server or EJB Server ?
        boolean isEJB = true;
        try {
            Service ejbservice = sm.getEjbService();
        } catch (ServiceException se) {
            isEJB = false;
        }
        adm.serverReady(isEJB);
        System.out.println(i18n.getMessage("Server.start.serverIsReady", serverName, Version.NUMBER));
        logger.log(BasicLevel.INFO, i18n.getMessage("Server.start.serverStarted", serverName, protocols));

        // Periodically count the number of active threads in this thread's group
        Thread thisThread = Thread.currentThread();
        ThreadGroup thGroup = thisThread.getThreadGroup();
        while (true) {
            threadCount = Thread.activeCount();
            groupCount = thGroup.activeGroupCount();
            Thread.sleep(10000);
        }
    }

    /**
     * get the number of active threads in the JOnAS server's initial thread group
     */
    protected int getActiveThreadsCount() {
        return threadCount;
    }
    /**
     * get the estimated number of active groups in the JOnAS server's initial thread group
     */
    public int getActiveGroupCount() {
        return groupCount;
    }

    /**
     * Get J2EEServer.
     */
    public J2EEServer getJ2EEServer() {
        return m_J2EEServer;
    }

    /**
     * initialize interception and carol
     */
    public String  initCarol(JProp props) throws Exception {

        // Intialize Carol
        CarolConfiguration.init(Log.getLoggerFactory());

        // Initialize Interceptors
        String sec = props.getValue(SECURITY_PROPAGATION, "false").trim();
        String trans = props.getValue(TRANSACTION_PROPAGATION, "false").trim();
        if (sec.equals("true")) {
            // jrmp interceptors
            CarolConfiguration.addInterceptors("jrmp", "org.objectweb.jonas.security.interceptors.jrmp.SecurityInitializer");
            // cmi interceptors
            CarolConfiguration.addInterceptors("cmi", "org.objectweb.jonas.security.interceptors.jrmp.SecurityInitializer");
            // iiop interceptors flag
            CarolConfiguration.addInterceptors("iiop", "org.objectweb.jonas.security.interceptors.iiop.SecurityInitializer");

        }

        if (trans.equals("true")) {
            // jrmp interceptors flag
            CarolConfiguration.addInterceptors("jrmp", "org.objectweb.jotm.jta.rmi.JTAInterceptorInitializer");
            // jrmp interceptors flag
            CarolConfiguration.addInterceptors("cmi", "org.objectweb.jotm.jta.rmi.JTAInterceptorInitializer");
            // iiop interceptors flag
            CarolConfiguration.addInterceptors("iiop", "org.objectweb.jotm.ots.OTSORBInitializer");
        }

        // intialize port number
        System.setProperty("rmi.server.port", props.getValue(ORB_PORT, DEF_ORB_PORT));

        String s = CarolConfiguration.getProtocols();
        StringTokenizer st = new StringTokenizer(s, ",", false);
        String pn = " ";

        int idx = 0;
        while (st.hasMoreElements()) {
            if (idx != 0) {
                pn = pn + ", ";
            }
            pn = pn + "rmi/" + st.nextToken().trim();
            idx++;
        }
        return pn;
    }

    /**
     * Stop the server without stopping the JVM
     */
    public void stop() throws Exception {
        if (adm != null) {
            System.out.println(i18n.getMessage("Server.stop.serverIsStopped", serverName));
            logger.log(BasicLevel.INFO, i18n.getMessage("Server.stop.serverStopped", serverName));
            adm.stopServer();
        }
    }

    /**
     * Stop the server and stop the JVM (MBean method)
     */
    public void kill() throws Exception {
        if (adm != null) {
            adm.killServer();
        }
    }


    /**
     * Server main routine.
     * No args.
     */
    public static void main(String args[]) {
        Server server = null;
        try {
            server = Server.getInstance();
            server.start();
        }  catch (JonasAlreadyStartedException e) {
            // if we get this exception this means that
            // we are trying to restart the same JOnAS
            System.err.println(i18n.getMessage("Server.main.alreadyStarted", SERVERNAME));
            System.exit(2);
        } catch (Exception e) {
            // We don't use here the trace system.
            System.err.println("JOnAS error: " + e);
            e.printStackTrace();
            System.err.println(SERVERNAME + " halting");
            try {
                if (server != null) {
                    server.stop();
                }
            } catch (Exception ignore) {
            }
            System.exit(2);
        }
    }

}
