/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Philippe Coq
 * --------------------------------------------------------------------------
 * $Id: JOnASStandardContext.java,v 1.6 2004/03/25 15:58:20 sauthieg Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.web.catalina50;

import java.net.URL;

import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.service.ServiceManager;

import org.apache.catalina.LifecycleException;
import org.apache.catalina.Valve;
import org.apache.catalina.authenticator.AuthenticatorBase;
import org.apache.catalina.core.StandardContext;

/**
 * Define a JOnAS context. It is used to check if a context has been
 * defined in server.xml and to use this context while deploying the war
 * instead of creating a new one
 * @author Florent Benoit
 * @author Philippe Coq (Tomcat 4.0)
 */

public final class JOnASStandardContext extends StandardContext {

    /**
     * Logger for this service.
     */
    private static Logger logger = null;

    /**
     * Unique instance of Catalina
     */
    private static CatalinaJWebContainerServiceImpl catalinaService = null;

    /**
     * This context was configured in server.xml ?
     */
    private boolean inServerXml = false;

    /**
     * JOnAS deployment descriptor
     */
    private String jonasDeploymentDescriptor = null;


    /**
     * Context was started ?
     */
    private boolean startedJStdx = false;

    private URL earURL = null;

    private URL warURL = null;

    private boolean inEarCase = false;

    private boolean java2DelegationModel = false;

    /**
     * Constructor of the Context
     * @param inServerXml this context was defined in server.xml
     */
    public JOnASStandardContext(boolean inServerXml, boolean java2DelegationModel, boolean inEarCase) {
        super();
        this.inServerXml = inServerXml;
        this.java2DelegationModel = java2DelegationModel;
        this.inEarCase = inEarCase;
        logger = Log.getLogger(Log.JONAS_WEB_PREFIX);
        try {
            catalinaService = (CatalinaJWebContainerServiceImpl) ServiceManager.getInstance().getWebContainerService();
        } catch (Exception e) {
            // should bever occurs
        }
    }

    /**
     * Constructor of the Context
     */
    public JOnASStandardContext() {
        this(true, true, false);
    }

    /**
     * Start the JOnAS context if catalina is started
     * @throws LifecycleException if the context can't be started
     */
    public synchronized void start() throws LifecycleException {
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "");
        }
        if (catalinaService != null && catalinaService.isTomcatStarted()) {
            if (logger.isLoggable(BasicLevel.DEBUG)) {
                logger.log(BasicLevel.DEBUG, "Tomcat in Web container service is started, starting the context...");
            }
            startedJStdx = true;
            super.start();
        }

    }

    /**
     * Stop this Context component.
     *
     * @exception LifecycleException if a shutdown error occurs
     */
    public synchronized void stop() throws LifecycleException {
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "");
        }
        if (startedJStdx) {
            startedJStdx = false;
            super.stop();
        }


    }

    /**
     * This context was defined in server.xml ?
     * @return true if this context was defined in server.xml
     */
    public boolean isInServerXml() {
        return inServerXml;
    }


    /**
     * Disable cache of the authenticator
     */
    public void unsetAuthenticationCaching() {
        Valve[] valves = getValves();
        for (int i = 0; i < valves.length; i++) {
            Valve valve = valves[i];
            if (valve instanceof AuthenticatorBase) {
                AuthenticatorBase authBase = (AuthenticatorBase) valve;
                authBase.setCache(false);
            }
        }
    }

    /**
     * Set the deployment descriptor content of jonas-web.xml file
     * @param jonasDeploymentDescriptor the content of jonas-web.xml
     */
    public void setJonasDeploymentDescriptor(String jonasDeploymentDescriptor) {
        this.jonasDeploymentDescriptor = jonasDeploymentDescriptor;
    }


    /**
     * Gets the deployment descriptor content of jonas-web.xml file
     * @return the content of jonas-web.xml file
     */
    public String getJonasDeploymentDescriptor() {
        return jonasDeploymentDescriptor;
    }

    public void setEarURL(URL earURL) {
        this.earURL = earURL;
    }

    public URL getEarURL() {
        return earURL;
    }

    public void setWarURL(URL warURL) {
        this.warURL = warURL;
    }

    public URL getWarURL() {
        return warURL;
    }

    /**
     * @return true if this web module is a part of a EAR
     */
    public boolean isInEarCase() {
        return inEarCase;
    }

    /**
     * @return true if the Java2 delegation model is used
     */
    public boolean isJava2DelegationModel() {
        return java2DelegationModel;
    }
}
