/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial Developer : Delplanque Xavier & Sauthier Guillaume
 * --------------------------------------------------------------------------
 * $Id: AbsWebServicesServiceImpl.java,v 1.16 2004/03/25 15:58:19 sauthieg Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.ws;

import java.io.File;
import java.io.UnsupportedEncodingException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.net.URLEncoder;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

import javax.naming.Context;
import javax.naming.NamingException;
import javax.xml.namespace.QName;

import org.objectweb.jonas.common.JProp;
import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.naming.CompNamingContext;
import org.objectweb.jonas.service.AbsServiceImpl;
import org.objectweb.jonas.service.ServiceException;
import org.objectweb.jonas.service.ServiceManager;
import org.objectweb.jonas.web.AbsJWebContainerServiceImpl;
import org.objectweb.jonas.web.JWebContainerServiceException;
import org.objectweb.jonas.ws.handler.WSDLHandler;
import org.objectweb.jonas.ws.handler.WSDLHandlerFactory;
import org.objectweb.jonas_lib.I18n;
import org.objectweb.jonas_web.deployment.api.WebContainerDeploymentDesc;
import org.objectweb.jonas_web.deployment.api.WebContainerDeploymentDescException;
import org.objectweb.jonas_web.deployment.lib.WebDeploymentDescManager;
import org.objectweb.jonas_ws.deployment.api.JaxRpcPortComponentDesc;
import org.objectweb.jonas_ws.deployment.api.PortComponentDesc;
import org.objectweb.jonas_ws.deployment.api.SSBPortComponentDesc;
import org.objectweb.jonas_ws.deployment.api.ServiceDesc;
import org.objectweb.jonas_ws.deployment.api.WSDLFile;
import org.objectweb.jonas_ws.deployment.api.WSDeploymentDesc;
import org.objectweb.jonas_ws.deployment.api.WSDeploymentDescException;
import org.objectweb.jonas_ws.deployment.lib.WSDeploymentDescManager;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;


/**
 * implements deployment process and declares commons abstract methods for all
 * (used) Web services engine. Actually we consider Glue and AXIS.
 *
 * @author Guillaume Sauthier
 * @author Xavier Delplanque
 */
public abstract class AbsWebServicesServiceImpl extends AbsServiceImpl
    implements WebServicesService, AbsWebServicesServiceImplMBean {

    /**
     * WebServices service configuration properties : Xml parsing with
     * validation
     */
    public static final String PARSINGWITHVALIDATION =
        "jonas.service.ws.parsingwithvalidation";

    /** WebServices service configuration properties : WSDL handlers */
    public static final String WSDL_HANDLERS = "jonas.service.ws.wsdlhandlers";

    /**
     * WebServices service configuration properties : base for web services
     * properties.
     */
    public static final String WS_BASE = "jonas.service.ws.";

    /** Logger for this service. */
    protected static Logger logger = null;

    /** Internationalization tool */
    protected static I18n i18n = I18n.getInstance(AbsWebServicesServiceImpl.class);

    /** Manager for WSDL files publishing. */
    protected WSDLManager wsdlManager = null;

    /** Reference to the ws deployment descriptor manager. */
    private WSDeploymentDescManager wsDDManager = null;

    /** Reference to the web container service */
    private AbsJWebContainerServiceImpl webService = null;

    /**
     * Initialize the service.
     *
     * @param ctx the configuration context of the service.
     *
     * @throws ServiceException if the initialization failed.
     */
    protected void doInit(Context ctx) throws ServiceException {
        // Init the logger
        logger = Log.getLogger(Log.JONAS_WS_PREFIX);

        // Set the XML parsing mode to no validation
        String parsingMode = "false";

        try {
            parsingMode = (String) ctx.lookup(PARSINGWITHVALIDATION);
        } catch (NamingException e) {
            ;
            // No problem if there is no value for 'parsingwithvalidation' 
            // (false by default)
        }

        WSDeploymentDescManager.setParsingWithValidation("true"
                                                         .equalsIgnoreCase(parsingMode));

        if ("false".equalsIgnoreCase(parsingMode)) {
            logger.log(BasicLevel.DEBUG,
                       "WebServices XML parsing without validation");
        } else {
            logger.log(BasicLevel.DEBUG,
                       "WebServices XML parsing with validation");
        }

        // Init the web services deployment desc managers.
        wsDDManager = WSDeploymentDescManager.getInstance();

        // Init the wsdl Manager
        wsdlManager = new WSDLManager();

        // Get publishing parameters in jonas properties
        JProp jp = null;

        try {
            jp = JProp.getInstance();
        } catch (Exception e) {
            String err = i18n.getMessage("AbsWebServicesServiceImpl.doInit.noJProp");
            logger.log(BasicLevel.ERROR, err + " " + e.getMessage());
            throw new WSServiceException(err, e);
        }

        String[] handlers = jp.getValueAsArray(WSDL_HANDLERS);

        if (handlers == null) {
            String err = i18n.getMessage("AbsWebServicesServiceImpl.doInit.noHandlers",
                                         WSDL_HANDLERS);
            logger.log(BasicLevel.ERROR, err);
            throw new WSServiceException(err);
        }

        // Build WSDLHandlers

        WSDLHandlerFactory factory = WSDLHandlerFactory.newInstance();

        // build each specified handlers and add it in the WSDLManager.
        for (int i = 0; i < handlers.length; i++) {

            // Create the handler
            WSDLHandler handler = factory.newHandler(handlers[i]);

            if (logger.isLoggable(BasicLevel.DEBUG)) {
                logger.log(BasicLevel.DEBUG, "Adding WSDLHandler '" + handlers[i] + "'");
            }
            
            // Add the created Handler
            wsdlManager.addHandler(handler);
        }
    }

    /**
     * Get, update and publish WSDL documents contained in a set of jars and
     * wars.
     *
     * @param ctx the context containing the configuration  to deploy the
     *        wars.<BR> This context contains the following parameters :<BR> 
     *        - jarUrls the list of the urls of the jars to deploy.<BR> 
     *        - warUrls the list of the urls of the wars to deploy.<BR> 
     *        - ejbClassLoader the classLoader of the ejbs(used for webapps too).<BR> 
     *        - earClassLoader the ear classLoader of the j2ee app.<BR> 
     *        - warCtxRootMapping the webapps contextRoots with associated war
     *        name. <BR> 
     *        - unpackDir the directory where EAR has been unpacked
     *        (optionnal for pure webapp webservices).
     *
     * @throws WSServiceException if an error occurs during  the deployment.
     */
    public void deployWebServices(Context ctx) throws WSServiceException {
        /**
         * For each Jar/WarUrls in the given ctx : 
         * - Get the jar/war deployment desc 
         * - Get the WSDL definition 
         * - Update the WSDL location 
         * - Publish the WSDL
         */

        // Cannot get webContainer at startup (cycles)
        if (webService == null) {
            try {
                webService =
                    (AbsJWebContainerServiceImpl) ServiceManager.getInstance().getService("web");
            } catch (Exception e) {
                String err = i18n.getMessage("AbsWebServicesServiceImpl.deployWebServices.noWeb");
                logger.log(BasicLevel.ERROR, err);
                throw new WSServiceException(err, e);
            }
        }

        // Get context parameters.
        URL[] jarUrls = null;
        URL[] warUrls = null;
        URL earURL = null;
        ClassLoader ejbClassLoader = null;
        ClassLoader earClassLoader = null;
        Map warCtxRootMapping = null;
        String unpackDir = null;

        // EAR & Webapp case common information
        try {
            jarUrls = (URL[]) ctx.lookup("jarUrls");
            warUrls = (URL[]) ctx.lookup("warUrls");
            unpackDir = (String) ctx.lookup("unpackDir");
        } catch (NamingException e) {
            String err = i18n.getMessage("AbsWebServicesServiceImpl.deployWebServices.ctxParamProblem");
            logger.log(BasicLevel.ERROR, err + e.getMessage());
            throw new WSServiceException(err, e);
        }

        // EAR case specific information
        try {
            earURL = (URL) ctx.lookup("earURL");
            earClassLoader = (ClassLoader) ctx.lookup("earClassLoader");
            warCtxRootMapping = (Map) ctx.lookup("warCtxRootMapping");
            ejbClassLoader = (ClassLoader) ctx.lookup("ejbClassLoader");
        } catch (NamingException ne) {
            ;
            // WebApp case
        }

        WSDeploymentDesc wsDD = null;

        // Deploy wars Web services.
        URLClassLoader loaderForCls = null;
        String earAppName = null;
        if (earURL != null) {
            String earName = new File(earURL.getFile()).getName();
            earAppName = earName.substring(0, earName.length() - 4);
        }

        String fileName = null;

        for (int i = 0; i < warUrls.length; i++) {
            // Get the name of a war to deploy.
            fileName = warUrls[i].getFile();
            logger.log(BasicLevel.DEBUG,
                       "Analyzing war '" + fileName + "' for web services");

            // Get the class loader for the current war.
            loaderForCls = webService.getClassLoader(warUrls[i],
                                                     earAppName,
                                                     ejbClassLoader);

            // Get the deployment descriptor from file.
            try {
                wsDD = wsDDManager.getDeploymentDesc(warUrls[i],
                                                     loaderForCls,
                                                     earClassLoader);
            } catch (WSDeploymentDescException wsdde) {
                String err = i18n.getMessage("AbsWebServicesServiceImpl.deployWebServices.wsddEx",
                                             fileName);
                logger.log(BasicLevel.ERROR, err + ": " + wsdde);
                throw new WSServiceException(err, wsdde);
            }

            if (wsDD != null) {
                // The current War contains Webservices
                try {
                    Context compCtx = new CompNamingContext(fileName);
                    compCtx.rebind("wsDD", wsDD);
                    compCtx.rebind("cl", loaderForCls);
                    if (earClassLoader != null) {
                        compCtx.rebind("earCL", earClassLoader);
                    }
                    compCtx.rebind("warURL", warUrls[i]);

                    // ear case
                    if (warCtxRootMapping != null) {
                        compCtx.rebind("warContextRoot",
                                       warCtxRootMapping.get(warUrls[i]));

                        // non-ear case
                    } else {
                        compCtx.rebind("warContextRoot", "");
                    }

                    doDeployWebServices(compCtx);
                } catch (NamingException ne) {
                    String err = i18n.getMessage("AbsWebServicesServiceImpl.deployWebServices.bindError",
                                                 fileName);
                    logger.log(BasicLevel.ERROR, err);
                    throw new WSServiceException(err, ne);
                }
            }
        }
        // End deploy wars Web services.

        // Deploy jars Web services.
        // Check warCtxRootMapping presence
        if ((jarUrls.length != 0) && (warCtxRootMapping == null)) {
            String err = i18n.getMessage("AbsWebServicesServiceImpl.deployWebServices.ctxRootMappingMissing");
            logger.log(BasicLevel.ERROR, err);
            throw new WSServiceException(err);
        }

        for (int i = 0; i < jarUrls.length; i++) {
            // Get the name of a war to deploy.
            fileName = jarUrls[i].getFile();
            logger.log(BasicLevel.DEBUG,
                       "Analyzing jar '" + fileName + "' for web services");

            URLClassLoader webClassLoader = null;

            // Get the deployment descriptor from file.
            try {
                wsDD = wsDDManager.getDeploymentDesc(jarUrls[i],
                                                     ejbClassLoader,
                                                     earClassLoader);
            } catch (WSDeploymentDescException wsdde) {
                String err = i18n.getMessage("AbsWebServicesServiceImpl.deployWebServices.wsddEx",
                                             fileName);
                logger.log(BasicLevel.ERROR, err + ": " + wsdde);
                throw new WSServiceException(err, wsdde);
            }

            if (wsDD != null) {
                // The current Jar contains Webservices
                URL warURL = null;

                try {
                    /*
                     * Default Naming Convention for Webapp linked to Ejb
                     * <ejb-name>.jar -> <ejb-name>.war
                     */
                    if (wsDD.getWarFile() != null) {
                        warURL =
                            new URL(unpackDir + "/" + wsDD.getWarFile());
                    } else {
                        String ejb = new File(jarUrls[i].getFile()).getName();
                        String war = ejb.substring(0, ejb.length() - 4);
                        warURL = new URL(unpackDir + "/" + war + ".war");
                    }

                    // check if webapp is present
                    boolean present = false;

                    for (int w = 0; w < warUrls.length; w++) {
                        if (warUrls[w].equals(warURL)) {
                            present = true;
                        }
                    }

                    // get the linked web classloader
                    webClassLoader = webService.getClassLoader(warURL,
                                                               earAppName,
                                                               ejbClassLoader);

                    if (!present) {
                        String err = i18n.getMessage("AbsWebServicesServiceImpl.deployWebServices.webappNotFound",
                                                     warURL);
                        logger.log(BasicLevel.ERROR, err);
                        throw new WSServiceException(err);
                    }
                } catch (MalformedURLException mue) {
                    String err = i18n.getMessage("AbsWebServicesServiceImpl.deployWebServices.mue");
                    logger.log(BasicLevel.ERROR, mue.getMessage());
                    throw new WSServiceException(err, mue);
                }

                try {
                    Context compCtx = new CompNamingContext(fileName);
                    compCtx.rebind("wsDD", wsDD);
                    
                    // Using WebClassLoader ()
                    compCtx.rebind("cl", webClassLoader);
                    compCtx.rebind("earCL", earClassLoader);

                    compCtx.rebind("warURL", warURL);

                    // warContextRoot cannot be null in ejb case
                    // because an ejb needs a webapp for webservices
                    compCtx.rebind("warContextRoot",
                                   warCtxRootMapping.get(warURL));
                    doDeployWebServices(compCtx);
                } catch (NamingException ne) {
                    String err = i18n.getMessage("AbsWebServicesServiceImpl.deployWebServices.bindError",
                                                 fileName);
                    logger.log(BasicLevel.ERROR, err);
                    throw new WSServiceException(err, ne);
                }
            }
        }

        // End deploy jars Web services.
    }

    /**
     * Deploy the given Web services.
     *
     * @param ctx Context used for parameter passing.
     *
     * @throws WSServiceException when error occurs.
     */
    protected void doDeployWebServices(Context ctx) throws WSServiceException {
        /* Context structure :
         * - wsDD  Web Services deployment descriptor.
         * - cl    Module ClassLoader
         * - earCL Application ClassLoader
         * - warContextRoot the context-root of the war (null in ejb case)
         * - warURL the url of war dispatching SOAP requests
         */
        WSDeploymentDesc wsDD = null;
        ClassLoader cl = null;
        ClassLoader earCL = null;
        String warContextRoot = null;
        URL warURL = null;

        try {
            wsDD = (WSDeploymentDesc) ctx.lookup("wsDD");
            cl = (ClassLoader) ctx.lookup("cl");
            warContextRoot = (String) ctx.lookup("warContextRoot");
            warURL = (URL) ctx.lookup("warURL");
        } catch (NamingException ne) {
            String err = i18n.getMessage("AbsWebServicesServiceImpl.doDeployWebServices.namingError");
            logger.log(BasicLevel.ERROR, err);
            throw new WSServiceException(err, ne);
        }

        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "Deploying WebServices for '" + warURL + "'");
        }

        try {
            earCL = (ClassLoader) ctx.lookup("earCL");
        } catch (NamingException ne) {
            ;
            // Nothing to do : non EAR case
        }

        List sds = wsDD.getServiceDescs();

        // for each WSDL contained in the component
        for (Iterator i = sds.iterator(); i.hasNext();) {
            ServiceDesc sd = (ServiceDesc) i.next();
            WSDLFile wsdl = sd.getWSDL();

            // get the endpoint URL for each port-component
            // update the WSDL Definition
            for (Iterator pc = sd.getPortComponents().iterator(); pc.hasNext();) {
                PortComponentDesc pcd = (PortComponentDesc) pc.next();
                String name = pcd.getName();
                QName portQName = pcd.getQName();
                URL endpoint = null;
                if (pcd instanceof JaxRpcPortComponentDesc)
                    endpoint = getEndpointURL((JaxRpcPortComponentDesc) pcd,
                                              warURL,
                                              warContextRoot);
                else
                    endpoint = getEndpointURL((SSBPortComponentDesc) pcd,
                                              warURL,
                                              cl,
                                              earCL,
                                              warContextRoot);
                    
                pcd.setEndpointURL(endpoint);
                wsdl.setLocation(portQName, endpoint);
            }

            // publish WSDL
            wsdlManager.publish(wsdl);
        }
    }

    /**
     * DOCUMENT ME!
     *
     * @param cl DOCUMENT ME!
     * @param wsdd DOCUMENT ME!
     */
    protected void deployEndpoint(
                                  ClassLoader      cl,
                                  WSDeploymentDesc wsdd) {
    }


    private URL getEndpointURL(JaxRpcPortComponentDesc jpcd,
                               URL warURL,
                               String contextRoot) 
        throws WSServiceException {

        logger.log(BasicLevel.DEBUG, "get endpoint for JaxRpc Port Component");
        
        return getEndpointURL(jpcd.getWebDesc(), 
                              warURL,
                              contextRoot,
                              jpcd.getServiceDesc().getName(),
                              jpcd.getName());

    }

    private URL getEndpointURL(SSBPortComponentDesc spcd,
                               URL warURL,
                               ClassLoader cl,
                               ClassLoader earCL,
                               String contextRoot) 
        throws WSServiceException {

        logger.log(BasicLevel.DEBUG, "get endpoint for StatelessSessionBean Port Component");

        WebContainerDeploymentDesc webDD = null;

        try {
            webDD =
                WebDeploymentDescManager.getInstance().getDeploymentDesc(warURL,
                                                                         cl,
                                                                         earCL);
        } catch (WebContainerDeploymentDescException e) {
            String err = i18n.getMessage("AbsWebServicesServiceImpl.getEndpointURL.webDDException",
                                         warURL.getFile());
            logger.log(BasicLevel.ERROR, err);
            throw new WSServiceException(err, e);
        }
        
        return getEndpointURL(webDD, 
                              warURL,
                              contextRoot,
                              spcd.getServiceDesc().getName(),
                              spcd.getName());

    }

    /**
     * Return the URL used to access the endpoint.
     *
     * @param webDD the WebContainer DD of the servlet
     * @param warURL the url of the war containing the servlet  dispatching the
     *        SOAP request to the servant.
     * @param warContextRoot the application defined context-root for the
     *        webapp (can be null).
     * @param servletName the name of servlet dispatching SOAP requets.
     * @param pcName port component name
     *
     * @return the URL used to access the endpoint.
     *
     * @throws WSServiceException When cannot create the endpoint URL.
     */
    private URL getEndpointURL(WebContainerDeploymentDesc webDD,
                               URL                        warURL,
                               String                     warContextRoot,
                               String                     servletName,
                               String                     pcName) throws WSServiceException {

        // Resolve :
        // - hostname
        // - http/https port
        // Priority order : 1. the ones specified in jonas-web.xml
        //                  2. default web container values (if any)

        logger.log(BasicLevel.DEBUG, "SOAP Servlet name '" + servletName + "'");

        String hostname = webDD.getHost();
        String port = webDD.getPort();
        String scheme = "";
        String contextRoot = getContextRoot(warContextRoot, webDD, warURL);
        String mapping = getServletMapping(servletName, webDD);

        // get default host value
        if (hostname == null) {
            // get hostname
            try {
                hostname = webService.getDefaultHost();
            } catch (JWebContainerServiceException e) {
                String err = i18n.getMessage("AbsWebServicesServiceImpl.getEndpointURL.noDefaultHost",
                                             warURL.getFile());
                logger.log(BasicLevel.ERROR, err);
                throw new WSServiceException(err, e);
            }
        }

        // get default port and scheme value
        if (port == null) {
            // get Http informations
            try {
                port = webService.getDefaultHttpPort();
                scheme = "http";
            } catch (JWebContainerServiceException e) {
                // Http Fails
                // Try Https informations
                try {
                    port = webService.getDefaultHttpsPort();
                    scheme = "https";
                } catch (JWebContainerServiceException e2) {
                    String err = i18n.getMessage("AbsWebServicesServiceImpl.getEndpointURL.noDefaultHTTPPort",
                                                 warURL.getFile());
                    logger.log(BasicLevel.ERROR, err);
                    throw new WSServiceException(err, e2);
                }
            }
        }

        // construct URL
        port = (port.equals("80")) ? "" : (":" + port);

        String encodedPortName;
        try {
            encodedPortName = URLEncoder.encode(pcName, "UTF-8");
        } catch (UnsupportedEncodingException e) {
            // should never occurs since utf-8 is mandatory on all JVM
            encodedPortName = null;
        }
        String url =
            scheme + "://" + hostname + port + "/" + contextRoot + "/"
            + mapping + "/" + encodedPortName;

        URL endpoint = null;

        try {
            endpoint = new URL(url);
        } catch (MalformedURLException mue) {
            String err = i18n.getMessage("AbsWebServicesServiceImpl.getEndpointURL.endpointURLError",
                                         pcName,
                                         url);
            logger.log(BasicLevel.ERROR, err);
            throw new WSServiceException(err, mue);
        }

        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG,
                       "Constructed URL for '" + pcName + "' : '"
                       + endpoint + "'");
        }

        return endpoint;
    }

    /**
     * Return the contextRoot used for webapp access.
     *
     * @param ctxRoot The application specified context-root ("" if not).
     * @param webDD The web DeploymentDesc of the dispatching webapp.
     * @param warURL url of the webapp file.
     *
     * @return The contextRoot used for webapp access.
     */
    private String getContextRoot(
                                  String                     ctxRoot,
                                  WebContainerDeploymentDesc webDD,
                                  URL                        warURL) {
        // Set the right context root for the web application, the priority is
        // the following :
        //    1 - context-root of application.xml
        //    2 - context-root of jonas-web.xml
        //    3 - context-root is the name of the file without .war.
        String contextRoot = null;

        if (ctxRoot == "") {
            String cRoot = webDD.getContextRoot();

            if (cRoot == null) {
                String file = new File(warURL.getFile()).getName();

                if (file.toLowerCase().endsWith(".war")) {
                    contextRoot = file.substring(0, file.length() - 4);
                } else {
                    //It's a directory which is deployed
                    contextRoot = file.substring(0, file.length());
                }
            } else {
                contextRoot = cRoot;
            }
        } else {
            contextRoot = ctxRoot;
        }

        return contextRoot;
    }

    /**
     * DOCUMENT ME!
     */
    public void locateWSDL() {
    }

    /**
     * Publish the specified WSDL Definition.
     *
     * @param wsdl The WSDL Definition to be published
     *
     * @see org.objectweb.jonas.ws.WSDLManager
     */
    public void publishWSDL(WSDLFile wsdl) {
    }

    /**
     * Empty Method for WebServices (no running application).
     *
     * @throws ServiceException never thrown
     */
    public void doStop() throws ServiceException {
    }

    /**
     * Empty Method for WebServices (no running application).
     *
     * @throws ServiceException never thrown
     */
    public void doStart() throws ServiceException {
    }

    /**
     * Return the unique url servlet mapping.
     *
     * @param servlet the servlet name.
     * @param webDD the web DeploymentDesc where mappings are stored.
     *
     * @return the unique url servlet mapping.
     *
     * @throws WSServiceException when multiple (or 0) mappings are found.
     */
    private String getServletMapping(String servlet,
                                     WebContainerDeploymentDesc webDD)
        throws WSServiceException {

        List mappings = webDD.getServletMappings(servlet);

        logger.log(BasicLevel.DEBUG, "mapping : " + mappings);

        if (mappings.size() != 1) {
            String err = i18n.getMessage("AbsWebServicesServiceImpl.getServletMapping.1mappingOnly",
                                         servlet);
            logger.log(BasicLevel.ERROR, err);
            throw new WSServiceException(err);
        }

        String mapping = (String) mappings.get(0);

        // keep only the first part of the mapping :
        // /services/*  becomes services
        StringTokenizer st = new StringTokenizer(mapping, "/");
        mapping = st.nextToken();

        return mapping;

    }

    public void removeCache(ClassLoader cl) {
        wsDDManager.removeCache(cl);
    }
}
