/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer: Florent BENOIT
 * --------------------------------------------------------------------------
 * $Id: ClientContainerDeploymentDesc.java,v 1.7 2004/03/19 14:31:53 sauthieg Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_client.deployment.api;


// import java
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.util.jar.JarFile;
import java.util.zip.ZipEntry;

import org.objectweb.jonas_client.deployment.rules.ApplicationClientRuleSet;
import org.objectweb.jonas_client.deployment.rules.JonasClientRuleSet;
import org.objectweb.jonas_client.deployment.xml.ApplicationClient;
import org.objectweb.jonas_client.deployment.xml.JonasClient;
import org.objectweb.jonas_client.deployment.xml.JonasSecurity;
import org.objectweb.jonas_lib.deployment.api.DeploymentDescException;
import org.objectweb.jonas_lib.deployment.api.EjbRefDesc;
import org.objectweb.jonas_lib.deployment.api.EnvEntryDesc;
import org.objectweb.jonas_lib.deployment.api.JDigester;
import org.objectweb.jonas_lib.deployment.api.JndiEnvRefsGroupDesc;
import org.objectweb.jonas_lib.deployment.api.ResourceEnvRefDesc;
import org.objectweb.jonas_lib.deployment.api.ResourceRefDesc;
import org.xml.sax.EntityResolver;
import org.xml.sax.ErrorHandler;

/**
 * This class do the parsing of the application-client.xml file and jonas-client.xml files and
 * contruct a data structure associated to these two files.
 * @author Florent Benoit
 * @author Philippe Coq
 */
public class ClientContainerDeploymentDesc extends JndiEnvRefsGroupDesc {

    /** 
     * The path to the application-client.xml file. 
     */
    public static final String CLIENT_FILE_NAME = "META-INF/application-client.xml";

    /** 
     * The path to the jonas-client.xml file. 
     */
    public static final String JONAS_CLIENT_FILE_NAME = "META-INF/jonas-client.xml";
 
    /** 
     *  The error handler to use
     */
    private static ErrorHandler errorHandler = null;

    /** 
     * The entity resolver to use
     */
    private static EntityResolver entityResolver = null;

    /**
     * Flag for parser validation
     */
    private static boolean parsingWithValidation = false;

    /**
     * Xml content of the web.xml file
     */
    private static String xmlContent = "";

    /**
     * Xml content of the jonas-web.xml file
     */

    private static String jonasXmlContent = "";
    /**
     * Digester used to parse application-client.xml
     */
    private static JDigester appClientDigester = null;

    /**
     * Digester use to parse jonas-client.xml
     */
    private static JDigester jonasAppClientDigester = null;

    /**
     * Rules to parse the application-client.xml
     */
    private static ApplicationClientRuleSet appClientRuleSet = new ApplicationClientRuleSet();

    /**
     * Rules to parse the jonas-client.xml
     */
    private static JonasClientRuleSet jonasAppClientRuleSet = new JonasClientRuleSet();

    /**
     * The name of the callback handler
     */
    private String callbackHandler = null;

    /**
     * The name of the jaas config file
     */
    private String jaasFile = null;

    /**
     * The name of the jaas entry
     */
    private String jaasEntry = null;

    /**
     * The username to use for the callback handler
     */
    private String username = null;

    /**
     * The password to use for the callback handler
     */
    private String password = null;

    /**
     * Construct an instance of a ClientContainerDeploymentDesc.<BR>
     * Constructor is private, call one of the static getInstance 
     * method instead.
     * @param classLoader the classloader for the classes.
     * @param applicationClient the data structure of the application-client (application-client.xml)
     * @param jonasClient the data structure of the jonas-client (jonas-client.xml)
     * @throws DeploymentDescException if the deployment 
     * descriptors are corrupted.
     */
    private ClientContainerDeploymentDesc(ClassLoader classLoader,
                                          ApplicationClient applicationClient,
                                          JonasClient jonasClient) 
        throws DeploymentDescException {
        super(classLoader, applicationClient, jonasClient, null);
        

        // callbackHandler
        callbackHandler = (applicationClient.getCallbackHandler() != null) ? applicationClient.getCallbackHandler() : null;
        
        JonasSecurity jonasSecurity = jonasClient.getJonasSecurity();
        if (jonasSecurity != null) {
            // jaas config file
            jaasFile = (jonasSecurity.getJaasfile() != null) ? jonasSecurity.getJaasfile() : null;

            // jaas entry
            jaasEntry = (jonasSecurity.getJaasentry() != null) ? jonasSecurity.getJaasentry() : null;

            // username
            username = (jonasSecurity.getUsername() != null) ? jonasSecurity.getUsername() : null;

            // password
            password = (jonasSecurity.getPassword() != null) ? jonasSecurity.getPassword() : null;
        }


    }

    /**
     * Get an instance of a Client deployment descriptor by parsing the application-client.xml 
     * and jonas-client.xml deployment descriptors.
     * @param clientFileName the fileName of the client file for the deployment 
     * descriptors.
     * @param classLoaderForCls the classloader for the classes.
     * @return a Client deployment descriptor by parsing the application-client.xml 
     * and jonas-client.xml deployment descriptors.
     * @throws ClientContainerDeploymentDescException if the deployment 
     * descriptors are corrupted.
     */
    public static ClientContainerDeploymentDesc getInstance(String clientFileName,
                                                            ClassLoader classLoaderForCls)
        throws DeploymentDescException {
        
        return getInstance(clientFileName, classLoaderForCls, null);
    }

    /**
     * Get an instance of a Client deployment descriptor by parsing the application-client.xml 
     * and jonas-client.xml deployment descriptors.
     * @param clientFileName the fileName of the client file for the deployment 
     * descriptors.
     * @param classLoaderForCls the classloader for the classes.
     * @param altClientXmlFilename the fileName to the application-client.xml for the alt-dd
     * tag in the Ear Case. This is used for specify an alternate DDesc file.
     * @return a Client deployment descriptor by parsing the application-client.xml 
     * and jonas-client.xml deployment descriptors.
     * @throws ClientContainerDeploymentDescException if the deployment 
     * descriptors are corrupted.
     */
    public static ClientContainerDeploymentDesc getInstance(String clientFileName,
                                                            ClassLoader classLoaderForCls,
                                                            String altClientXmlFilename)

        throws DeploymentDescException {

       // Reinit static values;
        xmlContent = "";
        jonasXmlContent = "";

        // clientjar file
        JarFile clientFile = null;

        // Input streams
        InputStream applicationClientInputStream = null;
        InputStream jonasClientInputStream = null;
        
        // ZipEntry
        ZipEntry applicationClientZipEntry = null;
        ZipEntry jonasClientZipEntry = null;

        // Clients
        ApplicationClient applicationClient;
        JonasClient jonasClient;

        // Build the file
        File fClient = new File(clientFileName);

        
        //Check if the file exists.
        if (!(fClient.exists())) {
            String err = "' " + clientFileName + "' was not found.";
            throw new ClientContainerDeploymentDescException(err);
        }

        //Check if the Alt deploymentDesc file exists.
        //But only if it's a non null value because it's optionnal.
        if ((altClientXmlFilename != null) && (!new File(altClientXmlFilename).exists())) {
            String err = "The file for the altdd tag for the EAR case '" + altClientXmlFilename + "' was not found.";
            throw new ClientContainerDeploymentDescException(err);
        }


        // load the application-client deployment descriptor data (META-INF/application-client.xml 
        // and META-INF/jonas-client.xml)
        //No alt-dd case
        if (altClientXmlFilename == null) {
            try {
                clientFile = new JarFile(clientFileName);
                
                //Lookup in the JAR
                //Check the client entry
                applicationClientZipEntry = clientFile.getEntry(CLIENT_FILE_NAME);
                if (applicationClientZipEntry == null) {
                    throw new ClientContainerDeploymentDescException("The entry '" + CLIENT_FILE_NAME + "' was not found in the file '" + clientFileName + "'.");
                }
                //Get the stream
                applicationClientInputStream = clientFile.getInputStream(applicationClientZipEntry);
            } catch (Exception e) {
                if (clientFile != null) {
                    try {
                        clientFile.close();
                    } catch (IOException ioe) {
                        //We can't close the file
                    }
                }
                throw new ClientContainerDeploymentDescException("Can not read the XML deployment descriptors of the client jar file '" + clientFileName + "'.", e);
            }
        } else {
            try {
                applicationClientInputStream = new FileInputStream(altClientXmlFilename);
            } catch (FileNotFoundException ioe) {
                throw new ClientContainerDeploymentDescException("The altDD file '" + altClientXmlFilename + "' was not found.");
            }
        }

        applicationClient = loadApplicationClient(new InputStreamReader(applicationClientInputStream), CLIENT_FILE_NAME);
        
        try {
            clientFile = new JarFile(clientFileName);
                
            // Lookup in the JAR
            // Check the client entry
            jonasClientZipEntry = clientFile.getEntry(JONAS_CLIENT_FILE_NAME);
            if (jonasClientZipEntry != null) {
                //Get the stream
                jonasClientInputStream = clientFile.getInputStream(jonasClientZipEntry);
            }
        } catch (Exception e) {
            if (clientFile != null) {
                try {
                    clientFile.close();
                } catch (IOException ioe) {
                    //We can't close the file
                }
            }
            throw new ClientContainerDeploymentDescException("Can not read the XML deployment descriptors of the client jar file '" + clientFileName + "'.", e);
        }

        // load jonas-client deployment descriptor data 
        // (META-INF/jonas-client.xml)
        if (jonasClientInputStream != null) {
            jonasClient = loadJonasClient(new InputStreamReader(jonasClientInputStream), 
                                          JONAS_CLIENT_FILE_NAME);
            try {
                jonasClientInputStream.close(); 
            } catch (IOException e) {
                // Nothing to do
            }
        } else {
            jonasClient = new JonasClient();
        }


        // instantiate client deployment descriptor
        return new ClientContainerDeploymentDesc(classLoaderForCls, 
                                                 applicationClient, 
                                                 jonasClient);
    }


    /** 
     * Load the application-client.xml file.
     * @param reader the reader of the XML file.
     * @param fileName the name of the file (application-client.xml).
     * @return a structure containing the result of the application-client.xml parsing.
     * @throws ClientContainerDeploymentDescException if the deployment descriptor 
     * is corrupted.
     */
    public static ApplicationClient loadApplicationClient(Reader reader, String fileName) 
        throws DeploymentDescException {

        ApplicationClient appc = new ApplicationClient();

        // Create if null
        if (appClientDigester == null) {
            // Create and initialize the digester
            appClientDigester = new JDigester(appClientRuleSet,
                                              parsingWithValidation,
                                              true,
                                              new AppClientDTDs(),
                                              new AppClientSchemas());
        }

        try {
            appClientDigester.parse(reader ,fileName, appc );
        } catch (DeploymentDescException  e) {
            throw e;
        } finally {
            appClientDigester.push(null);
        }
        return appc;
    }

    /**
     * Load the jonas-client.xml file.
     * @param is the InputStream of the XML file.reader the Reader of the XML file.
     * @param fileName the name of the file (jonas-client.xml).
     * @return a structure containing the result of the jonas-client.xml parsing.
     * @throws ClientContainerDeploymentDescException if the deployment 
     * descriptor is corrupted.
     */
    public static JonasClient loadJonasClient(Reader reader, 
                                              String fileName) 
        throws DeploymentDescException {

        JonasClient jc = new JonasClient();

        // Create if null
        if (jonasAppClientDigester == null) {
            jonasAppClientDigester = new JDigester(jonasAppClientRuleSet,
                                                   parsingWithValidation,
                                                   true,
                                                   new JonasAppClientDTDs(),
                                                   new JonasAppClientSchemas()); 
        }

        try {
            jonasAppClientDigester.parse(reader, fileName, jc);

        } catch (DeploymentDescException e) {
            throw e;
        } finally {
            jonasAppClientDigester.push(null);
        }
        return jc;
    }



    /**
     * Get the name of the jaas configuration file
     * @return the name of the jaas configuration file
     */
    public String getJaasFile() {
        return jaasFile;
    }

    /**
     * Get the entry in the jaas configuration file
     * @return the entry in the jaas configuration file
     */
    public String getJaasEntry() {
        return jaasEntry;
    }

    /**
     * Get the username used for a callback handler
     * @return the username used for a callback handler
     */
    public String getUsername() {
        return username;
    }

    /**
     * Get the password used for a callback handler
     * @return the password used for a callback handler
     */
    public String getPassword() {
        return password;
    }

    /**
     * Get the callback handler of this client application.
     * @return the callback handler of this client application.
     */
    public String getCallbackHandler() {
        return callbackHandler;
    }

    /**
     * Return a String representation of the ClientContainerDeploymentDesc.
     * @return a String representation of the ClientContainerDeploymentDesc.
     */
    public String toString() {
        StringBuffer ret = new StringBuffer();
        
        // Return the displayName
        ret.append("\ngetDisplayName()=" + getDisplayName());
        
        // Return the resource-env-ref
        ResourceEnvRefDesc[] rer = getResourceEnvRefDesc();
        for (int i = 0; i < rer.length; i++) {
            ret.append("\ngetResourceEnvRefDesc(" + i + ")=" + rer[i].getClass().getName());
            ret.append(rer[i].toString());
        }

        // Return the resource-ref
        ResourceRefDesc[] resourceRefDesc = getResourceRefDesc();
        for (int i = 0; i < resourceRefDesc.length; i++) {
            ret.append("\ngetResourceRefDesc(" + i + ")=" + resourceRefDesc[i].getClass().getName());
            ret.append(resourceRefDesc[i].toString());
        }
 
        // Return the env-entry
        EnvEntryDesc[] envEntries = getEnvEntryDesc();
        for (int i = 0; i < envEntries.length; i++) {
            ret.append("\ngetEnvEntryDesc(" + i + ")=" + envEntries[i].getClass().getName());
            ret.append(envEntries[i].toString());
        }

        // Return the ejb-ref
        EjbRefDesc[] ejbRefDesc = getEjbRefDesc();
        for (int i = 0; i < ejbRefDesc.length; i++) {
            ret.append("\ngetEjbRefDesc(" + i + ")=" + ejbRefDesc[i].getClass().getName());
            ret.append(ejbRefDesc[i].toString());
        }
        
        // Return the callbackHandler
        ret.append("\ngetCallbackHandler()=" + getCallbackHandler());

        return ret.toString();
    }

    /**
     * Get the current ErrorHandler.
     * @return the current ErrorHandler.
     */
    public static ErrorHandler getErrorHandler() {
        return errorHandler;
    }
    
    /**
     * Get the current EntityResolver.
     * @return the current EntityResolver.
     */
    public static EntityResolver getEntityResolver() {
        return entityResolver;
    }

    /**
     * Controls whether the parser is reporting all validity errors.
     * @return if true, all external entities will be read.
     */
    public static boolean getParsingWithValidation() {
        return ClientContainerDeploymentDesc.parsingWithValidation;
    }

    /**
     * Set the specified ErrorHandler. If the application does not register
     * an ErrorHandler, all error events reported by the reader are ignored.
     * @param errorHandler the Error Handler to set.
     */
    public static void setErrorHandler(ErrorHandler errorHandler) {
        ClientContainerDeploymentDesc.errorHandler = errorHandler;
    }

    /**
     * Set the specified EntityResolver.
     * @param entityResolver the Entity Resolver to set.
     */
    public static void setEntityResolver(EntityResolver entityResolver) {
        ClientContainerDeploymentDesc.entityResolver = entityResolver;
    }

    /**
     * Controls whether the parser is reporting all validity errors.
     * @param validation if true, all external entities will be read.
     */
    public static void setParsingWithValidation(boolean validation) {
        ClientContainerDeploymentDesc.parsingWithValidation = validation;
    }

}
