/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Florent BENOIT & Ludovic BERT
 * --------------------------------------------------------------------------
 * $Id: EarDeploymentDesc.java,v 1.8 2004/03/19 14:31:51 sauthieg Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ear.deployment.api;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Vector;
import java.util.jar.JarFile;
import java.util.zip.ZipEntry;

import org.objectweb.jonas_ear.deployment.rules.ApplicationRuleSet;
import org.objectweb.jonas_ear.deployment.xml.Application;
import org.objectweb.jonas_ear.deployment.xml.Module;
import org.objectweb.jonas_ear.deployment.xml.SecurityRole;
import org.objectweb.jonas_ear.deployment.xml.Web;
import org.objectweb.jonas_lib.deployment.api.AbsDeploymentDesc;
import org.objectweb.jonas_lib.deployment.api.DeploymentDescException;
import org.objectweb.jonas_lib.deployment.api.JDigester;
import org.xml.sax.EntityResolver;
import org.xml.sax.ErrorHandler;

/**
 * This class extends the AbsDeploymentDescriptor class of JOnAS
 * It provides a description of the specific EAR desployment descriptor
 * @author Florent Benoit
 * @author Ludovic Bert
 * @author Helene Joanin
 */

public class EarDeploymentDesc extends AbsDeploymentDesc {

    /**
     * Path of the application.xml deploymnet descriptor file
     */
    private static final String APPLICATION_FILE_NAME = "META-INF/application.xml";

    /**
     * Digester used to parse application.xml
     */
    private static JDigester earDigester = null;

    /**
     * Rules to parse the application.xml
     */
    private static ApplicationRuleSet appRuleSet = new ApplicationRuleSet();

    /** 
     *  The error handler to use
     */
    private static ErrorHandler errorHandler = null;

    /** 
     * The entity resolver to use
     */
    private static EntityResolver entityResolver = null;

    /**
     * Flag for parser validation
     */
    private static boolean parsingWithValidation = true;

    /**
     * Vector of connectorTags (Connector objects).
     */
    private Vector connectorTags = null;

    /**
     * Vector of altDD for connectors
     */
    private Vector altDDConnectors = null;

    /**
     * Vector of ejbTags (Ejb objects).
     */
    private Vector ejbTags = null;

    /**
     * Vector of altDD for EJBs
     */
    private Vector altDDEjbs = null;

    /**
     * Vector of webTags (Web objects).
     */
    private Vector webTags = null;

    /**
     * Vector of clientTags (Java objects).
     */
    private Vector clientTags = null;

    /**
     * Vector of altDD for wars
     */
    private Vector altDDWebs = null;

    /**
     * Vector of altDD for clients
     */
    private Vector altDDClients = null;

    /**
     * Vector of security roles
     */
    private Vector securityRolesNames = null;

    /**
     * Xml content of application.xml file
     */
    private static String xmlContent = null;
    

    /**
     * Construct an instance of a EarDeploymentDesc.
     * Called by the getInstance() method.
     * @param classLoaderForCls ClassLoader of the classes .
     * @param application application.xml parsed file.
     * @throws EarDeploymentDescException if we can't build an instance.
     */
    private EarDeploymentDesc(ClassLoader classLoaderForCls, Application application)
        throws EarDeploymentDescException {

        // test classloader
        if (classLoaderForCls == null) {
            throw new EarDeploymentDescException("DeploymentDesc: Classloader is null");
        }

        ejbTags = new Vector();
        connectorTags = new Vector();
        webTags = new Vector();
        clientTags = new Vector();
        altDDEjbs = new Vector();
        altDDClients = new Vector();
        altDDConnectors = new Vector();
        altDDWebs = new Vector();
        securityRolesNames = new Vector();

        // display name
        displayName = application.getDisplayName();

        //Tags
        for (Iterator i = application.getModuleList().iterator(); i.hasNext();) {
            Module module = (Module) i.next();
            String ejb = module.getEjb();
            String connector = module.getConnector();
            String java = module.getJava();
            Web web = module.getWeb();
            String altDD = module.getAltDd();
            if (ejb != null) {
                ejbTags.add(ejb);
                altDDEjbs.add(altDD);
            } else if (connector != null) {
                connectorTags.add(connector);
                altDDConnectors.add(altDD);
            } else if (java != null) {
                clientTags.add(java);
                altDDClients.add(altDD);
            } else if (web != null) {
                webTags.add(web);
                altDDWebs.add(altDD);
            }
        }
        for (Iterator i = application.getSecurityRoleList().iterator(); i.hasNext();) {
            SecurityRole securityRole = (SecurityRole) i.next();
            if (securityRole != null) {
                if (securityRole.getRoleName() != null) {
                    securityRolesNames.add(securityRole.getRoleName());
                }
            }
        }

    }
    
    /**
     * Get the ejb tags of the application.xml file.
     * @return the ejb tags of the application.xml file.
     */
    public String[] getEjbTags() {
        String[] tmp = new String[ejbTags.size()];
        ejbTags.copyInto(tmp);
        return tmp;
    }

    /**
     * Get the alt-dd of the ejbs of the application.xml file.
     * @return the alt-dd of the ejbs of the application.xml file.
     */
    public String[] getAltDDEjbs() {
        String[] tmp = new String[altDDEjbs.size()];
        altDDEjbs.copyInto(tmp);
        return tmp;
    }

    /**
     * Get the client tags of the application.xml file.
     * @return the client tags of the application.xml file.
     */
    public String[] getClientTags() {
        String[] tmp = new String[clientTags.size()];
        clientTags.copyInto(tmp);
        return tmp;
    }

    /**
     * Get the alt-dd of the clients of the application.xml file.
     * @return the alt-dd of the clients of the application.xml file.
     */
    public String[] getAltDDClients() {
        String[] tmp = new String[altDDClients.size()];
        altDDClients.copyInto(tmp);
        return tmp;
    }

    /**
     * Get the web tags of the application.xml file.
     * @return the web tags of the application.xml file.
     */
    public Web[] getWebTags() {
        Web[] tmp = new Web[webTags.size()];
        webTags.copyInto(tmp);
        return tmp;
    }

    /**
     * Get the alt-dd of the wars of the application.xml file.
     * @return the alt-dd of the wars of the application.xml file.
     */
    public String[] getAltDDWebs() {
        String[] tmp = new String[altDDWebs.size()];
        altDDWebs.copyInto(tmp);
        return tmp;
    }

    /**
     * Get the connector tags of the application.xml file.
     * @return the connector tags of the application.xml file.
     */
    public String[] getConnectorTags() {
        String[] tmp = new String[connectorTags.size()];
        connectorTags.copyInto(tmp);
        return tmp;
    }

    /**
     * Get the alt-dd of the connectors of the application.xml file.
     * @return the alt-dd of the connectors of the application.xml file.
     */
    public String[] getAltDDConnectors() {
        String[] tmp = new String[altDDConnectors.size()];
        altDDConnectors.copyInto(tmp);
        return tmp;
    }

    /**
     * Get the security-role names tags
     * @return the security roles names.
     */
    public String[] getSecurityRolesNames() {
        String[] tmp = new String[securityRolesNames.size()];
        securityRolesNames.copyInto(tmp);
        return tmp;
    }

    /**
     * Get the content of the xml file
     * @return the content of the application xml file
     */
    public String getXmlContent() {
        return xmlContent;
    }

    /**
     * Get an instance of an EAR deployment descriptor by parsing the 
     * application.xml deployment descriptor.
     * @param earFileName the fileName of the war file for the deployment descriptors.
     * @param classLoaderForCls the classloader for the classes.
     * @return an EAR deployment descriptor by parsing the application.xml 
     * deployment descriptor.
     * @throws EarDeploymentDescException if the deployment descriptor is 
     * corrupted.
     */
    public static EarDeploymentDesc getInstance(String earFileName,
                                                ClassLoader classLoaderForCls)
        throws EarDeploymentDescException {

        //ear file
        JarFile earFile = null;

        //Input Stream
        InputStream appInputStream = null;

        //ZipEntry
        ZipEntry appZipEntry = null;

        Application application;
 

        //Check if the file exists.
        File earF = new File(earFileName);
        if (!earF.exists()) {
            throw new EarDeploymentDescException("The file '" + earFileName + "' was not found.");
        }

        // load deployment descriptor data (META-INF/application.xml)
        try {
            // Is it a directory ?
            if (earF.isDirectory()) {
                //lookup a META-INF/application.xml file
                File earXmlF = new File(earFileName, APPLICATION_FILE_NAME);
                if (!earXmlF.exists()) {
                    String err = "You have choose to deploy an ear directory but there is no " + APPLICATION_FILE_NAME + " file in the directory " + earFileName;
                    throw new EarDeploymentDescException(err);
                }
                appInputStream = new FileInputStream(earXmlF);
                xmlContent = xmlContent(appInputStream);
                // necessary to have a not empty InputStream !!!
                appInputStream = new FileInputStream(earXmlF);
            } else {
                // it is an .ear file
                earFile = new JarFile(earFileName);

                //Check the application entry
                appZipEntry = earFile.getEntry(APPLICATION_FILE_NAME);
                if (appZipEntry == null) {
                    throw new EarDeploymentDescException("The entry '" + APPLICATION_FILE_NAME + "' was not found in the file '" + earFileName + "'.");
                }

                //Get the stream
                appInputStream = earFile.getInputStream(appZipEntry);
                xmlContent = xmlContent(appInputStream);
                // necessary to have a not empty InputStream !!!
                appInputStream = earFile.getInputStream(appZipEntry);
            }
        } catch (Exception e) {
            if (earFile != null) {
                try {
                    earFile.close();
                } catch (IOException ioe) {
                    //We can't close the file
                }
            }
            throw new EarDeploymentDescException("Cannot read the XML deployment descriptors of the ear file '" + earFileName + "'.", e);
        }

        application = loadApplication(new InputStreamReader(appInputStream), APPLICATION_FILE_NAME);
        try {
            appInputStream.close();
        } catch (IOException e) {
        }

        // instantiate deployment descriptor
        return new EarDeploymentDesc(classLoaderForCls, application);
    }


    /**
     * Load the application.xml file.
     * @param is the InputStream of the XML file.
     * @param fileName the name of the file (application.xml).
     * @throws EarDeploymentDescException if the deployment descriptor is 
     * corrupted.
     * @return an application object.
     */
    public static Application loadApplication(Reader reader, String fileName)
        throws EarDeploymentDescException {

        Application app = new Application();

        // Create if earDigester is null
        if (earDigester == null) {
            try {
                // Create and initialize the digester
                earDigester = new JDigester(appRuleSet,
                                            parsingWithValidation,
                                            true,
                                            new EarDTDs(),
                                            new EarSchemas());
            } catch (DeploymentDescException  e) {
                throw new EarDeploymentDescException(e);
            }
        }

        try {
            earDigester.parse(reader, fileName, app);
        } catch (DeploymentDescException  e) {
            throw new EarDeploymentDescException(e);
        } finally {
            earDigester.push(null);
        }

        return app;
    }

    /**
     * Return the content of the inut stream
     * @param in the given input stream
     * @return the content of the inut stream
     */
    private static String xmlContent(InputStream in) throws IOException {

        StringBuffer sb = new StringBuffer();
        String line;
        BufferedReader br = new BufferedReader(new InputStreamReader(in));
        // loop until the end
        while ((line = br.readLine()) != null) {
            sb.append(line);
            sb.append("\n");
        }
        return sb.toString();
    }


    /** 
     * Return a String representation of the EarDeploymentDesc.
     * @return a String representation of the EarDeploymentDesc.
     */
    public String toString() {
        
        StringBuffer ret = new StringBuffer();
        ret.append("\ndisplay-name=" + displayName);
        ret.append("\nconnectors=");
        for (Enumeration e = connectorTags.elements(); e.hasMoreElements();) {
            ret.append(e.nextElement() + ",");
        }
        ret.append("\nejbs=");
        for (Enumeration e = ejbTags.elements(); e.hasMoreElements();) {
            ret.append(e.nextElement() + ",");
        }
        ret.append("\nwebs=");
        for (Enumeration e = webTags.elements(); e.hasMoreElements();) {
            ret.append(((Web)e.nextElement()).getWebUri() + ",");
        }
        ret.append("\njavas=");
        for (Enumeration e = clientTags.elements(); e.hasMoreElements();) {
            ret.append(e.nextElement() + ",");
        }
        ret.append("\nsecurity-roles-names=");
        for (Enumeration e = securityRolesNames.elements(); e.hasMoreElements();) {
            ret.append(e.nextElement() + ",");
        }

        return ret.toString();
    }

    /**
     * Get the current ErrorHandler.
     * @return the current ErrorHandler.
     */
    public static ErrorHandler getErrorHandler() {
        return errorHandler;
    }

    /**
     * Get the current EntityResolver.
     * @return the current EntityResolver.
     */
    public static EntityResolver getEntityResolver() {
        return entityResolver;
    }

    /**
     * Controls whether the parser is reporting all validity errors.
     * @return if true, all external entities will be read.
     */
    public static boolean getParsingWithValidation() {
        return EarDeploymentDesc.parsingWithValidation;
    }

    /**
     * Set the specified ErrorHandler. If the application does not register
     * an ErrorHandler, all error events reported by the reader are ignored.
     * @param errorHandler the Error Handler to set.
     */
    public static void setErrorHandler(ErrorHandler errorHandler) {
        EarDeploymentDesc.errorHandler = errorHandler;
    }

    /**
     * Set the specified EntityResolver.
     * @param entityResolver the Entity Resolver to set.
     */
    public static void setEntityResolver(EntityResolver entityResolver) {
        EarDeploymentDesc.entityResolver = entityResolver;
    }

    /**
     * Controls whether the parser is reporting all validity errors.
     * @param validation if true, all external entities will be read.
     */
    public static void setParsingWithValidation(boolean validation) {
        EarDeploymentDesc.parsingWithValidation = validation;
    }


}
