/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: JEntitySwitch.java,v 1.34 2004/03/19 14:31:48 sauthieg Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.rmi.NoSuchObjectException;
import java.rmi.RemoteException;

import javax.ejb.EJBException;
import javax.ejb.EntityBean;
import javax.ejb.ObjectNotFoundException;
import javax.ejb.TimedObject;
import javax.ejb.Timer;
import javax.ejb.TimerService;
import javax.transaction.Transaction;

import org.objectweb.jonas_ejb.deployment.api.EntityDesc;
import org.objectweb.jonas_ejb.deployment.api.MethodDesc;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * JEntitySwitch is used internally to synchronize accesses to the entity context
 * and thus to the entity bean instance.
 * All parts common to EJBObject and EJBLocalObject should be here.
 * Different policies can be applied to manage context/instance pairs:
 * - only 1 pair (container manages the transaction isolation)
 * - 1 pair for each transaction (transaction isolation managed by DataBase)
 * - 2 pairs (1 for transactional accesses, 1 for non transaction accesses)
 * @author Philippe Durieux
 * @author Philippe Coq
 */
public abstract class JEntitySwitch {

    /**
     * The Factory for this bean
     */
    protected JEntityFactory bf;

    /**
     * The Primary Key for this bean instance.
     */
    protected Object pk = null;

    /**
     * The EJBLocalObject, or null if bean has no local interface.
     */
    protected JEntityLocal local = null;

    /**
     * The EJBObject, or null if bean has no remote interface.
     */
    protected JEntityRemote remote = null;

    /**
     * shared=true if the bean can be modify outside this container.
     */
    protected boolean shared;

    /**
     * Lock policy used for this entity bean.
     * Possible values are
     * 0 = LOCK_CONTAINER_READ_UNCOMMITTED (1 instance)
     * 1 = LOCK_CONTAINER_SERIALIZED (1 instance)
     * 2 = LOCK_CONTAINER_READ_COMMITTED (2 instances)
     * 3 = LOCK_DATABASE (n instances)
     */
    protected int lockpolicy;

    /**
     * True if a transaction is mandatory for all modifying methods. In this case,
     * all methods outside tranction is considered as read-only.
     */
    protected boolean txUpdates;

    /**
     * The Timer Service
     */
    protected TimerService myTimerService = null;

    /**
     * nb of non transacted requests running
     */
    protected int countIH = 0;

    /**
     * nb of threads waiting (synchronization)
     */
    protected int waiters = 0;

    /**
     * True if this instance may have been modified outside transactions.
     * Avoids to put it twice in dirty list.
     */
    protected boolean inDirtyList = false;

    /**
     * True if instance has been modified by a Transaction.
     * This means that reading outside transaction should reload it before use.
     */
    protected boolean mustReload = false;

    /**
     * True when context/instance has been discarded, to avoids that
     * passivate store its state on storage
     * Meaningful only outside tx
     */
    protected boolean discarded = false;

    // transaction on which a Context has been registered.
    // used only for policies CS and CRC
    protected Transaction runningtx = null;

    // transaction that has modified an instance
    // used only for policies DB and CRU
    protected Transaction writingtx = null;

    /**
     * last transaction blocked by synchronization.
     * this is to do a very basic deadlock detection.
     */
    protected Transaction blockedtx = null;

    // ident is used mainly for debugging
    protected static int counter = 0;
    protected String ident;

    protected boolean lazyregister = false;

    abstract JEntityContext getContext4Tx(Transaction tx);
    abstract void setContext4Tx(Transaction tx, JEntityContext ctx);
    abstract void removeContext4Tx(Transaction tx);
    abstract public JEntityContext mapICtx(Transaction tx, JEntityContext bctx, boolean forced, boolean holdit);
    abstract protected void initpolicy(JEntityFactory bf);
    abstract public boolean passivateIH(boolean passivation);
    abstract public void endIH();
    abstract public void notifyWriting(Transaction tx, JEntityContext bctx);
    abstract public void releaseICtx(Transaction tx);

    /**
     * empty constructor. Object is initialized via init() because it is
     * implemented differently according to jorm mappers.
     */
    public JEntitySwitch() {
    }

    /**
     * constructor.
     * A new object is build when a new PK is known in the container,
     * either when a new bean is created, or when a find occurs.
     * For create(), PK is not known yet when this object is build.
     * @param bf The Entity Factory
     * @param pk The Primary Key
     */
    public void init(JEntityFactory bf, Object pk) {
        this.bf = bf;
        this.pk = pk;
        if (pk == null) {
            TraceEjb.logger.log(BasicLevel.ERROR, ident + "Init Entity Switch with a null PK!");
            throw new EJBException("Init Entity Switch with a null PK!");
        }

        shared = bf.isShared();
        initpolicy(bf);

        // generate new ident (debug)
        ident = "<" + counter + ":" + pk + ">";
        if (TraceEjb.isDebugTxlistener()) TraceEjb.txlistener.log(BasicLevel.DEBUG, ident + lazyregister);
        counter++;

        countIH = 0;
        waiters = 0;

        // Create EJBObject if bean has a Remote Interface
        if (bf.getHome() != null) {
            try {
                remote = ((JEntityHome) bf.getHome()).createRemoteObject();
                remote.setEntitySwitch(this);
            } catch (RemoteException e) {
                throw new EJBException("cannot create Remote Object", e);
            }
        }

        // Create EJBLocalObject if bean has a Local Interface
        if (bf.getLocalHome() != null) {
            local = ((JEntityLocalHome) bf.getLocalHome()).createLocalObject();
            local.setEntitySwitch(this);
        }
    }

    /**
     * @return the underlaying EJBLocalObject
     */
    public JEntityLocal getLocal() {
        return local;
    }

    /**
     * @return the underlaying EJBObject
     */
    public JEntityRemote getRemote() {
        return remote;
    }

    /**
     * Obtains the TimerService associated for this Entity Bean (one / pk)
     * @return a JTimerService instance.
     */
    public TimerService getEntityTimerService() {
        if (myTimerService == null) {
            // TODO : Check that instance implements TimedObject ?
            myTimerService = new JTimerService(this);
        }
        return myTimerService;
    }

    /**
     * Notify a timeout for this bean and this Pk
     * @param timer timer whose expiration caused this notification.
     */
    public void notifyTimeout(Timer timer) {
        if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident);

        // Making this method generic leads to some restrictions listed here after :
        // - Transactional attribute is enforced to TX_REQUIRES_NEW.
        // - security is not checked.
        // LATER: Make it non generic and move it in JEntityLocal.vm and JEntityRemote.vm

        RequestCtx rctx = bf.preInvoke(MethodDesc.TX_REQUIRES_NEW, "");
        try {
            JEntityContext bctx = getICtx(rctx.currTx);
            EntityBean eb = bctx.getInstance();
            if (eb instanceof TimedObject) {
                TimedObject instance = (TimedObject) eb;
                instance.ejbTimeout(timer);
            } else {
                throw new EJBException("The bean does not implement the `TimedObject` interface");
            }
        } catch (EJBException e) {
            rctx.sysExc = e;
            throw e;
        } catch (RuntimeException e) {
            rctx.sysExc = e;
            throw new EJBException("RuntimeException thrown by an enterprise Bean", e);
        } catch (Error e) {
            rctx.sysExc = e;
            throw new EJBException("Error thrown by an enterprise Bean" + e);
        } catch (RemoteException e) {
            rctx.sysExc = e;
            throw new EJBException("Remote Exception raised:", e);
        } finally {
            try {
                bf.postInvoke(rctx);
            } finally {
                if (rctx.sysExc != null) {
                    discardICtx(rctx.currTx);
                } else {
                    releaseICtx(rctx.currTx);
                }
            }
        }
	}

    /**
     * @return the Primary Key Object for this instance.
     */
    public Object getPrimaryKey() {
        if (pk == null) {
            throw new EJBException("No Primary Key yet");
        }
        return pk;
    }

    /**
     * bind a JEntityContext for a create method.
     * No need to synchronize since object was unknown before. (TODO ?)
     * @param tx - the Transaction object
     * @param bctx - the JEntityContext to bind
     */
    public void bindICtx(Transaction tx, JEntityContext bctx) {
        mapICtx(tx, bctx, true, false);
    }

    /**
     * Try to bind a JEntityContext if none already bound.
     * Called by finder methods
     * @param tx - the Transaction object
     * @param bctx The Entity Context
     */
    public synchronized void tryBindICtx(Transaction tx, JEntityContext bctx) throws ObjectNotFoundException {

        // Don't bind if already a context.
        if (getContext4Tx(tx) != null) {
            TraceEjb.context.log(BasicLevel.DEBUG, ident + "context already mapped!");
            bf.releaseJContext(bctx);
            if (getContext4Tx(tx).isMarkedRemoved()) {
                TraceEjb.context.log(BasicLevel.DEBUG, ident + "Removed in the current tx");
                throw new ObjectNotFoundException("Removed in the current tx");
            }
            return;
        }

        // No synchronization, since this is used for readonly methods (finder);
        bctx.initEntityContext(this);
        bctx.activate(true);
        setContext4Tx(tx, bctx); // after activate (CRU => full parallellism)
        if (! lazyregister) {
            if (tx == null) {
                if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "IH find");
            } else {
                if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "IT find");
                registerCtx(tx, bctx);
            }
        }
    }

    /**
     * bind a JEntityContext for a remove method.
     * called in case of remove(pk) or remove(handle)
     * @param tx - the Transaction object
     * @param newctx - the JEntityContext to bind
     * @return the BeanContext
     */
    public JEntityContext getICtx(Transaction tx, JEntityContext newctx) {
        return mapICtx(tx, newctx, false, false);
    }

    /**
     * Get a context/instance associated with this transaction
     * Called at each request on the bean (including remove)
     * @param tx - the Transaction object
     * @return the BeanContext
     */
    public JEntityContext getICtx(Transaction tx) {
        return mapICtx(tx, null, false, true);
    }

    public Transaction getBlockedTx() {
        return blockedtx;
    }

    public Transaction getBlockingTx() {
        return runningtx;
    }

    /**
     * Discard a context/instance at end of request. A problem occured
     * on this insance and it must be discarded.
     * @param tx - transaction associated to this context
     */
    public synchronized void discardICtx(Transaction tx) {
        if (tx == null) {
            JEntityContext jec = getContext4Tx(tx);
            countIH--;
            if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "discardICtx IH count=" + countIH);
            if (countIH == 0) {
                if (TraceEjb.isDebugContext()) TraceEjb.context.log(BasicLevel.DEBUG, ident + "discardICtx!");
                bf.releaseJContext(jec);
                removeContext4Tx(tx);
            } else {
                discarded = true;
            }
        } else {
            // nothing to do now. Wait for tx completed.
            if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "discardICtx IT");
        }
    }

    /**
     * This transaction is now over. We can dispose of the instance
     * for another transaction or discard it.
     * @param tx the transaction object
     * @param committed true if transaction was committed.
     */
    public synchronized void txCompleted(Transaction tx, boolean committed) {
        JEntityContext jec = getContext4Tx(tx);
        runningtx = null;
        if (writingtx != null && tx.equals(writingtx)) {
            writingtx = null;
        }

        if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "txCompleted " + committed);

        // The create was eventually rolled back. We must remove PK entry.
        if (jec.isNewInstance() && !committed) {
            if (TraceEjb.isDebugContext()) TraceEjb.context.log(BasicLevel.DEBUG, ident + "create rolled back!");
            bf.releaseJContext(jec);
            removeContext4Tx(tx);
            noLongerUsed();
            return;
        }
            
        if (jec.isMarkedRemoved()) {
            if (TraceEjb.isDebugContext()) TraceEjb.context.log(BasicLevel.DEBUG, ident + "remove!");
            bf.releaseJContext(jec);
            removeContext4Tx(tx);
            if (committed) {
                noLongerUsed(); // remove is effective !
            }
        } else {
            if (shared || !committed) {
                if (TraceEjb.isDebugContext()) TraceEjb.context.log(BasicLevel.DEBUG, jec + " passivated!");
                jec.passivate();
                bf.releaseJContext(jec);
                removeContext4Tx(tx);
            } else {
                // do this before reallocating the Context for another tx (inside this lock)
                jec.detachTx();
            }
            // If special instance for no transactional accesses, it must be reloaded
            if (lockpolicy == EntityDesc.LOCK_CONTAINER_READ_COMMITTED || 
                lockpolicy == EntityDesc.LOCK_DATABASE) {
                mustReload = true;
            }
        }
        if (waiters > 0) {
            TraceEjb.synchro.log(BasicLevel.DEBUG, ident + " notify");
            notifyAll();
        }
    }

    /**
     * Remove this EJB and
     * UnExport the remote object.
     */
    protected void noLongerUsed() {
        if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident);
        if (remote != null) {
            try {
                remote.unexportObject();
            } catch (NoSuchObjectException e) {
                TraceEjb.logger.log(BasicLevel.ERROR, ident + " unexport entity failed: " + e);
            }
        }
        bf.removeEJB(getPrimaryKey());
    }

    /**
     * register a Context on the transaction, as a Synchronization.
     * this will be used later to store instance state when needed :
     * before a finder, or at beforeCompletion, and to release instance
     * at commit.
     * @param tx Transaction object
     * @param bctx The Context to be registered
     */
    protected void registerCtx(Transaction tx, JEntityContext bctx) {
        // register this context to the transaction, as a Synchronization.
        try {
            TraceEjb.context.log(BasicLevel.DEBUG, bctx + "registered!");
            bf.registerContext(tx, bctx);
        } catch (IllegalStateException e) {
            TraceEjb.logger.log(BasicLevel.ERROR, ident + "Transaction is in an illegal state");
            throw e;
        }
        runningtx = tx;
        bctx.setRunningTx(tx);
    }

    /**
     * @return lock policy for this bean
     */
    public int getPolicy() {
        return lockpolicy;
    }


    /**
     * @return ejb instance's state
     * state values are 0=in-tx, 1=out-tx, 2=idle, 3=passive, 4=removed.
     * we don't synchronize this method to avoid jadmin blocks
     */
    public int getState() {
        JEntityContext jec = getContext4Tx(null); // TODO: change this null !
        if (jec == null) {
            return 3;
        }
        if (runningtx != null) {
            return 0;
        }
        if (jec.isMarkedRemoved()) {
            return 4;
        }
        if (inDirtyList) {
            return 1;
        } else {
            return 2;
        }
    }

}
