/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: JEntitySwitchDB.java,v 1.3 2004/03/19 14:31:48 sauthieg Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.util.HashMap;

import javax.transaction.SystemException;
import javax.transaction.Transaction;

import org.objectweb.jonas_ejb.deployment.api.EntityDesc;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * JEntitySwitch is used internally to synchronize accesses to the entity context
 * and thus to the entity bean instance.
 * All parts common to EJBObject and EJBLocalObject should be here.
 * Different policies can be applied to manage context/instance pairs:
 * - only 1 pair (container manages the transaction isolation)
 * - 1 pair for each transaction (transaction isolation managed by DataBase)
 * - 2 pairs (1 for transactional accesses, 1 for non transaction accesses)
 * @author Philippe Durieux
 * @author Philippe Coq
 */
public class JEntitySwitchDB extends JEntitySwitch {

    /**
     * EntityContext for non-transacted requests
     */
    protected JEntityContext ihContext = null;

    /**
     * Map of EntityContext's for transactions
     * Key is the Transaction.
     */
    protected HashMap itsContext;

    /**
     * empty constructor. Object is initialized via init() because it is
     * implemented differently according to jorm mappers.
     */
    public JEntitySwitchDB() {
        lockpolicy = EntityDesc.LOCK_DATABASE;
        txUpdates = true;
        itsContext = new HashMap();
    }

    protected void initpolicy(JEntityFactory bf) {
        lazyregister = ! bf.isPrefetch();
    }

    /**
     * @param tx The Transaction
     * @return the JEntityContext used for this tx
     */
    protected JEntityContext getContext4Tx(Transaction tx) {
        if (tx == null ) {
            return ihContext;
        } else {
            return (JEntityContext) itsContext.get(tx);
        }
    }

    /**
     * @param tx The Transaction
     * @param the JEntityContext used for this tx
     */
    protected void setContext4Tx(Transaction tx, JEntityContext ctx) {
        if (tx == null) {
            ihContext = ctx;
        } else {
            itsContext.put(tx, ctx);
        }
    }

    /**
     *
     */
    protected void removeContext4Tx(Transaction tx) {
        if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + " get rid of this Context");
        // reset these flags
        discarded = false;
        // free this Context
        if (tx == null) {
            ihContext = null;
        } else {
            itsContext.remove(tx);
        }
    }

    /**
     * Map a context and its instance.
     * @param tx - the Transaction object
     * @param bctx - the JEntityContext to bind if not null
     * @param forced - force to take this context. (case of create)
     * @param holdit - increment count to hold it, a release will be called later.
     * @return JEntityContext actually mapped
     */
    public synchronized JEntityContext mapICtx(Transaction tx, JEntityContext bctx, boolean forced, boolean holdit) {

        // Synchronization.
        if (tx == null) {
            // ------------- IH begin -------------
            if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "mapICtx IH");
        } else {
            // ------------- IT begin -------------
            if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "mapICtx IT");
        }

        // Choose the context to use.
        boolean newtrans = false;
        JEntityContext jec = getContext4Tx(tx);
        if (forced) {
            // If the new context is enforced, we must first release the older
            if (jec != null) {
                if (TraceEjb.isDebugContext()) TraceEjb.context.log(BasicLevel.DEBUG, ident + "new context is enforced!");
                bf.releaseJContext(jec);
                removeContext4Tx(tx);
            }
            jec = bctx;
            setContext4Tx(tx, jec);
            jec.initEntityContext(this);
            newtrans = true;
        } else {
            if (jec != null) {
                // Reuse the Context for this transaction.
                // If a context was supplied, release it first.
                if (bctx != null) {
                    if (TraceEjb.isDebugContext()) TraceEjb.context.log(BasicLevel.DEBUG, ident + "a context was supplied!");
                    bf.releaseJContext(bctx);
                }
                jec.reuseEntityContext();
            } else {
                if (bctx != null) {
                    jec = bctx;
                } else {
                    // no Context available : get one from the pool.
                    jec = (JEntityContext) bf.getJContext();
                }
                jec.initEntityContext(this);
                jec.activate(true);
                setContext4Tx(tx, jec); // after activate
                newtrans = true;
            }
        }

        if (tx != null) {
            // Register Context now, except if lazy register possible, or no new transaction for policies 1 and 2.
            if (newtrans && ! lazyregister) {
                if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "mapICtx IT: new tx, registerSynchronization");
                registerCtx(tx, jec);
            }
        } else {
            if (holdit) {
                countIH++;
                if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "mapICtx IH count=" + countIH);
                if ((shared || mustReload) && countIH == 1) {
                    // reload state that could have been modified by transactions.
                    jec.activate(false);
                    mustReload = false;
                }
            }
        }

        return jec;
    }

    /**
     * Release a context/instance at end of request.
     * @param tx - transaction associated to this context
     */
    public synchronized void releaseICtx(Transaction tx) {

        if (tx == null) {
            // ------------- IH end -------------
            JEntityContext jec = getContext4Tx(tx);
            countIH--;
            if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "releaseICtx IH count=" + countIH);
            if (countIH == 0) {
                if (discarded) {
                    if (TraceEjb.isDebugContext()) TraceEjb.context.log(BasicLevel.DEBUG, ident + "discarded!");
                    bf.releaseJContext(jec);
                    removeContext4Tx(tx);
                } else {
                    if (jec.isMarkedRemoved()) {
                        if (TraceEjb.isDebugContext()) TraceEjb.context.log(BasicLevel.DEBUG, ident + "Marked Removed!");
                        bf.releaseJContext(jec);
                        removeContext4Tx(tx);
                        noLongerUsed();
                    }
                }
                if (waiters > 0) {
                    TraceEjb.synchro.log(BasicLevel.DEBUG, ident + " notify");
                    notifyAll();
                }
            }
        } else {
            // ------------- IT end -------------
            // nothing to do now. Wait for tx completed.
            if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "releaseICtx IT");

        }
    }

    public boolean passivateIH(boolean passivation) { 
        return true;
    }

    public void endIH() {
        return;
    }

    /**
     * This transaction has just modified this instance. (CMP2 only)
     * @param tx transaction
     */
    public synchronized void notifyWriting(Transaction tx, JEntityContext bctx) {
        TraceEjb.swapper.log(BasicLevel.DEBUG, ident);
        if (writingtx != null) {
            TraceEjb.logger.log(BasicLevel.WARN, "Conflict: 2 transactions attempt to write the same instance");
            try {
                tx.setRollbackOnly();
            } catch (SystemException e) {
                TraceEjb.logger.log(BasicLevel.ERROR, "cannot setRollbackOnly");
            }
        } else {
            if (lazyregister) {
                // Instance must be written later.
                // In case of prefetch, this has already be done (= !lazyregister)
                registerCtx(tx, bctx);
            }
            writingtx = tx;
        }
    }

}
