/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Helene Joanin
 * Contributor(s): Lutris Technologies Inc.
 *
 * --------------------------------------------------------------------------
 * $Id: BeanDesc.java,v 1.36 2004/03/18 14:29:38 benoitf Exp $
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas_ejb.deployment.api;

import java.lang.reflect.Modifier;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Properties;

import javax.jms.MessageListener;

import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import org.objectweb.jonas_lib.deployment.api.DeploymentDescException;
import org.objectweb.jonas_lib.deployment.api.EjbLocalRefDesc;
import org.objectweb.jonas_lib.deployment.api.EjbRefDesc;
import org.objectweb.jonas_lib.deployment.api.EnvEntryDesc;
import org.objectweb.jonas_lib.deployment.api.JndiEnvRefsGroupDesc;
import org.objectweb.jonas_lib.deployment.api.ResourceEnvRefDesc;
import org.objectweb.jonas_lib.deployment.api.ResourceRefDesc;
import org.objectweb.jonas_lib.deployment.api.SecurityRoleRefDesc;
import org.objectweb.jonas_lib.deployment.xml.RunAs;
import org.objectweb.jonas_lib.deployment.xml.SecurityRole;
import org.objectweb.jonas_lib.deployment.xml.SecurityRoleRef;

import org.objectweb.jonas_ejb.container.TraceEjb;
import org.objectweb.jonas_ejb.deployment.xml.AssemblyDescriptor;
import org.objectweb.jonas_ejb.deployment.xml.CommonEjb;
import org.objectweb.jonas_ejb.deployment.xml.ContainerTransaction;
import org.objectweb.jonas_ejb.deployment.xml.JonasCommonEjb;
import org.objectweb.jonas_ejb.deployment.xml.Method;
import org.objectweb.jonas_ejb.deployment.xml.MethodPermission;
import org.objectweb.jonas_ejb.lib.BeanNaming;
import org.objectweb.jonas_ejb.lib.JavaType;

import org.objectweb.jonas_ws.deployment.api.ServiceRefDesc;

/**
 * Class to hold meta-information related to any enterprise-beans
 * @author Christophe Ney [cney@batisseurs.com]
 * @author Helene Joanin
 * @author Philippe Coq
 * Contributor(s):
 * Florent Benoit & Ludovic Bert : ejb_link
 * Jeff Mesnil
 */
public abstract class BeanDesc extends JndiEnvRefsGroupDesc {

    protected static final int BEAN_TRANSACTION_TYPE = 1;
    protected static final int CONTAINER_TRANSACTION_TYPE = 2;
    protected static final String[] TRANS = {
        null,
        "BEAN_TRANSACTION_TYPE",
        "CONTAINER_TRANSACTION_TYPE"
    };

    private static final String ejb10PropertiesPrefix = "ejb10-properties/";
    private static final String REMOTEDEFAULTSUFFIX = "Home";
    private static final String LOCALDEFAULTSUFFIX = "_L";

    protected String ejbName = null;
    protected String jndiName = null;
    protected String jndiLocalName = null;
    protected Class home = null;
    protected Class remote = null;
    protected Class localhome = null;
    protected Class local = null;
    protected Class ejbClass = null;
    protected Properties ejb10EnvProps = new Properties();

    protected Logger logger = null;
    
    // array of methods meta-info
    private ArrayList methodDesc = new ArrayList();
    // used by deprebcated method
    private HashMap methodDescMap = new HashMap();

    /**
     * List of SecurityRoleRefDesc for JACC
     */
    protected List securityRoleRefDescList = null;

    protected DeploymentDesc deploymentDesc = null;

    private String fullWrpRemoteName = null;
    private String wrpRemoteName = null;

    private String fullWrpHomeName = null;
    private String wrpHomeName = null;

    private String fullWrpLocalName = null;
    private String wrpLocalName = null;

    private String fullWrpLocalHomeName = null;
    private String wrpLocalHomeName = null;

    private String fullWrpHandleName = null;
    private String wrpHandleName = null;

    protected String fullDerivedBeanName = null;
    protected String derivedBeanName = null;

    /**
     * Run-as role
     */
    private String runAsRole = null;


    protected int poolMin = 0;
    protected int cacheMax = 0;

    /**
     * constructor to be used by parent node
     */
    public BeanDesc(ClassLoader classLoader, 
                    CommonEjb ejb,
                    JonasCommonEjb jonasEjb,
                    AssemblyDescriptor asd,
                    String fileName)
        throws DeploymentDescException {

        super(classLoader, ejb, jonasEjb, fileName);
        securityRoleRefDescList = new LinkedList();
        ejbName = (ejb.getEjbName() != null) ? ejb.getEjbName(): null;

        // If no value given specifically, we use ejbName+Home as JNDI name.
        jndiName = (jonasEjb.getJndiName() != null) ? jonasEjb.getJndiName() : ejbName + REMOTEDEFAULTSUFFIX;
        // If no value given specifically, we use jndiName+_L as JNDI local name.
        jndiLocalName = (jonasEjb.getJndiLocalName() != null) ? jonasEjb.getJndiLocalName() : jndiName + LOCALDEFAULTSUFFIX;

        // classes
        if (ejb.getHome() != null) {
            try {
                this.home = classLoader.loadClass(ejb.getHome());
                // check home extends javax.ejb.EJBHome
                if (!javax.ejb.EJBHome.class.isAssignableFrom(this.home)) {
                    throw new DeploymentDescException("Home class '" + ejb.getHome()
                                                      + "' doesn't not extends javax.ejb.EJBHome");
                }
            } catch (ClassNotFoundException e) {
                throw new DeploymentDescException("Home class not found for bean " + ejbName, e);
            }
        }

        if (ejb.getRemote() != null) {
            try {
                this.remote = classLoader.loadClass(ejb.getRemote());
                // check remote extends javax.ejb.EJBObject
                if (!javax.ejb.EJBObject.class.isAssignableFrom(this.remote)) {
                    throw new DeploymentDescException("Remote class '" + 
                                                      ejb.getRemote()+ 
                                                      "' doesn't not extends javax.ejb.EJBObject");
                }
            } catch (ClassNotFoundException e) {
                throw new DeploymentDescException("Remote class not found for bean " + ejbName, e);
            }
        }

        if (ejb.getLocalHome() != null) {
            try {
                this.localhome = classLoader.loadClass(ejb.getLocalHome());
                // check localhome extends javax.ejb.EJBLocalHome
                if (!javax.ejb.EJBLocalHome.class.isAssignableFrom(this.localhome)) {
                    throw new DeploymentDescException("LocalHome class '" + ejb.getLocalHome()+ "' doesn't not extends javax.ejb.EJBLocalHome");
                }
            } catch (ClassNotFoundException e) {
                throw new DeploymentDescException("LocalHome class not found for bean " + ejbName, e);
            }
        }

        if (ejb.getLocal() != null) {
            try {
                this.local = classLoader.loadClass(ejb.getLocal());
                // check local extends javax.ejb.EJBLocalObject
                if (!javax.ejb.EJBLocalObject.class.isAssignableFrom(this.local)) {
                    throw new DeploymentDescException("Local class '" + ejb.getLocal()
                                                      + "' doesn't not extends javax.ejb.EJBLocalObject");
                }
            } catch (ClassNotFoundException e) {
                throw new DeploymentDescException("Local class not found for bean " + ejbName, e);
            }
        }

        try {
            this.ejbClass = classLoader.loadClass(ejb.getEjbClass());
            // check ejb-class extends javax.ejb.EnterpriseBean
            if (!javax.ejb.EnterpriseBean.class.isAssignableFrom(this.ejbClass)) {
                throw new DeploymentDescException("Ejb class '" + ejb.getEjbClass()
                                                  + "' doesn't not extends javax.ejb.EnterpriseBean");
            }
        } catch (ClassNotFoundException e) {
            throw new DeploymentDescException("Ejb class not found for bean " + ejbName, e);
        }
        EnvEntryDesc [] envt = getEnvEntryDesc();
        for (int i = 0; i < envt.length; i++) {
            if (envt[i].getName().startsWith(ejb10PropertiesPrefix)) {
                String newName = envt[i].getName().substring(ejb10PropertiesPrefix.length());
                String value = new String();
                if (envt[i].hasValue()) {
                    value = envt[i].getValue().toString();
                }
                ejb10EnvProps.put(newName, value);
            }
        }
 
        // run-as
        if (ejb.getSecurityIdentity() != null) {
            RunAs runAsElem = ejb.getSecurityIdentity().getRunAs();
            if (runAsElem != null) {
                String roleName = runAsElem.getRoleName();
                if (roleName != null) {
                    runAsRole = roleName;
                    boolean runAsRoleFound = false;
                    for (Iterator j = asd.getSecurityRoleList().iterator(); j.hasNext();) {
                        SecurityRole securityRole = (SecurityRole) j.next();
                        // role name of a security-role element defined in the assembly descriptor
                        String r = securityRole.getRoleName();
                        if (r.equals(runAsRole)) {
                            runAsRoleFound = true;
                        }
                    }
                    //if there are no matching run-as and security-role
                    if (!runAsRoleFound) {
                        throw new DeploymentDescException("There is no security-role defined for the run-as role: " + runAsRole);
                    }
                }
            }
        }


        // populate array with all methods
        try {
            java.lang.reflect.Method[] m;
            int beginIdx = methodDesc.size();
            int len = 0;
            if ((this.home != null) && (this.remote != null)) {
                // session bean or entity bean with remote interface
                m = this.home.getMethods();
                for (int i = 0; i < m.length; i++) {
                    if (!m[i].getDeclaringClass().equals(javax.ejb.EJBHome.class)
                        || m[i].getName().equals("remove")) {
                        addMethodDesc(m[i]);
                        len++;
                        // check RemoteException is thrown
                        checkRemoteException(m[i], true);
                        /*
                        // arguments and return types must be valid types for RMI
                        if (!m[i].getDeclaringClass().equals(javax.ejb.EJBHome.class)
                            && !m[i].getName().equals("findByPrimaryKey")) {
                            checkValidTypesForRmi(m[i]);
                        }
                        */
                    }
                }

                m = this.remote.getMethods();
                for (int i = 0; i < m.length; i++) {
                    if (!m[i].getDeclaringClass().equals(javax.ejb.EJBObject.class)
                        || m[i].getName().equals("remove")) {
                        addMethodDesc(m[i]);
                        len++;
                        // check RemoteException is thrown
                        checkRemoteException(m[i], true);
                        /*
                        // arguments and return types must be valid for RMI
                        if (!m[i].getDeclaringClass().equals(javax.ejb.EJBObject.class)) {
                            checkValidTypesForRmi(m[i]);
                        }
                        */
                    }
                }
            }
            if ((this.localhome != null) && (this.local != null)) {
                // session bean or entity bean with local interface
                m = this.localhome.getMethods();
                for (int i = 0; i < m.length; i++) {
                    if (!m[i].getDeclaringClass().equals(javax.ejb.EJBLocalHome.class)
                        || m[i].getName().equals("remove")) {
                        addMethodDesc(m[i]);
                        len++;
                        // check RemoteException is NOT thrown
                        checkRemoteException(m[i], false);
                    }
                }
                m = this.local.getMethods();
                for (int i = 0; i < m.length; i++) {
                    if (!m[i].getDeclaringClass().equals(javax.ejb.EJBLocalObject.class)
                        || m[i].getName().equals("remove")) {
                        addMethodDesc(m[i]);
                        len++;
                        // check RemoteException is NOT thrown
                        checkRemoteException(m[i], false);
                    }
                }
            }
            if (this.home == null && this.localhome == null) {
                // message driven bean
                addMethodDesc(MessageListener.class.getMethods()[0]);
                len++;
            }
            if (javax.ejb.EntityBean.class.isAssignableFrom(this.ejbClass)) {
                // ejbSelect methods
                m = this.ejbClass.getMethods();
                for (int i = 0; i < m.length; i++) {
                    if (m[i].getName().startsWith("ejbSelect")
                        && Modifier.isAbstract(m[i].getModifiers())) {
                        addMethodDesc(m[i]);
                        len++;
                    }
                }
            }
            orderMethodDesc(methodDesc, beginIdx, len);
        } catch (SecurityException e) {
            throw new DeploymentDescException("Cannot use java reflection on " + ejbName);
        }

        // populate HashMap with all security role references
        // for each security-role-ref of the bean
        if (ejb.getSecurityRoleRefList() != null) {
            for (Iterator i = ejb.getSecurityRoleRefList().iterator(); i.hasNext();) {
                SecurityRoleRef sRoleRef = (SecurityRoleRef) i.next();
                // test if the assembly-descriptor element exists
                if (asd == null) {
                    throw new DeploymentDescException("No assembly-descriptor element (and security-role elements) whereas security-role-ref element(s) are defined");
                } else {
                    // test if the assembly-descriptor element contains security-role elements
                    if (asd.getSecurityRoleList().isEmpty()) {
                        throw new DeploymentDescException("No security-role elements defined in the assembly-descriptor element  whereas security-role-ref element(s) are defined");
                    }
                }
                // test if the role-link element of the security-role-ref element exists
                if (sRoleRef.getRoleLink() == null) {
                    throw new DeploymentDescException("No role-link element in a security-role-ref element with role-name element : " + sRoleRef.getRoleName());
                }
                String roleLink = sRoleRef.getRoleLink();
                // correspond to role-name element in security-role-ref element
                String roleRefName = sRoleRef.getRoleName();
                // to know if there are corresponding roleLink and roleRefName
                boolean correspondingRoles = false;
                for (Iterator j = asd.getSecurityRoleList().iterator(); j.hasNext();) {
                    SecurityRole securityRole = (SecurityRole) j.next();
                    // role name of a security-role element defined in the assembly descriptor
                    String roleName = securityRole.getRoleName();
                    if (roleName.equals(roleLink)) {
                        securityRoleRefDescList.add(new SecurityRoleRefDesc(ejbName, sRoleRef, true));
                        correspondingRoles = true;
                    }
                }
                //if there are no matching roleLink and roleName
                if (!correspondingRoles) {
                    throw new DeploymentDescException("No corresponding role-name in a security-role element for role-link : " + roleLink);
                }
            }
        }

        // check if there are corresponding role-name element in security-role
        // and method-permission element of the Assembly Descriptor
        if (asd != null) {
            verifyRoleName(asd);
            // set  method descriptor values for method permissions
            for (Iterator i = asd.getMethodPermissionList().iterator(); i.hasNext();) {
                MethodPermission methodPermission = (MethodPermission) i.next();
                for (Iterator j = methodPermission.getMethodList().iterator(); j.hasNext();) {
                    Method m = ((Method) j.next());
                    // check method applies to current bean
                    if (!m.getEjbName().equals(ejbName)) {
                        continue;
                    }
                    // selection of interface home/remote
                    Class pClass = null;
                    if (m.getMethodIntf() != null) {
                        if (m.getMethodIntf().equals("Home")) {
                            pClass = javax.ejb.EJBHome.class;
                        } else if (m.getMethodIntf().equals("Remote")) {
                            pClass = javax.ejb.EJBObject.class;
                        } else if (m.getMethodIntf().equals("LocalHome")) {
                            pClass = javax.ejb.EJBLocalHome.class;
                        } else if (m.getMethodIntf().equals("Local")) {
                            pClass = javax.ejb.EJBLocalObject.class;
                        } else {
                            throw new DeploymentDescException(m.getMethodIntf()
                                                              + " is invalid value for method-intf on bean "+ ejbName);
                        }
                    }
                    for (int k = 0; k < methodDesc.size(); k++) {
                        MethodDesc md = (MethodDesc) methodDesc.get(k);
                        int matchStatus = md.matchPattern(pClass, m.getMethodName(), m.getMethodParams());
                        for (Iterator l = methodPermission.getRoleNameList().iterator(); l.hasNext();) {
                            String roleName = (String) l.next();
                            // Add only if it greater than APPLY_TO_NOTHING
                            if (matchStatus > MethodDesc.APPLY_TO_NOTHING) {
                                md.addRoleName(roleName);
                            }
                        }
                    }
                }
            }
        }

        // set method descriptor values for container transactions
        if (asd != null) {
            // Loop on ContainerTransaction list
            Iterator i = asd.getContainerTransactionList().iterator();
            while (i.hasNext()) {
                ContainerTransaction cotr = (ContainerTransaction) i.next();
                // Loop on Method list
                Iterator j = cotr.getMethodList().iterator();
                while (j.hasNext()) {
                    Method m = (Method) j.next();

                    // check method applies to current bean
                    if (!m.getEjbName().equals(ejbName)) {
                        continue;
                    }

                    // selection of interface home/remote
                    Class pClass = null;
                    if (m.getMethodIntf() != null) {
                        String mintf = m.getMethodIntf();
                        if (mintf.equals("Home")) {
                            pClass = javax.ejb.EJBHome.class;
                        } else if (mintf.equals("Remote")) {
                            pClass = javax.ejb.EJBObject.class;
                        } else if (mintf.equals("LocalHome")) {
                            pClass = javax.ejb.EJBLocalHome.class;
                        } else if (mintf.equals("Local")) {
                            pClass = javax.ejb.EJBLocalObject.class;
                        } else {
                            throw new DeploymentDescException(mintf
                                                              + " is invalid value for method-intf on bean "
                                                              + ejbName);
                        }
                    }

                    // loop on methods in the array
                    boolean found = false;
                    for (int k = 0; k < methodDesc.size(); k++) {
                        MethodDesc md = (MethodDesc) methodDesc.get(k);
                        int matchStatus = md.matchPattern(pClass, m.getMethodName(), m.getMethodParams());
                        if (matchStatus != MethodDesc.APPLY_TO_NOTHING) {
                            found = true;
                        }
                        // set if proper level of assignation
                        String transAttribute = null;
                        if (cotr.getTransAttribute() != null) {
                            transAttribute = cotr.getTransAttribute();
                        }
                        md.overwriteTxAttribute(transAttribute, matchStatus);
                    }
                    if (!found) {
                        // The method defined via the <method> element is not a declared method
                        throw new DeploymentDescException("Invalid container-transaction: method '"
                                                          + MethodDesc.methodElementToString(m)
                                                          + "' not declared in bean " + ejbName);
                    }
                }
            }
        }

        String packageName = "";
        String ejbIdentifier = getIdentifierFromEjbName(ejbName);

        if (getRemoteClass() != null) {
            packageName = BeanNaming.getPackageName(getRemoteClass().getName());
            wrpRemoteName = new String("JOnAS" + ejbIdentifier + "Remote");
            wrpHandleName = new String("JOnAS" + ejbIdentifier + "Handle");
            fullWrpRemoteName = BeanNaming.getClassName(packageName, wrpRemoteName);
            fullWrpHandleName = BeanNaming.getClassName(packageName, wrpHandleName);
        }

        if (getLocalClass() != null) {
            packageName = BeanNaming.getPackageName(getLocalClass().getName());
            wrpLocalName = new String("JOnAS" + ejbIdentifier + "Local");
            fullWrpLocalName = BeanNaming.getClassName(packageName, wrpLocalName);
        }

        if (getHomeClass() != null) {
            packageName = BeanNaming.getPackageName(getHomeClass().getName());
            wrpHomeName = new String("JOnAS" + ejbIdentifier + "Home");
            fullWrpHomeName = BeanNaming.getClassName(packageName, wrpHomeName);
        }

        if (getLocalHomeClass() != null) {
            packageName = BeanNaming.getPackageName(getLocalHomeClass().getName());
            wrpLocalHomeName = new String("JOnAS" + ejbIdentifier + "LocalHome");
            fullWrpLocalHomeName = BeanNaming.getClassName(packageName, wrpLocalHomeName);
        }

        fullDerivedBeanName = getEjbClass().getName();
        derivedBeanName = BeanNaming.getBaseName(getEjbClass().getName());

    }

    /**
     * Build a valid java identifier from the ejb name
     */
    protected String getIdentifierFromEjbName(String ejbName) {
        char[] tabChar = new char[ejbName.length()];
        for (int i = 0; i < ejbName.length(); i++) {
            if (!Character.isJavaIdentifierPart(ejbName.charAt(i))) {
                tabChar[i] = '_';
            } else {
                tabChar[i] = ejbName.charAt(i);
            }
        }
        return new String(tabChar);
    }

    /**
     * Check if method throws RemoteException
     * @exception DeploymentDescException thrown for non-valid method
     */
    public void checkRemoteException(java.lang.reflect.Method m, boolean shouldThrow) throws DeploymentDescException {
        Class[] excepts = m.getExceptionTypes();
        for (int j = 0; j < excepts.length; j++) {
            if (RemoteException.class.isAssignableFrom(excepts[j])) {
                if (shouldThrow) {
                    return;
                } else {
                    throw new DeploymentDescException("Method " + m.getName() + " of "
                                                      + m.getDeclaringClass()
                                                      + " should NOT throw RemoteException");
                }
            }
        }
        if (shouldThrow) {
            throw new DeploymentDescException("Method " + m.getName() + " of "
                                              + m.getDeclaringClass() + " must throw RemoteException");
        } else {
            return;
        }

    }

    /**
     * Check if argument and return types are valid types for RMI-IIOP.
     * (Primitive types, remote objects, or non-remote objects that implement
     * the java.io.Serializable interface)
     * Only a warning is logging if it's not the case.
     */
    public void checkValidTypesForRmi(java.lang.reflect.Method m) {
        boolean isValid = true;
        Class []paramTypes = m.getParameterTypes();
        for (int i = 0; i < paramTypes.length; i++) {
            if (!JavaType.isValidForRmi(paramTypes[i])) {
                isValid = false;
            }
        }
        if (!JavaType.isValidForRmi(m.getReturnType())
            && !JavaType.isVoid(m.getReturnType())
            && !JavaType.isCollecOrEnum(m.getReturnType())) {
            // Allow Collection or Enumeration type (finder methods)
            isValid = false;
        }
        if (!isValid) {
            // TODO: the trace of this warning uses a static variable into TraceEjb classe
            // because this.logger is not yet initialized
            TraceEjb.dd.log(BasicLevel.WARN,
                            "Method " + m.getName() + " of " + m.getDeclaringClass()
                            + " may be have an invalid argument or return type for RMI"); 
        }
    }

    /**
     * Check that the bean descriptor is valid
     * @exception DeploymentDescException thrown for non-valid bean
     */
    public void check() throws DeploymentDescException {
        // check business methods transaction attribute
        for (int i = 0; i < methodDesc.size(); i++) {
            checkTxAttribute(((MethodDesc) methodDesc.get(i)));
        }
    }

    /**
     * check that trans-attribute is valid for bean
     */
    protected abstract void checkTxAttribute(MethodDesc md) throws DeploymentDescException;

    /**
     * check that there are always corresponding role-name element
     * in security-role and method-permission elements
     */
    protected void verifyRoleName(AssemblyDescriptor asd) throws DeploymentDescException {
        boolean verify;
        List methodPermission = asd.getMethodPermissionList();
        if (methodPermission != null) {
            for (Iterator i = methodPermission.iterator(); i.hasNext();) {
                MethodPermission mPermission = (MethodPermission) i.next();
                for (Iterator j = mPermission.getRoleNameList().iterator(); j.hasNext();) {
                    verify = false;
                    String mpRoleName = (String) j.next();
                    for (Iterator k = asd.getSecurityRoleList().iterator(); k.hasNext();) {
                        String srRoleName = ((SecurityRole) k.next()).getRoleName();
                        if (mpRoleName.equals(srRoleName)) {
                            verify = true;
                        }
                    }
                    if (!verify) {
                        throw new DeploymentDescException("a method-permission element defined a role-name element \"" + mpRoleName + "\" which doesn't correspond to any security-role element of the Deployment Descriptor");
                    }
                }
            }
        }
    }

    /**
     * Get the enterprise bean's name.
     * @return Name of the bean
     */
    public String getEjbName() {
        return ejbName;
    }

    /**
     * Get the name to associate with the enterprise Bean in the JNDI name space.
     * Invoking this method is disallowed for message driven bean.
     * (This information is JOnAS specific). Used by Server
     * @return Jndi name of the bean
     */
    public String getJndiName() {
        return jndiName;
    }

    /**
     * Get the name to associate with the enterprise Bean in the JNDI name space
     * for local access (inside the same JOnAS Server)
     * @return Jndi local name of the bean
     */
    public String getJndiLocalName() {
        return jndiLocalName;
    }

    /**
     * Get bean's home interface.
     * Invoking this method is disallowed for message driven bean.
     * Used by GenIC and Server
     * @return Class for the home interface or null
     */
    public Class getHomeClass() {
        return home;
    }

    /**
     * Get bean's remote interface.
     * Invoking this method is disallowed for message driven bean.
     * Used by GenIC and Server
     * @return Class for the remote interface or null
     */
    public Class getRemoteClass() {
        return remote;
    }

    /**
     * Get bean's local home interface.
     * Invoking this method is disallowed for message driven bean.
     * Used by GenIC and Server
     * @return Class for the home interface or null
     */
    public Class getLocalHomeClass() {
        return localhome;
    }

    /**
     * Get bean's local interface.
     * Invoking this method is disallowed for message driven bean.
     * Used by GenIC and Server
     * @return Class for the Local interface or null
     */
    public Class getLocalClass() {
        return local;
    }

    /**
     * Get bean's class.
     * Used by GenIC and Server
     * @return Class for the bean
     */
    public Class getEjbClass() {
        return ejbClass;
    }



    /**
     * Get bean's EJB 1.0 environment properties
     */
    public Properties getEjb10Environment() {
        Properties ret = new Properties(ejb10EnvProps);
        return ret;
    }


    /**
     * Get descriptor attached to a given method.
     * Used by GenIC
     * @param method Method from the Home/Remote/MessageListener interface
     * @return The method descriptor (never returns null)
     * @deprecated use getMethodDescIterator or getMethodDesc(int index) instead
     */
    public MethodDesc getMethodDesc(java.lang.reflect.Method method) {

        if (!methodDescMap.containsKey(method)) {
            throw new Error(method.getName()
                            + " is not a method of home/remote interfaces for bean "
                            + getEjbName());
        }
        return (MethodDesc) methodDescMap.get(method);
    }

    /**
     * get an iterator for all MethodDesc of the bean
     */
    public Iterator getMethodDescIterator() {
        return methodDesc.iterator();
    }

    /**
     * get a MethodDesc from its index
     */
    public MethodDesc getMethodDesc(int index) {
        return (MethodDesc) methodDesc.get(index);
    }


    /**
     * Gets the security-role-ref of the beans
     * can access to the EJBRoleRefPermission and role-link
     * @return the security-role-ref of the beans
     */
    public List getSecurityRoleRefDescList() {
        return securityRoleRefDescList;
    }


    /**
     * @return min size of the pool
     */
    public int getPoolMin() {
        return poolMin;
    }

    /**
     * @return max size of the cache
     */
    public int getCacheMax() {
        return cacheMax;
    }


    /**
     * @return the run-as role name
     */
    public String getRunAsRole() {
        return runAsRole;
    }



    /**
     * String representation of the object for test purpose
     * @return String representation of this object
     */
    public String toString() {
        StringBuffer ret = new StringBuffer();
        ret.append("\ngetDisplayName()=" + getDisplayName());
        ret.append("\ngetEjbName()=" + getEjbName());
        if (jndiName != null) {
            ret.append("\ngetJndiName()=" + getJndiName());
        }
        if (home != null) {
            ret.append("\ngetHomeClass()=" + getHomeClass().toString());
        }
        if (remote != null) {
            ret.append("\ngetRemoteClass()=" + getRemoteClass().toString());
        }
        if (localhome != null) {
            ret.append("\ngetLocalHomeClass()=" + getLocalHomeClass().toString());
        }
        if (local != null) {
            ret.append("\ngetLocalClass()=" + getLocalClass().toString());
        }
        ret.append("\ngetEjbClass()=" + getEjbClass().toString());
        EnvEntryDesc[] envE = getEnvEntryDesc();
        for (int i = 0; i < envE.length; i++) {
            ret.append("\ngetEnvEntryDesc(" + i + ")=" + envE[i].toString());
        }
        ServiceRefDesc[] sr = getServiceRefDesc();
        for (int i = 0; i < sr.length; i++) {
            ret.append("\ngetServiceRefDesc(" + i + ")=" + sr[i].toString());
        }
        ResourceEnvRefDesc[] resEnvR = getResourceEnvRefDesc();
        for (int i = 0; i < resEnvR.length; i++) {
            ret.append("\ngetResourceEnvRefDesc(" + i + ")=" + resEnvR[i].toString());
        }
        ResourceRefDesc[] resR = getResourceRefDesc();
        for (int i = 0; i < resR.length; i++) {
            ret.append("\ngetResourceRefDesc(" + i + ")=" + resR[i].toString());
        }
        EjbRefDesc[] ejbR = getEjbRefDesc();
        for (int i = 0; i < ejbR.length; i++) {
            ret.append("\ngetEjbRefDesc(" + i + ")=" + ejbR[i].toString());
        }
        EjbLocalRefDesc[] ejbLR = getEjbLocalRefDesc();
        for (int i = 0; i < ejbLR.length; i++) {
            ret.append("\ngetEjbLocalRefDesc(" + i + ")=" + ejbLR[i].toString());
        }
        for (int i = 0; i < methodDesc.size(); i++) {
            MethodDesc md = (MethodDesc) methodDesc.get(i);
            ret.append("\ngetMethodDesc(" + i + ")=" + md.getClass().getName());
            ret.append(md.toString());
        }

        if (securityRoleRefDescList != null) {
            for (Iterator i = securityRoleRefDescList.iterator(); i.hasNext();) {
                SecurityRoleRefDesc sRoleRefDesc = (SecurityRoleRefDesc) i.next();
                ret.append("\nsecurity-role-ref, role-name = '" + sRoleRefDesc.getRoleName() + "', role-link = '" + sRoleRefDesc.getRoleLink() + "'");
            }
        }
        ret.append("\ngetFullWrpRemoteName() = " + getFullWrpRemoteName());
        ret.append("\ngetFullWrpHomeName() = " + getFullWrpHomeName());
        ret.append("\ngetFullWrpLocalName() = " + getFullWrpLocalName());
        ret.append("\ngetFullWrpLocalHomeName() = " + getFullWrpLocalHomeName());
        ret.append("\ngetFullWrpHandleName() = " + getFullWrpHandleName());
        ret.append("\ngetFullDerivedBeanName() = " + getFullDerivedBeanName());
        ret.append("\ngetRunAsRole() = " + getRunAsRole());
        return ret.toString();
    }

    /**
     * MethodDesc constructor factory method to ovewrite
     */
    protected MethodDesc newMethodDescInstance(java.lang.reflect.Method meth, int index) {
        return new MethodDesc(this, meth, index);
    }

    /**
     * add a new method descriptor to the bean
     */
    protected MethodDesc addMethodDesc(java.lang.reflect.Method meth) {
        MethodDesc md = newMethodDescInstance(meth, methodDesc.size());
        methodDesc.add(md);
        methodDescMap.put(meth, md);
        return md;
    }

    /**
     * returns the full name of the wrapper class for the bean
     */
    public String getFullWrpRemoteName() {
        return fullWrpRemoteName;
    }

    /**
     * returns the full name of the wrapper class for the bean
     */
    public String getFullWrpLocalName() {
        return fullWrpLocalName;
    }

    /**
     * returns the name of the wrapper class for the bean
     */
    public String getWrpRemoteName() {
        return wrpRemoteName;
    }

    /**
     * returns the name of the wrapper class for the bean
     */
    public String getWrpLocalName() {
        return wrpLocalName;
    }

    /**
     * returns the full name of the Handle class for the bean
     */
    public String getFullWrpHandleName() {
        return fullWrpHandleName;
    }

    /**
     * returns the name of the Handle class for the bean
     */
    public String getWrpHandleName() {
        return wrpHandleName;
    }

    /**
     * returns the full name of the wrapper home class for the bean
     */
    public String getFullWrpHomeName() {
        return fullWrpHomeName;
    }

    /**
     * returns the full name of the wrapper local home class for the bean
     */
    public String getFullWrpLocalHomeName() {
        return fullWrpLocalHomeName;
    }

    /**
     * returns the name of the wrapper home class for the bean
     */
    public String getWrpHomeName() {
        return wrpHomeName;
    }

    /**
     * returns the name of the wrapper local home class for the bean
     */
    public String getWrpLocalHomeName() {
        return wrpLocalHomeName;
    }

    /**
     * returns the full name of the derived class of the bean
     */
    public String getFullDerivedBeanName() {
        return fullDerivedBeanName;
    }

    /**
     * returns the name of the derived class of the bean
     */
    public String getDerivedBeanName() {
        return derivedBeanName;
    }

    /**
     * get the parent deployment descriptor
     */
    public DeploymentDesc getDeploymentDesc() {
        return deploymentDesc;
    }

    /**
     * set the parent deployment descriptor
     */
    public void setDeploymentDesc(DeploymentDesc deploymentDesc) {
        this.deploymentDesc = deploymentDesc;
        logger = deploymentDesc.getLogger();
    }

    private void orderMethodDesc(ArrayList al, int begin, int length) {
        MethodDesc[] mds = new MethodDesc[length];
        for (int i = 0; i < length; i++) {
            mds[i] = (MethodDesc) al.get(begin + i);
        }
        Arrays.sort(mds, 0, length, new MethodDescComparator());
        for (int i = 0; i < length; i++) {
            al.set(begin + i, mds[i]);
            mds[i].setIndex(begin + i);
        }
    }

    private class MethodDescComparator implements Comparator {
        public int compare(Object o1, Object o2) {
            MethodDesc md1 = (MethodDesc) o1;
            MethodDesc md2 = (MethodDesc) o2;
            int res = md1.getMethod().getDeclaringClass().getName().compareTo(md2.getMethod().getDeclaringClass().getName());
            if (res == 0) {
                res = md1.getMethod().getName().compareTo(md2.getMethod().getName());
                if (res == 0) {
                    // Same method name => compare the parameter types
                    Class[] p1 = md1.getMethod().getParameterTypes();
                    Class[] p2 = md2.getMethod().getParameterTypes();
                    res = p1.length - p2.length;
                    if (res == 0) {
                        //Same parameter size
                        for (int i = 0;
                             i < p1.length && (res = p1[i].getName().compareTo(p2[i].getName())) == 0;
                             i++);
                    }
                }
            }
            return res;
        }
    }
}

