/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Eric HARDESTY
 * --------------------------------------------------------------------------
 * $Id: RarDeploymentDesc.java,v 1.7 2003/12/01 08:09:28 coqp Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_rar.deployment.api;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.util.Iterator;
import java.util.Vector;
import java.util.jar.JarFile;
import java.util.zip.ZipEntry;

import javax.naming.Context;
import javax.naming.NamingException;

import org.objectweb.jonas.naming.CompNamingContext;

import org.objectweb.jonas_lib.deployment.api.AbsDeploymentDesc;
import org.objectweb.jonas_lib.deployment.api.DeploymentDescException;
import org.objectweb.jonas_lib.deployment.api.JDigester;

import org.objectweb.jonas_rar.deployment.rules.ConnectorRuleSet;
import org.objectweb.jonas_rar.deployment.rules.JonasConnectorRuleSet;
import org.objectweb.jonas_rar.deployment.xml.ConfigProperty;
import org.objectweb.jonas_rar.deployment.xml.Connector;
import org.objectweb.jonas_rar.deployment.xml.JonasConfigProperty;
import org.objectweb.jonas_rar.deployment.xml.JonasConnector;

import org.xml.sax.EntityResolver;
import org.xml.sax.ErrorHandler;

/**
 * This class extends the AbsDeploymentDescriptor class of JOnAS
 * It provides a description of the specific RAR desployment descriptor
 * @author Eric Hardesty
 */

public class RarDeploymentDesc extends AbsDeploymentDesc {

    /**
     * Path of the ra.xml deploymnet descriptor file
     */
    private static final String RA_FILE_NAME = "META-INF/ra.xml";

    /**
     * Path of the jonas-ra.xml deploymnet descriptor file
     */
    private static final String JONAS_RA_FILE_NAME = "META-INF/jonas-ra.xml";

    /**
     *  The error handler to use
     */
    private static ErrorHandler errorHandler = null;

    /**
     * The entity resolver to use
     */
    private static EntityResolver entityResolver = null;

    /**
     * Flag for parser validation
     */
    private static boolean parsingWithValidation = true;

    /**
     * Connector
     */
    private static Connector connector = null;

    /**
     * Jonas Connector
     */
    private static JonasConnector jonasConnector = null;

    /**
     * Digester use to parse ra.xml
     */
    private static JDigester connectorDigester = null;

    /**
     * Digester use to parse jonas-ra.xml
     */
    private static JDigester jonasConnectorDigester = null;

    /**
     * Rules to parse the ra.xml
     */
    private static ConnectorRuleSet connectorRuleSet = new ConnectorRuleSet();

    /**
     * Rules to parse the jonas-ra.xml
     */
    private static JonasConnectorRuleSet jonasConnectorRuleSet =
        new JonasConnectorRuleSet();

    /**
     * Vector of raConfigPropTags (config-property objects).
     */
    private Vector raConfigPropTags = null;

    /**
     * Vector of jonasConfigPropTags (config-property objects).
     */
    private Vector jonasConfigPropTags = null;

    /**
     * Construct an instance of a RarDeploymentDesc.
     * Called by the getInstance() method.
     * @param classLoader ClassLoader of the classes .
     * @param connector ra.xml parsed file.
     * @param jonasConnector jonas-ra.xml parsed file.
     * @throws RarDeploymentDescException if we can't build an instance.
     */
    private RarDeploymentDesc(
        ClassLoader classLoader,
        Connector connector,
        JonasConnector jonasConnector)
        throws RarDeploymentDescException {

        // test classloader
        if (classLoader == null) {
            throw new RarDeploymentDescException("DeploymentDesc: Classloader is null");
        }

        if (connector != null) {
            // display name
            displayName =
                (connector.getDisplayName() != null)
                    ? connector.getDisplayName()
                    : null;

            raConfigPropTags = new Vector();
            //Tags
            for (Iterator i =
                connector
                    .getResourceadapter()
                    .getConfigPropertyList()
                    .iterator();
                i.hasNext();
                ) {
                raConfigPropTags.add((ConfigProperty) i.next());
            }
        }

        if (jonasConnector != null) {
            jonasConfigPropTags = new Vector();
            //Tags
            for (Iterator i =
                jonasConnector.getJonasConfigPropertyList().iterator();
                i.hasNext();
                ) {
                jonasConfigPropTags.add((JonasConfigProperty) i.next());
            }
        }

    }

    /**
     * Set the config-property tags of the ra.xml file.
     * @param conn the Connector objeect with the config-property tags of the ra.xml file.
     */
    public void setRaConfigPropTags(Connector conn) {
        //Tags
        for (Iterator i =
            conn.getResourceadapter().getConfigPropertyList().iterator();
            i.hasNext();
            ) {
            if (raConfigPropTags == null) {
                raConfigPropTags = new Vector();
            }
            raConfigPropTags.add((ConfigProperty) i.next());
        }
        if (raConfigPropTags == null) {
            return;
        }
        ConfigProperty[] tmp = new ConfigProperty[raConfigPropTags.size()];
        raConfigPropTags.copyInto(tmp);
    }

    /**
     * Get the config-property tags of the ra.xml file.
     * @return the config-property tags of the ra.xml file.
     */
    public ConfigProperty[] getRaConfigPropTags() {
        if (raConfigPropTags == null) {
            return null;
        }
        ConfigProperty[] tmp = new ConfigProperty[raConfigPropTags.size()];
        raConfigPropTags.copyInto(tmp);
        return tmp;
    }

    /**
     * Get the config-property tags of the jonas-ra.xml file.
     * @return the config-property tags of the jonas-ra.xml file.
     */
    public JonasConfigProperty[] getJonasConfigPropTags() {
        if (jonasConfigPropTags == null) {
            return null;
        }
        if (jonasConfigPropTags != null) {
            JonasConfigProperty[] tmp =
                new JonasConfigProperty[jonasConfigPropTags.size()];
            jonasConfigPropTags.copyInto(tmp);
            return tmp;
        }
        return null;
    }

    /**
     * Get an instance of an RAR deployment descriptor by parsing the
     * ra.xml and jonas-ra.xml deployment descriptors.
     * @param rarFileName the fileName of the rar file for the deployment descriptors.
     * @param classLoader the classloader for the classes.
     * @return an RAR deployment descriptor by parsing the ra.xml & jonas-ra.xml
     * deployment descriptors.
     * @throws DeploymentDescException if the deployment descriptors are
     * corrupted.
     */
    public static RarDeploymentDesc getInstance(
        String rarFileName,
        ClassLoader classLoader)
        throws DeploymentDescException {

        // The context to give for the creation of an instance
        // of RarDeploymentDesc
        Context contctx = null;
        try {
            contctx = new CompNamingContext(rarFileName);
            contctx.rebind("rarFileName", rarFileName);
            contctx.rebind("classloader", classLoader);
        } catch (Exception e) {
            // An exception occurred trying to bind context
            //            logger.log(BasicLevel.ERROR, "Error when binding context for RarDeploymentDesc for '" + rarFileName + "'");
            //            logger.log(BasicLevel.ERROR, e.getMessage());

            throw new RarDeploymentDescException(
                "Error during the deployment",
                e);
        }
        // want both ra.xml & jonas-ra.xml
        return getInstance(contctx);
    }

    /**
     * Get an instance of an RAR deployment descriptor by parsing the
     * ra.xml and jonas-ra.xml deployment descriptors.
     * @param ctx the context which contains the configuration in order to load the deployment descriptors.
     *   There are 6 possible parameters :
     *     - rarFileName is the path of the RAR file (required param).
     *     - classLoader is the classloader (required param)
     *     - altDD is the optional deployment descriptor (optional param).
     * @return an RAR deployment descriptor by parsing the ra.xml(or altdd) & jonas-ra.xml
     * deployment descriptors.
     * @throws DeploymentDescException if the deployment descriptors are
     * corrupted.
     */
    public static RarDeploymentDesc getInstance(Context ctx)
        throws DeploymentDescException {

        String rarFileName = null;
        ClassLoader classLoader = null;
        try {
            rarFileName = (String) ctx.lookup("rarFileName");
            classLoader = (ClassLoader) ctx.lookup("classloader");
        } catch (NamingException e) {
            String err = "Error while getting parameter from context param ";
            throw new RarDeploymentDescException(err, e);
        }

        // optional parameter
        String raDeployDesc = RA_FILE_NAME;
        try {
            raDeployDesc = (String) ctx.lookup("altDD");
        } catch (NamingException e) {
            ; // no alt DD
        }

        //ear file
        JarFile rarFile = null;

        //Input Stream
        InputStream raInputStream = null;
        InputStream jonasRaInputStream = null;

        //ZipEntry
        ZipEntry raZipEntry = null;
        ZipEntry jonasRaZipEntry = null;

        connector = null;
        jonasConnector = null;

        // Check if the file exists.
        if (!new File(rarFileName).exists()) {
            throw new RarDeploymentDescException(
                "The file '" + rarFileName + "' was not found.");
        }

        // load deployment descriptor data (META-INF/ra.xml)
        boolean setupRa = true;
        boolean setupJonasRa = true;
        try {
            rarFile = new JarFile(rarFileName);
            ;

            //Check the ra entry
            raZipEntry = rarFile.getEntry(raDeployDesc);
            if (raZipEntry == null) {
                connector = null;
                setupRa = false;
            } else {
                //Get the stream
                raInputStream = rarFile.getInputStream(raZipEntry);
            }

            //Check the jonas-ra entry
            jonasRaZipEntry = rarFile.getEntry(JONAS_RA_FILE_NAME);
            if (jonasRaZipEntry == null) {
                jonasConnector = null;
                setupJonasRa = false;
            } else {
                //Get the stream
                jonasRaInputStream = rarFile.getInputStream(jonasRaZipEntry);
            }

        } catch (Exception e) {
            if (rarFile != null) {
                try {
                    rarFile.close();
                } catch (IOException ioe) {
                    ; // Can't close the file
                }
            }
            throw new RarDeploymentDescException(
                "Cannot read the XML deployment descriptors of the rar file '"
                    + rarFileName
                    + "'.",
                e);
        }

        if (setupRa) {
            connector =
                loadConnector(
                    new InputStreamReader(raInputStream),
                    raDeployDesc);
            try {
                raInputStream.close();
            } catch (IOException e) {
                ; // can't close
            }
        }

        if (setupJonasRa) {
            jonasConnector =
                loadJonasConnector(
                    new InputStreamReader(jonasRaInputStream),
                    JONAS_RA_FILE_NAME);
            try {
                jonasRaInputStream.close();
            } catch (IOException e) {
                ; // can't close
            }
        }

        // instantiate deployment descriptor
        return new RarDeploymentDesc(classLoader, connector, jonasConnector);
    }

    /**
     * Load the ra.xml file.
     * @param reader the Reader of the XML file.
     * @param fileName the name of the file (ra.xml).
     * @throws DeploymentDescException if the deployment descriptor is
     * corrupted.
     * @return a connector object.
     */
    public static Connector loadConnector(Reader reader, String fileName)
        throws DeploymentDescException {

        Connector connector = new Connector();

        // Create if null
        if (connectorDigester == null) {
            connectorDigester =
                new JDigester(
                    connectorRuleSet,
                    parsingWithValidation,
                    true,
                    new ConnectorDTDs(),
                    new ConnectorSchemas());
        }

        try {
            connectorDigester.parse(reader, fileName, connector);
        } catch (DeploymentDescException e) {
            throw e;
        } finally {
            connectorDigester.push(null);
        }
        return connector;
    }

    /**
     * Load the jonas-ra.xml file.
     * @param reader the Reader of the XML file.
     * @param fileName the name of the file (jonas-ra.xml).
     * @throws DeploymentDescException if the deployment descriptor is
     * corrupted.
     * @return a JonasConnector object.
     */
    public static JonasConnector loadJonasConnector(
        Reader reader,
        String fileName)
        throws DeploymentDescException {

        JonasConnector jonasConnector = new JonasConnector();

        // Create if null
        if (jonasConnectorDigester == null) {
            jonasConnectorDigester =
                new JDigester(
                    jonasConnectorRuleSet,
                    parsingWithValidation,
                    true,
                    new JonasConnectorDTDs(),
                    new JonasConnectorSchemas());
        }

        try {
            jonasConnectorDigester.parse(reader, fileName, jonasConnector);
        } catch (DeploymentDescException e) {
            throw e;
        } finally {
            jonasConnectorDigester.push(null);
        }
        return jonasConnector;
    }

    /**
     * Return a String representation of the RarDeploymentDesc.
     * @return a String representation of the RarDeploymentDesc.
     */
    public String toString() {
        String ret = "Connector :";

        return ret;
    }

    /**
     * Get the current Connector.
     * @return the current Connector.
     */
    public static Connector getConnector() {
        return connector;
    }

    /**
     * Get the current ErrorHandler.
     * @return the current ErrorHandler.
     */
    public static ErrorHandler getErrorHandler() {
        return errorHandler;
    }

    /**
     * Get the current EntityResolver.
     * @return the current EntityResolver.
     */
    public static EntityResolver getEntityResolver() {
        return entityResolver;
    }

    /**
     * Get the current JonasConnector.
     * @return the current JonasConnector.
     */
    public static JonasConnector getJonasConnector() {
        return jonasConnector;
    }

    /**
     * Controls whether the parser is reporting all validity errors.
     * @return if true, all external entities will be read.
     */
    public static boolean getParsingWithValidation() {
        return RarDeploymentDesc.parsingWithValidation;
    }

    /**
     * Set the specified ErrorHandler. If the application does not register
     * an ErrorHandler, all error events reported by the reader are ignored.
     * @param errorHandler the Error Handler to set.
     */
    public static void setErrorHandler(ErrorHandler errorHandler) {
        RarDeploymentDesc.errorHandler = errorHandler;
    }

    /**
     * Set the specified EntityResolver.
     * @param entityResolver the Entity Resolver to set.
     */
    public static void setEntityResolver(EntityResolver entityResolver) {
        RarDeploymentDesc.entityResolver = entityResolver;
    }

    /**
     * Controls whether the parser is reporting all validity errors.
     * @param validation if true, all external entities will be read.
     */
    public static void setParsingWithValidation(boolean validation) {
        RarDeploymentDesc.parsingWithValidation = validation;
    }

}
