/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Florent BENOIT & Ludovic BERT
 * --------------------------------------------------------------------------
 * $Id: WebContainerDeploymentDesc.java,v 1.20 2004/03/19 14:31:54 sauthieg Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_web.deployment.api;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Vector;
import java.util.jar.JarFile;
import java.util.zip.ZipEntry;

import org.objectweb.jonas_lib.deployment.api.DeploymentDescException;
import org.objectweb.jonas_lib.deployment.api.EjbLocalRefDesc;
import org.objectweb.jonas_lib.deployment.api.EjbRefDesc;
import org.objectweb.jonas_lib.deployment.api.EnvEntryDesc;
import org.objectweb.jonas_lib.deployment.api.JDigester;
import org.objectweb.jonas_lib.deployment.api.JndiEnvRefsGroupDesc;
import org.objectweb.jonas_lib.deployment.api.ResourceEnvRefDesc;
import org.objectweb.jonas_lib.deployment.api.ResourceRefDesc;
import org.objectweb.jonas_lib.deployment.xml.SecurityRole;
import org.objectweb.jonas_web.deployment.rules.JonasWebAppRuleSet;
import org.objectweb.jonas_web.deployment.rules.WebAppRuleSet;
import org.objectweb.jonas_web.deployment.xml.JonasWebApp;
import org.objectweb.jonas_web.deployment.xml.Servlet;
import org.objectweb.jonas_web.deployment.xml.ServletMapping;
import org.objectweb.jonas_web.deployment.xml.WebApp;
import org.objectweb.jonas_ws.deployment.api.ServiceRefDesc;
import org.xml.sax.EntityResolver;
import org.xml.sax.ErrorHandler;


/**
 * This class do the parsing of the web.xml file and jonas-web.xml files and
 * contruct a data structure associated to these two files.
 * 03/03 : Can read web.xml and jonas-web.xml if the url of the war is a directory.
 * @author Ludovic Bert
 * @author Florent Benoit
 * @author Philippe Coq
 */
public class WebContainerDeploymentDesc extends JndiEnvRefsGroupDesc {

    /** 
     * The path to the web.xml file. 
     */
    public static final String WEB_FILE_NAME = "WEB-INF/web.xml";

    /** 
     * The path to the jonas-web.xml file. 
     */
    public static final String JONAS_WEB_FILE_NAME = "WEB-INF/jonas-web.xml";

    /** 
     *  The error handler to use
     */
    private static ErrorHandler errorHandler = null;

    /** 
     * The entity resolver to use
     */
    private static EntityResolver entityResolver = null;

    /**
     * Flag for parser validation
     */
    private static boolean parsingWithValidation = true;

    /**
     * Xml content of the web.xml file
     */
    private static String xmlContent = "";

    /**
     * Xml content of the jonas-web.xml file
     */
    private static String jonasXmlContent = "";

    /**
     * Digester use to parse web.xml
     */
    private static JDigester webAppDigester = null;

    /**
     * Digester use to parse jonas-web.xml
     */
    private static JDigester jonasWebAppDigester = null;

    /**
     * Rules to parse the web.xml
     */
    private static WebAppRuleSet webAppRuleSet = new WebAppRuleSet();

    /**
     * Rules to parse the jonas-web.xml
     */
    private static JonasWebAppRuleSet jonasWebAppRuleSet = new JonasWebAppRuleSet();

    /**
     * URLPattern Constraints 
     */
    private static SecurityConstraintListDesc securityConstraintListDesc = null;

    /**
     * List of URLs
     */
    private static List securityRoleList = null;

    /**
     * The host on which the web application must be depoyed (can be null in
     * ear case).
     */
    private String host = null;

    /**
     * The context root of the web application (can be null in ear case).
     */
    private String contextRoot = null;

    /**
     * The port number of the prefered connector to access the web application 
     * (can be null if defaults values can be found). Used only for WebServices.
     */
    private String port = null;
 
    /**
     * The delegation model used by the class loader context follows the java 2 delegation model ?
     */
    private boolean java2DelegationModel = true;

    /**
     * List of mapping between servlets names and servlets classes
     */
    private Map servlets = new HashMap();

    /**
     * List of mapping between servlets names and servlets url-mapping
     */
    private Hashtable servletsUrlMapping = new Hashtable();

    /**
     * Name of the war File. 
     */
    private String warFileName = "";

    /**
     * Construct an instance of a WebContainerDeploymentDesc.<BR>
     * Constructor is private, call one of the static getInstance 
     * method instead.
     * @param fileName the name of the warFile
     * @param classLoader the classloader for the classes.
     * @param webApp the data structure of the web-app (web.xml)
     * @param jonasWebApp the data structure of the jonas-web-app 
     * (jonas-web.xml)
     * @throws DeploymentDescException if the deployment 
     * descriptors are corrupted.
     */
    private WebContainerDeploymentDesc(String fileName,
                                       ClassLoader classLoader,
                                       WebApp webApp, JonasWebApp jonasWebApp) 
        throws DeploymentDescException {

        super(classLoader, webApp, jonasWebApp, fileName);


        // host
        host = (jonasWebApp.getHost() != null) ? jonasWebApp.getHost() : null;
        // context-root
        contextRoot = (jonasWebApp.getContextRoot() != null) ? jonasWebApp.getContextRoot() : null;
        // port
        port = (jonasWebApp.getPort() != null) ? jonasWebApp.getPort() : null;
  
        // Class loader delegation model
        String delegationModel = (jonasWebApp.getJava2DelegationModel() != null) ? jonasWebApp.getJava2DelegationModel() : "true";
        if (delegationModel.equalsIgnoreCase("false")) {
            java2DelegationModel = false;
        } else if (delegationModel.equalsIgnoreCase("true")) {
            java2DelegationModel = true;
        } else {
            throw new WebContainerDeploymentDescException("The java2 delegation model could be 'true' or 'false', not '" + delegationModel + "'.");
        }

        // Security Role List
        SecurityRole securityRole = null;
        securityRoleList = new ArrayList();
        for (Iterator itSecurityRole = webApp.getSecurityRoleList().iterator(); itSecurityRole.hasNext();) {
            securityRole = (SecurityRole) itSecurityRole.next();
            securityRoleList.add(new SecurityRoleDesc(securityRole));
        }


        // servlet classes + names
        List servletList = webApp.getServletList();
        for (Iterator i = servletList.iterator(); i.hasNext();) {
            Servlet servlet = (Servlet) i.next();
            if (servlet.getServletName() != null) {
                ServletDesc servletDesc = new ServletDesc(servlet);
                servlets.put(servlet.getServletName(), servletDesc);
            }
        }


        // Construct servletsUrlMapping
        List urlMappings = webApp.getServletMappingList();
        for (Iterator i = servletList.iterator(); i.hasNext();) {
            Servlet servlet = (Servlet) i.next();
            String name = servlet.getServletName().trim();
            Vector mappings = new Vector();
            for (Iterator m = urlMappings.iterator(); m.hasNext();) {
                ServletMapping sm = (ServletMapping) m.next();
                if (sm.getServletName().trim().equals(name)) {
                    mappings.add(sm.getUrlPattern().trim());
                }
            }
            servletsUrlMapping.put(name, mappings);
        }

        // Build SecurityConstraintListDesc
        securityConstraintListDesc = new SecurityConstraintListDesc(webApp);

    }

    /** 
     * Get an instance of a WEB deployment descriptor by parsing the web.xml 
     * and jonas-web.xml deployment descriptors.
     * @param warFileName the fileName of the war file for the deployment 
     * descriptors.
     * @param classLoaderForCls the classloader for the classes.
     * @param altWebXmlFilename the fileName to the web.xml for the alt-dd
     * tag in the Ear Case. This is used for specify an alternate DDesc file.
     * @return a WEB deployment descriptor by parsing the web.xml and 
     * jonas-web.xml deployment descriptors.
     * @throws DeploymentDescException if the deployment 
     * descriptors are corrupted.
     */
    public static WebContainerDeploymentDesc getInstance(String warFileName,
                                                         ClassLoader classLoaderForCls,
                                                         String altWebXmlFilename)
        throws DeploymentDescException {
 
        // Reinit static values;
        xmlContent = "";
        jonasXmlContent = "";

        //war file
        JarFile warFile = null;

        //Input streams
        InputStream webInputStream = null;
        InputStream jonasWebInputStream = null;
  
        //ZipEntry
        ZipEntry webZipEntry = null;
        ZipEntry jonasWebZipEntry = null;

        //Webapps
        WebApp webApp;
        JonasWebApp jonasWebApp;

        //Build the file
        File fWar = new File(warFileName);

  
        //Check if the file exists.
        if (!(fWar.exists())) {
            String err = "' " + warFileName + "' was not found.";
            throw new WebContainerDeploymentDescException(err);
        }


        //Check if the Alt deploymentDesc file exists.
        //But only if it's a non null value because it's optionnal.
        if ((altWebXmlFilename != null) && (!new File(altWebXmlFilename).exists())) {
            String err = "The file for the altdd tag for the EAR case '" + altWebXmlFilename + "' was not found.";
            throw new WebContainerDeploymentDescException(err);
        }


        // load the web-app deployment descriptor data (WEB-INF/web.xml 
        // and WEB-INF/jonas-web.xml)
        try {

            //No alt-dd case
            if (altWebXmlFilename == null) {
 
                //If warFile is a directory, there is no jar. Check file in the directory
                if (fWar.isDirectory()) {
                    //lookup a WEB-INF/web.xml file
                    File webXmlF = new File(warFileName, WEB_FILE_NAME);
                    if (!webXmlF.exists()) {
                        String err = "You have choose to deploy a war directory but there is no " + WEB_FILE_NAME + " file in the directory " + warFileName;
                        throw new WebContainerDeploymentDescException(err);
                    }
                    webInputStream = new FileInputStream(webXmlF);
                    xmlContent = xmlContent(webInputStream);
                    webInputStream = new FileInputStream(webXmlF);
                } else {
                    warFile = new JarFile(warFileName);
                    //Lookup in the JAR
                    //Check the web entry
                    webZipEntry = warFile.getEntry(WEB_FILE_NAME);
                    if (webZipEntry == null) {
                        throw new WebContainerDeploymentDescException("The entry '" + WEB_FILE_NAME + "' was not found in the file '" + warFileName + "'.");
                    }
                    //Get the stream
                    webInputStream = warFile.getInputStream(webZipEntry);
                    xmlContent = xmlContent(webInputStream);
                    webInputStream = warFile.getInputStream(webZipEntry);
                }
            } else {
                webInputStream = new FileInputStream(altWebXmlFilename);
                xmlContent = xmlContent(webInputStream);
                webInputStream = new FileInputStream(altWebXmlFilename);
            }
   
            //This is a directory
            if (fWar.isDirectory()) {
                //lookup a WEB-INF/jonas-web.xml file
                File webJXmlF = new File(warFileName, JONAS_WEB_FILE_NAME);
                if (webJXmlF.exists()) {
                    jonasWebInputStream = new FileInputStream(webJXmlF);
                    jonasXmlContent = xmlContent(jonasWebInputStream);
                    jonasWebInputStream = new FileInputStream(webJXmlF);
                }
            } else {
                warFile = new JarFile(warFileName);
                //Check the jonas web entry
                jonasWebZipEntry = warFile.getEntry(JONAS_WEB_FILE_NAME);

                //Get the stream
                if (jonasWebZipEntry != null) {
                    jonasWebInputStream = warFile.getInputStream(jonasWebZipEntry);
                    jonasXmlContent = xmlContent(jonasWebInputStream);
                    jonasWebInputStream = warFile.getInputStream(jonasWebZipEntry);

                }
            }

        } catch (Exception e) {
            if (warFile != null) {
                try {
                    warFile.close();
                } catch (IOException ioe) {
                    ; // We can't close the file
                }
            }
            throw new WebContainerDeploymentDescException("Cannot read the XML deployment descriptors of the war file '" + warFileName + "'.", e);
        }

        webApp = loadWebApp(new InputStreamReader(webInputStream), WEB_FILE_NAME);
        try {
            webInputStream.close();
        } catch (IOException e) {
            ; // Nothing to do
        }




        // load jonas-web-app deployment descriptor data 
        // (WEB-INF/jonas-web.xml)
        if (jonasWebInputStream != null) {
            jonasWebApp = loadJonasWebApp(new InputStreamReader(jonasWebInputStream), JONAS_WEB_FILE_NAME);
            try {
                jonasWebInputStream.close();
            } catch (IOException e) {
                ; // Nothing to do
            }
        } else {
            jonasWebApp = new JonasWebApp();
        }


        // instantiate web deployment descriptor
        return new WebContainerDeploymentDesc(warFileName, classLoaderForCls, webApp, jonasWebApp);
    }

    /** 
     * Get an instance of a WEB deployment descriptor by parsing the web.xml 
     * and jonas-web.xml deployment descriptors.
     * @param warFileName the fileName of the war file for the deployment 
     * descriptors.
     * @param classLoaderForCls the classloader for the classes.
     * @return a WEB deployment descriptor by parsing the web.xml and 
     * jonas-web.xml deployment descriptors.
     * @throws DeploymentDescException if the deployment 
     * descriptors are corrupted.
     */
    public static WebContainerDeploymentDesc getInstance(String warFileName, ClassLoader classLoaderForCls)
        throws DeploymentDescException {

        return getInstance(warFileName, classLoaderForCls, null);
    }

    /** 
     * Load the web.xml file.
     * @param reader the reader of the XML file.
     * @param fileName the name of the file (web.xml).
     * @return a structure containing the result of the web.xml parsing.
     * @throws DeploymentDescException if the deployment descriptor 
     * is corrupted.
     */
    public static WebApp loadWebApp(Reader reader, String fileName) 
        throws DeploymentDescException {
  
        WebApp webApp = new WebApp();

        // Create if null
        if (webAppDigester == null) {
            webAppDigester = new JDigester(webAppRuleSet,
                                       parsingWithValidation,
                                       true,
                                       new WebAppDTDs(),
                                       new WebAppSchemas());
        }


        try {
            webAppDigester.parse(reader, fileName, webApp);
        } catch (DeploymentDescException e) {
            throw e;
        } finally {
            webAppDigester.push(null);
        }
        return webApp;
    }

    /**
     * Load the jonas-web.xml file.
     * @param reader the Reader of the XML file.
     * @param fileName the name of the file (jonas-web.xml).
     * @return a structure containing the result of the jonas-web.xml parsing.
     * @throws DeploymentDescException if the deployment 
     * descriptor is corrupted.
     */
    public static JonasWebApp loadJonasWebApp(Reader reader, 
                                              String fileName) 
        throws DeploymentDescException {

        JonasWebApp jonasWebApp = new JonasWebApp();

        // Create if null
        if (jonasWebAppDigester == null) {
            jonasWebAppDigester = new JDigester(jonasWebAppRuleSet,
                                                parsingWithValidation,
                                                true,
                                                new JonasWebAppDTDs(),
                                                new JonasWebAppSchemas()); 
        }

        try {
            jonasWebAppDigester.parse(reader, fileName, jonasWebApp);
        } catch (DeploymentDescException e) {
            throw e;
        } finally {
            jonasWebAppDigester.push(null);
        }
        return jonasWebApp;
    }



    /**
     * Get the context root of this web application.
     * @return the context root of this web application.
     */
    public String getContextRoot() {
        return contextRoot;
    }

    /**
     * Context classloader must follow the java2 delegation model ?
     * @return true if the context's classloader must follow the java 2 delegation model.
     */
    public boolean getJava2DelegationModel() {
        return java2DelegationModel;
    }

    /**
     * Get the host on which the web application must be deployed.
     * @return the host on which the web application must be deployed.
     */
    public String getHost() {
        return host;
    }

    /**
     * Get the prefered port of the connector used to access the web application.
     * @return the prefered port of the connector used to access the web application.
     */
    public String getPort() {
        return port;
    }

    /**
     * Return a String representation of the WebContainerDeploymentDesc.
     * @return a String representation of the WebContainerDeploymentDesc.
     */
    public String toString() {
        StringBuffer ret = new StringBuffer();
  
        // Return the displayName
        ret.append("\ngetDisplayName()=" + getDisplayName());
  
        // Return the resource-env-ref
        ResourceEnvRefDesc[] rer = getResourceEnvRefDesc();
        for (int i = 0; i < rer.length; i++) {
            ret.append("\ngetResourceEnvRefDesc(" + i + ")=" + rer[i].getClass().getName());
            ret.append(rer[i].toString());
        }

        // Return the resource-ref
        ResourceRefDesc[] resourceRefDesc = getResourceRefDesc();
        for (int i = 0; i < resourceRefDesc.length; i++) {
            ret.append("\ngetResourceRefDesc(" + i + ")=" + resourceRefDesc[i].getClass().getName());
            ret.append(resourceRefDesc[i].toString());
        }
 
        // Return the env-entry
        EnvEntryDesc[] envEntries = getEnvEntryDesc();
        for (int i = 0; i < envEntries.length; i++) {
            ret.append("\ngetEnvEntryDesc(" + i + ")=" + envEntries[i].getClass().getName());
            ret.append(envEntries[i].toString());
        }

        // Return the ejb-ref
        EjbRefDesc[] ejbRefDesc = getEjbRefDesc();
        for (int i = 0; i < ejbRefDesc.length; i++) {
            ret.append("\ngetEjbRefDesc(" + i + ")=" + ejbRefDesc[i].getClass().getName());
            ret.append(ejbRefDesc[i].toString());
        }
  
        // Return the ejb-local-ref
        EjbLocalRefDesc[] ejbLocalRefDesc = getEjbLocalRefDesc();
        for (int i = 0; i < ejbLocalRefDesc.length; i++) {
            ret.append("\ngetEjbLocalRefDesc(" + i + ")=" + ejbLocalRefDesc[i].getClass().getName());
            ret.append(ejbLocalRefDesc[i].toString());
        }

        // Return the service-ref-name
        ServiceRefDesc[] svcRef = getServiceRefDesc();
        for (int i = 0; i < svcRef.length; i++) {
            ret.append("\ngetServiceRefDesc(" + i + ")=" + svcRef[i].getClass().getName());
            ret.append(svcRef[i].toString());
        }

        // Return the host
        ret.append("\ngetHost()=" + getHost());

        // Return the context-root
        ret.append("\ngetContextRoot()=" + getContextRoot());
   
        return ret.toString();
    }


    /**
     * Gets the constraint list
     * @return the constraint list
     */
    public static SecurityConstraintListDesc getSecurityConstraintListDesc() {
        return securityConstraintListDesc;
    }

    /**
     * Gets the list of security roles
     * @return the list of security roles
     */
    public List getSecurityRoleList() {
        return securityRoleList;
    }


    /**
     * Get the current ErrorHandler.
     * @return the current ErrorHandler.
     */
    public static ErrorHandler getErrorHandler() {
        return errorHandler;
    }
 
    /**
     * Get the current EntityResolver.
     * @return the current EntityResolver.
     */
    public static EntityResolver getEntityResolver() {
        return entityResolver;
    }

    /**
     * Controls whether the parser is reporting all validity errors.
     * @return if true, all external entities will be read.
     */
    public static boolean getParsingWithValidation() {
        return WebContainerDeploymentDesc.parsingWithValidation;
    }

    /**
     * Set the specified ErrorHandler. If the application does not register
     * an ErrorHandler, all error events reported by the reader are ignored.
     * @param errorHandler the Error Handler to set.
     */
    public static void setErrorHandler(ErrorHandler errorHandler) {
        WebContainerDeploymentDesc.errorHandler = errorHandler;
    }

    /**
     * Set the specified EntityResolver.
     * @param entityResolver the Entity Resolver to set.
     */
    public static void setEntityResolver(EntityResolver entityResolver) {
        WebContainerDeploymentDesc.entityResolver = entityResolver;
    }

    /**
     * Controls whether the parser is reporting all validity errors.
     * @param validation if true, all external entities will be read.
     */
    public static void setParsingWithValidation(boolean validation) {
        WebContainerDeploymentDesc.parsingWithValidation = validation;
    }

    /**
     * Return the content of the inut stream
     * @param in the given input stream
     * @return the content of the inut stream
     * @throws IOException if the file can't be read
     */
    private static String xmlContent(InputStream in) throws IOException {

        StringBuffer sb = new StringBuffer();
        String line;
        BufferedReader br = new BufferedReader(new InputStreamReader(in));
        // loop until the end
        while ((line = br.readLine()) != null) {
            sb.append(line);
            sb.append("\n");
        }
        return sb.toString();
    }


    /**
     * Return the content of the web.xml file
     * @return the content of the web.xml file
     */
    public static String getXmlContent() {
        return xmlContent;
    }

    /**
     * Return the content of the jonas-web.xml file
     * @return the content of the jonas-web.xml file
     */
    public static String getJOnASXmlContent() {
        return jonasXmlContent;
    }

    /**
     * Gets the list of Servlets
     * @return list of Servlets
     */
    public Collection getServletDescList() {
        return servlets.values();
    }

    /**
     * Return a list of all servlets name available
     * @return a list of all servlets name available
     */
    public String[] getServletsName() {
        String[] st = new String[servlets.size()];
        return (String[]) servlets.keySet().toArray(st);
    }

    /**
     * Return the classname of the given servlet
     * @param servName name of the given servlet
     * @return the classname of the given servlet
     */
    public String getServletClassname(String servName) {
        return ((ServletDesc) servlets.get(servName)).getServletClass();
    }

    /**
     * Return the list of urlMapping of the given servlet
     * @param servName name of the given servlet
     * @return the list of urlMapping of the given servlet
     */
    public List getServletMappings(String servName) {
        return (List) servletsUrlMapping.get(servName);
    }

}
