/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial Developer : Delplanque Xavier & Sauthier Guillaume
 * --------------------------------------------------------------------------
 * $Id: MappingFile.java,v 1.8 2003/11/28 14:15:58 sauthieg Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ws.deployment.api;

import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.xml.namespace.QName;

import org.objectweb.jonas_lib.deployment.api.DeploymentDescException;
import org.objectweb.jonas_lib.deployment.api.JDigester;

import org.objectweb.jonas_ws.deployment.rules.JavaWsdlMappingRuleSet;
import org.objectweb.jonas_ws.deployment.xml.JavaWsdlMapping;
import org.objectweb.jonas_ws.deployment.xml.JavaXmlTypeMapping;
import org.objectweb.jonas_ws.deployment.xml.PackageMapping;


/**
 * this Class is used to manipulate jaxrpc-mapping-file. This file contains
 * informations for mapping between XML namespaces and java packages. We
 * actually support just a few part of this file. According with JSR 921, this
 * file must contain class mapping information (exceptions/faults,
 * types/classes, portTypes/interfaces ...).
 *
 * @author Guillaume Sauthier
 * @author Xavier Delplanque
 * @author Helene Joanin
 */
public class MappingFile {

    /**
     * Digester used to parse jaxrpc_mapping.xml
     */
    private static JDigester mfDigester = null;

    /**
     * Rules to parse the jaxrpc_mapping.xml
     */
    private static JavaWsdlMappingRuleSet mfRuleSet = new JavaWsdlMappingRuleSet();

    /** Flag for parser validation */
    private boolean parsingWithValidation = true;

    /** XML JavaWsdlMapping object */
    private JavaWsdlMapping jwMapping = null;

    /** Associates a namespace to its package. It's the cache. */
    private Map namespacePackageMapping;

    /** Associates an XML QName to its java classname */
    private Map xmlType2javaClassname;

    /**
     * Constructor : creates a MappingFile object.
     *
     * @param is mapping file Reader.
     * @param name file name (only for messages).
     * @param validate Flag for parser validation.
     *
     * @throws WSDeploymentDescException When reading/parsing fails.
     */
    public MappingFile(Reader is, String name, boolean validate)
        throws WSDeploymentDescException {

        parsingWithValidation = validate;

        // init vars
        namespacePackageMapping = new Hashtable();
        xmlType2javaClassname = new Hashtable();

        // load and fill the object
        jwMapping  = loadMappingFile(is, name);
        fillPackageMapping(jwMapping);
        fillXmlTypeMapping(jwMapping);

    }

    /**
     * Constructor : creates a MappingFile object.
     *
     * @param is mapping file InputStream.
     * @param name file name (only for messages).
     *
     * @throws WSDeploymentDescException When reading/parsing fails.
     */
    public MappingFile(InputStream is, String name)
        throws WSDeploymentDescException {
        this(new InputStreamReader(is), name, true);
    }

    /**
     * Constructor : creates a MappingFile object.
     *
     * @param is mapping file InputStream.
     * @param name file name (only for messages).
     * @param validate Flag for parser validation.
     *
     * @throws WSDeploymentDescException When reading/parsing fails.
     */
    public MappingFile(InputStream is, String name, boolean validate)
        throws WSDeploymentDescException {
        this(new InputStreamReader(is), name, validate);
    }


    /**
     * Fill up the package mapping hashtable
     *
     * @param mapping jaxrpc-mapping-file root element.
     */
    private void fillPackageMapping(JavaWsdlMapping mapping) {
        List pml = mapping.getPackageMappingList();

        for (Iterator i = pml.iterator(); i.hasNext(); ) {
            // get each mapping file package mappings
            PackageMapping pm = (PackageMapping) i.next();

            String pt = pm.getPackageType();
            String nuri = pm.getNamespaceURI();
            
            //add in the hashtable the namespace to package relation (cache)
            namespacePackageMapping.put(nuri, pt);
        }
    }

    /**
     * Fill up the xml type to java class map
     *
     * @param mapping jaxrpc-mapping-file root element.
     */
    private void fillXmlTypeMapping(JavaWsdlMapping mapping) {

        List jxml = mapping.getJavaXmlTypeMappingList();
        for (Iterator i = jxml.iterator(); i.hasNext();) {
            // get each java -> xml type mapping
            JavaXmlTypeMapping jxtm = (JavaXmlTypeMapping) i.next();
            
            String javaName = jxtm.getClassType();
            QName xmlName = jxtm.getRootTypeQname().getQName();
            
            xmlType2javaClassname.put(xmlName, javaName);
        }
    }

    /**
     * Load the MappingFile file
     *
     * @param reader The mapping file InputStream
     * @param fileName mapping filename
     *
     * @return Zeus Root Element of the xml file.
     *
     * @throws WSDeploymentDescException When parsing fails.
     */
    private JavaWsdlMapping loadMappingFile(Reader reader, String fileName)
        throws WSDeploymentDescException {

        JavaWsdlMapping jwm = new JavaWsdlMapping();

        // Create if mfDigester is null
        if (mfDigester == null) {
            try {
                // Create and initialize the digester
                mfDigester = new JDigester(mfRuleSet,
                                           parsingWithValidation,
                                           true,
                                           null,
                                           new JaxrpcMappingSchemas());
            } catch (DeploymentDescException  e) {
                throw new WSDeploymentDescException(e);
            }
        }

        try {
            mfDigester.parse(reader, fileName, jwm);
        } catch (DeploymentDescException  e) {
            throw new WSDeploymentDescException(e);
        } finally {
            mfDigester.push(null);
        }

        return jwm;
    }

    /**
     * return the mapping between XML namespaces and Java packages defined in
     * the jaxrpc-mapping file.
     *
     * @return the mapping between XML namespaces and Java packages
     */
    public Map getMappings() {
        return namespacePackageMapping;
    }

    /**
     * Return the package associated with the specified namespaceURI (can be
     * null).
     *
     * @param namespaceURI the namespace key to retrieve the package name
     *
     * @return the package associated with the specified namespaceURI. (null if
     *         namespace not present).
     */
    public String getMapping(String namespaceURI) {
        return (String) namespacePackageMapping.get(namespaceURI);
    }

    /**
     * Return the Java classname representing the xml type.
     *
     * @param xmlType the QName of the xml type
     *
     * @return the Java classname representing the xml type.
     */
    public String getClassname(QName xmlType) {
        return (String) xmlType2javaClassname.get(xmlType);
    }

    /**
     * Return an iterator traversing the list of xmlType mappings.
     *
     * @return an iterator traversing the list of xmlType mappings.
     */
    public Iterator getXmlTypeMappings() {
        return xmlType2javaClassname.keySet().iterator();
    }

    /**
     * Return the XML JavaWsdlMapping object
     * @return the XML JavaWsdlMapping object
     */
    public JavaWsdlMapping getXmlJavaWsdlMapping() {
        return jwMapping;
    }

    /**
     * Build a string representation of MappingFile
     *
     * @return String representation of the mapping file.
     */
    public String toString() {
        StringBuffer ret = new StringBuffer();

        ret.append("\n" + getClass().getName());
        ret.append("\ngetMappings()=" + namespacePackageMapping);
        ret.append("\ngetXmlTypeMappings()=" + xmlType2javaClassname);

        return ret.toString();
    }

    /**
     * Return true if the 2 objects are equals in value.
     *
     * @param other the object to compare.
     *
     * @return true if objects are equals in value, else false.
     */
    public boolean equals(Object other) {
        if (this == other) {
            return true;
        }
        if (other == null) {
            return false;
        }
        if (!(other instanceof MappingFile)) {
            return false;
        }
        MappingFile ref = (MappingFile) other;
        if (!namespacePackageMapping.equals(ref.getMappings())) {
            return false;
        }
        // After all theses tests, the 2 objects are equals in value
        return true;
    }

}
