/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial Developer : Delplanque Xavier & Sauthier Guillaume
 * --------------------------------------------------------------------------
 * $Id: PortComponentDesc.java,v 1.9 2004/03/26 08:57:04 sauthieg Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ws.deployment.api;

import java.net.URL;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import javax.xml.namespace.QName;

import org.objectweb.jonas_lib.deployment.api.DeploymentDescException;
import org.objectweb.jonas_lib.deployment.api.HandlerDesc;
import org.objectweb.jonas_lib.deployment.xml.Handler;
import org.objectweb.jonas_ws.deployment.xml.PortComponent;



/**
 * This class provides methodes to manipulate portComponent's attributes.
 *
 * @author Guillaume Sauthier
 * @author Xavier Delplanque
 */

public abstract class PortComponentDesc {

    /** The name of the Port (must be unique) */
    private String name;

    /** The interface name of the endpoint/port */
    private String sei;

    /** the sib link, can be ejb or servlet link */
    protected String sibLink;

    /** The classname of bean implementing the service */
    protected String sib;

    /** The list of Handlers the Port will run */
    private Vector handlers = new Vector();

    /** The QName of the WSDL port */
    private QName portQName;

    /** The URL where to access the port-component (null before resolving) */
    private URL endpoint = null;

    /** parent container service */
    private ServiceDesc parent = null;

    /**
     * Creates a new PortComponentDesc object.
     *
     * @param jarCL the module (ejbjar or war) classloader.
     * @param pc webservices portcomponent element
     *
     * @throws WSDeploymentDescException When construction fails.
     */
    protected PortComponentDesc(ClassLoader jarCL,
                                PortComponent pc,
                                ServiceDesc parent)
        throws WSDeploymentDescException {
        
        this.parent = parent;

        // set name
        name = pc.getPortComponentName();
        if (name.equals("")) {
            String err = "port-component-name must have a value.";
            throw new WSDeploymentDescException(err);
        }

        // set sei
        sei = pc.getServiceEndpointInterface();
        if (sei.equals("")) {
            String err = "service-endpoint-interface must have a value.";
            throw new WSDeploymentDescException(err);
        }
            
        // set and init handlers
        List hl = pc.getHandlerList();
        Handler h = null;

        for (int i = 0; i < hl.size(); i++) {
            // for each reference, build and add a HandlerDesc Object
            if (hl.get(i) != null) {
                // get the standard dd handler
                h = (Handler) hl.get(i);
                // build and add a new handler
                try {
                    handlers.add(new HandlerDesc(jarCL, h));
                } catch (DeploymentDescException dde) {
                    throw new WSDeploymentDescException(dde);
                }
            }
        }

        // set portQName
        portQName = pc.getWsdlPort().getQName();

    }

    /**
     * Return the parent ServiceDesc of the PortComponent.
     *
     * @return the parent ServiceDesc of the PortComponent.
     */
    public ServiceDesc getServiceDesc() {
        return parent;
    }

    /**
     * Return the name of the PortComponent.
     *
     * @return the name of the PortComponent.
     */
    public String getName() {
        return name;
    }

    /**
     * Return the Service Endpoint Interface name.
     *
     * @return the Service Endpoint Interface name.
     */
    public String getServiceEndpointInterface() {
        return sei;
    }

    /**
     * Return the implementation bean. It's the bean classname 
     * which do the real work.
     *
     * @return the implementation bean classname.
     */
    public String getSIBClassname() {
        return sib;
    }

    /**
     * Set the sib class name.
     *
     * @param sibClassName the sib class name.
     *
     * @deprecated sib class set with the setSessionStatelessDesc 
     *             or setWebDesc methods.
     */
    protected void setSIBClassname(String sibClassName) {
        this.sib = sibClassName;
    }

    /**
     * Return the WSDL's Port QName, the PortComponent is asssociated with.
     *
     * @return the port's QName
     */
    public QName getQName() {
        return portQName;
    }

    /**
     * Return the list of Handlers the PortComponent is associated with.
     *
     * @return the list of Handlers the PortComponent is associated with.
     */
    public List getHandlers() {
        return handlers;
    }

    /**
     * Return the service-impl-bean value. the sib can be an ejb or a servlet.
     *
     * @return the sib name.
     */
    public String getSibLink() {
        return sibLink;
    }

    /**
     * Return true if the Service Impl Bean is an EJB.
     *
     * @return true if the Service Impl Bean is an EJB.
     */
    public abstract boolean hasBeanImpl();

    /**
     * Return true if the Service Impl Bean is a JaxRpc component.
     *
     * @return true if the Service Impl Bean is a JaxRpc component.
     */
    public abstract boolean hasJaxRpcImpl();

    /**
     * Return the URL where the port-component can be accessed.
     *
     * @return the URL where the port-component can be accessed.
     */
    public URL getEndpointURL() {
        return endpoint;
    }

    /**
     * Set the Endpoint URL of the port-component.
     *
     * @param url the resolved endpoint URL.
     */
    public void setEndpointURL(URL url) {
        endpoint = url;
    }

    /**
     * Setter method for J2EE component linking.
     * @param desc the descriptor of the component implementing the endpoint.
     * @throws WSDeploymentDescException when desc is an unknown type.
     */
    public abstract void setDesc(Object desc) throws WSDeploymentDescException;

    /**
     * return a String representation of the PortComponent
     */
    public String toString() {

        StringBuffer sb = new StringBuffer();
        
        sb.append("\n" + getClass().getName());
        sb.append("\ngetName()=" + getName());
        sb.append("\ngetServiceEndpointInterface()=" + getServiceEndpointInterface());
        sb.append("\ngetSibLink()=" + getSibLink());
        sb.append("\ngetSIBClassname()=" + getSIBClassname());
        sb.append("\ngetQName()=" + getQName());

        for (Iterator i = getHandlers().iterator(); i.hasNext(); )
            sb.append("\ngetHandlers()=" + ((Handler) i.next()).toString());
       

        return sb.toString();

    }
}
