/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial Developer : Delplanque Xavier & Sauthier Guillaume
 * --------------------------------------------------------------------------
 * $Id: WSDeploymentDescManager.java,v 1.11 2004/03/19 14:31:53 sauthieg Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ws.deployment.lib;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import org.objectweb.jonas.common.Log;
import org.objectweb.jonas_ejb.deployment.api.BeanDesc;
import org.objectweb.jonas_ejb.deployment.api.DeploymentDesc;
import org.objectweb.jonas_ejb.deployment.api.SessionStatelessDesc;
import org.objectweb.jonas_ejb.deployment.lib.EjbDeploymentDescManager;
import org.objectweb.jonas_lib.deployment.api.DeploymentDescException;
import org.objectweb.jonas_lib.deployment.api.JDigester;
import org.objectweb.jonas_web.deployment.api.WebContainerDeploymentDesc;
import org.objectweb.jonas_web.deployment.api.WebContainerDeploymentDescException;
import org.objectweb.jonas_web.deployment.lib.WebDeploymentDescManager;
import org.objectweb.jonas_ws.deployment.api.PortComponentDesc;
import org.objectweb.jonas_ws.deployment.api.ServiceDesc;
import org.objectweb.jonas_ws.deployment.api.WSDeploymentDesc;
import org.objectweb.jonas_ws.deployment.api.WSDeploymentDescException;
import org.objectweb.jonas_ws.deployment.rules.JonasWebservicesRuleSet;
import org.objectweb.jonas_ws.deployment.rules.WebservicesRuleSet;
import org.objectweb.jonas_ws.deployment.xml.JonasWebservices;
import org.objectweb.jonas_ws.deployment.xml.Webservices;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;


/**
 * This class provide a way for managing the WSDeploymentDesc. Note that there
 * is 1 instance of the WSDeploymentDescManager on each JOnAS server.
 *
 * @author Guillaume Sauthier
 * @author Xavier Delplanque
 * @author Helene Joanin
 */
public class WSDeploymentDescManager {

    /**
     * Digester used to parse webservices.xml and jonas-webservices.xml
     */
    private static JDigester wsDigester = null;
    private static JDigester jwsDigester = null;

    /**
     * Rules to parse the webservices.xml
     */
    private static WebservicesRuleSet wsRuleSet = new WebservicesRuleSet();
    /**
     * Rules to parse the jonas-webservices.xml
     */
    private static JonasWebservicesRuleSet jwsRuleSet = new JonasWebservicesRuleSet();

    /** Flag for parser validation */
    private static boolean parsingWithValidation = true;

    /** The unique instance of the WSDeploymentDescManager */
    private static WSDeploymentDescManager unique = null;

    /** is the manager instanciated ? */
    private static boolean isInstanciated = false;

    /** Logger for the deployment desc manager. */
    private static Logger logger = Log.getLogger(Log.JONAS_WS_PREFIX);

    /** Associates module or webservices dd URL  and a WSDeploymentDesc */
    private HashMap urlWsddBindings;

    /** Associates a given ClassLaoder (ear/ejb/web) to a list of URL containing (or not) WSDD */
    private Map classLoader2URLs;

    /** Reference on the EjbDeploymentDescManager. */
    private EjbDeploymentDescManager ejbDDManager = null;

    /** Reference on the WebDeploymentDescManager. */
    private WebDeploymentDescManager wcDDManager = null;

    /**
     * Constructor : creates a WSDeploymentDescManager object. Used only when
     * loading files in the running JOnAS instance.
     */
    private WSDeploymentDescManager() {
        urlWsddBindings = new HashMap();
        classLoader2URLs = new Hashtable();
        ejbDDManager = EjbDeploymentDescManager.getInstance();
        wcDDManager = WebDeploymentDescManager.getInstance();
    }

    /**
     * Get the unique instance of the WSDeploymentDescManager.
     *
     * @return the instance of the WSDeploymentDescManager.
     */
    public static WSDeploymentDescManager getInstance() {
        if (!isInstanciated) {
            isInstanciated = true;
            unique = new WSDeploymentDescManager();

        }
        return unique;
    }

    /**
     * Get the specified WebService deployment descriptor. Used by WsGen.
     *
     * @param file module (ejbjar or war). It is required that file is not a
     *        directory.
     * @param jarCL the classloader where classes are stored.
     *
     * @return the module webservices deployment descriptor if it exists, null
     *         else.
     *
     * @throws WSDeploymentDescException when WSDeploymentDesc cannot be
     *         created with the given files.
     */
    public static WSDeploymentDesc getDeploymentDesc(String      file,
                                                     ClassLoader jarCL)
        throws WSDeploymentDescException {

        URL fileUrl = null;
        try {
            fileUrl = new File(file).toURL();
        } catch (MalformedURLException mue) {
            throw new WSDeploymentDescException(mue);
        }

        WSDeploymentDesc wsdd = loadDeploymentDesc(fileUrl, jarCL);

        // webservices.xml not present in file, no resolution needed
        if (wsdd == null) {
            return null;

        }

        // Resolve ejb-link and servlet-link elements
        List sdl = wsdd.getServiceDescs();

        for (int i = 0; i < sdl.size(); i++) {
            List pcdl = ((ServiceDesc) sdl.get(i)).getPortComponents();

            for (int j = 0; j < pcdl.size(); j++) {
                // Resolve the ejb-link and servlet-link of each port components
                PortComponentDesc pcd = (PortComponentDesc) pcdl.get(j);
                String sibLink = pcd.getSibLink();

                // a link is defined, we resolve it
                if (pcd.hasBeanImpl()) {
                    SessionStatelessDesc desc =
                        getBeanDesc(file, sibLink, jarCL);
                    pcd.setDesc(desc);

                } else if (pcd.hasJaxRpcImpl()) {
                    WebContainerDeploymentDesc desc =
                        getWebDesc(file, sibLink, jarCL);
                    pcd.setDesc(desc);

                }
            }
        }

        return wsdd;
    }

    /**
     * Get the specified ws deployment descriptor and put it in the cache if it
     * is not in.
     *
     * @param url module (ejbjar or war). It can be a directory or a jar file.
     * @param jarCL classloader used to load bean classes.
     * @param earCL the parent classloader (the ear classloader). Null when not
     *        in the case of an ear application.
     *
     * @return the module webservices deployment descriptor if it exists, null
     *         else.
     *
     * @throws WSDeploymentDescException when WSDeploymentDesc cannot be
     *         created with the given files.
     */
    public WSDeploymentDesc getDeploymentDesc(URL         url,
                                              ClassLoader jarCL,
                                              ClassLoader earCL)
        throws WSDeploymentDescException {

        WSDeploymentDesc wsdd = null;
        File moduleFile = new File(url.getFile());

        if (moduleFile.exists()) {
            // if the module has yet been loaded
            if (!urlWsddBindings.containsKey(url)) {
                wsdd = getDeploymentDescriptor(url, jarCL, earCL);
            } else {  // module already loaded
                // Can be null if no webservices.xml has been found in specified module
                wsdd = (WSDeploymentDesc) urlWsddBindings.get(url);
            }
        } else {
            throw new WSDeploymentDescException("'" + moduleFile.getName()
                                                + "' doesn't exists");
        }

        // Store binding
        ClassLoader keyCL = (earCL != null) ? earCL : jarCL;
        List urls = (List) classLoader2URLs.get(keyCL);
        if (urls == null) {
            urls = new Vector();
            classLoader2URLs.put(keyCL, urls);
        }
        urls.add(url);

        return wsdd;
    }

    /**
     * return the specified webservices deployment descriptor and put it in
     * cache.
     *
     * @param url an existing module (ejbjar or war). It can be a directory or
     *        a jar file
     * @param jarCL classloader used to load war or ejbjar classes.
     * @param earCL the parent classloader (the ear classloader). Null when not
     *        in the case of an ear application.
     *
     * @return the module webservices deployment descriptor if it exists, null
     *         else.
     *
     * @throws WSDeploymentDescException when WSDeploymentDesc cannot be
     *         created with the given files.
     */
    private WSDeploymentDesc getDeploymentDescriptor(URL         url,
                                                     ClassLoader jarCL,
                                                     ClassLoader earCL)
        throws WSDeploymentDescException {

        WSDeploymentDesc wsdd = loadDeploymentDesc(url, jarCL);

        // add in cache the deployment descriptor and return it
        try {
            urlWsddBindings.put(url, wsdd);

        } catch (NullPointerException npe) {
            throw new WSDeploymentDescException(npe);

        }

        // No WSDeploymentDesc, return null
        if (wsdd == null) {
            return null;

        }

        // Resolve ejb-link and servlet-link elements
        List sdl = wsdd.getServiceDescs();

        for (int i = 0; i < sdl.size(); i++) {
            List pcdl = ((ServiceDesc) sdl.get(i)).getPortComponents();

            for (int j = 0; j < pcdl.size(); j++) {
                // Resolve the ejb-link and servlet-link of each port components
                PortComponentDesc pcd = (PortComponentDesc) pcdl.get(j);
                String sibLink = pcd.getSibLink();

                // a link is defined, we resolve it
                if (pcd.hasBeanImpl()) {
                    SessionStatelessDesc desc =
                        getBeanDesc(url.getFile(), sibLink, jarCL, earCL);
                    pcd.setDesc(desc);

                } else if (pcd.hasJaxRpcImpl()) {
                    WebContainerDeploymentDesc desc =
                        getWebDesc(url.getFile(), sibLink, jarCL, earCL);
                    pcd.setDesc(desc);

                }
            }
        }

        return wsdd;
    }

    /**
     * Load webservices.xml in the WSDeploymentDesc structure (null if
     * webservices.xml not found). Can take as parameter a directory (like an
     * unpacked jar root directory), a XML file name (path to
     * (META-INF|WEB-INF)/webservices.xml) or a packed jar.
     *
     * @param url the url to load (can be a directory, xml, jar).
     * @param cl the classLoader to use for loading files.
     *
     * @return WSDeploymentDesc object.
     *
     * @throws WSDeploymentDescException when parsing error.
     */
    private static WSDeploymentDesc loadDeploymentDesc(URL url, ClassLoader cl)
        throws WSDeploymentDescException {

        URLClassLoader ucl = (URLClassLoader) cl;
        
        // get the Webservices object for this module
        InputStream isWebservices = null;

        // get the JonasWebservices object for this module
        InputStream isJonasWebservices = null;
            
        URL wsXml = null;
        URL jwsXml = null;

        try {
            
            // EjbJar archive
            if (url.getFile().toLowerCase().endsWith(".jar")) {

                boolean foundInEjb = false;
                for (Enumeration ejbWebservices = ucl.findResources("META-INF/webservices.xml");
                     ejbWebservices.hasMoreElements() && !foundInEjb;) {
                    wsXml = (URL) ejbWebservices.nextElement();
                    if (isResourceInFile(url, wsXml)) {
                        foundInEjb = true;
                    }
                }

                // if DD found in Ejb
                if (foundInEjb) {
                    isWebservices = wsXml.openStream();
                    
                    foundInEjb = false;
                    // try load jonas DD
                    for (Enumeration ejbJonasWebservices =
                             ucl.findResources("META-INF/jonas-webservices.xml");
                         ejbJonasWebservices.hasMoreElements() && !foundInEjb; ) {
                        jwsXml = (URL) ejbJonasWebservices.nextElement();
                        if (isResourceInFile(url, jwsXml)) {
                            foundInEjb = true;
                        }                
                    }
                    if (foundInEjb) {
                        isJonasWebservices = jwsXml.openStream();
                    }
                }
                
                // webservices.xml not found in Ejb, Try web location
            } else if (url.getFile().toLowerCase().endsWith(".war")) {

                boolean foundInWeb = false;
                for (Enumeration webWebservices = ucl.findResources("WEB-INF/webservices.xml");
                     webWebservices.hasMoreElements() && !foundInWeb;) {
                    wsXml = (URL) webWebservices.nextElement();
                    if (isResourceInFile(url, wsXml)) {
                        foundInWeb = true;
                    }
                }

                // if DD found in Web
                if (foundInWeb) {
                    isWebservices = wsXml.openStream();
                    
                    foundInWeb = false;
                    // try load jonas DD
                    for (Enumeration webJonasWebservices =
                             ucl.findResources("WEB-INF/jonas-webservices.xml");
                         webJonasWebservices.hasMoreElements() && !foundInWeb; ) {
                        jwsXml = (URL) webJonasWebservices.nextElement();
                        if (isResourceInFile(url, jwsXml)) {
                            foundInWeb = true;
                        }                
                    }
                    if (foundInWeb) {
                        isJonasWebservices = jwsXml.openStream();
                    }
                }
            }
        } catch (IOException ioe) {
            String err = "Cannot open Streams on WebServices Deployment Descriptor for '"
                + url + "'";
            throw new WSDeploymentDescException(err, ioe);
        }
            
        // webservices.xml not found in module, return null
        if (isWebservices == null) {
            return null;   
        }
        
        // Load and Parse files
        Webservices webservices = loadWebservices(new InputStreamReader(isWebservices),
                                                  wsXml.getFile());
        JonasWebservices jonasWebservices = null;
        
        if (isJonasWebservices != null) {
            jonasWebservices =
                loadJonasWebservices(new InputStreamReader(isJonasWebservices),
                                     jwsXml.getFile());
        }

        try {
            if (isWebservices != null) {
                isWebservices.close();
            }
            if (isJonasWebservices != null) {
                isJonasWebservices.close();
            }
        } catch (IOException ioe) {
            // nothing to do, just log a warning
            String err = "Cannot close InputStreams for '" + url + "'";
            logger.log(BasicLevel.WARN, err);
        }
        
        // Create DeploymentDesc
        WSDeploymentDesc wsdd =
            new WSDeploymentDesc(cl, logger, webservices, jonasWebservices);

        return wsdd;

    }

    /**
     * @return true if the URL resource is located in the specified file.
     */
    private static boolean isResourceInFile(URL file, URL resource) {

        boolean within = false;

        File f = new File(file.getFile());
        if (f.isDirectory()) {
            // test directory
            // resource need to be of the form :
            // <resource> = <file>/META-INF/webservices.xml for example
            if (resource.toString().startsWith(file.toString())) {
                within = true;
                logger.log(BasicLevel.DEBUG, "Directory case. found '" 
                           + resource + "' in '" + file + "'");
            }
        } else if (f.isFile()) {
            if (f.getPath().endsWith(".xml")) {
                // test XML
                // resource need to be of the form :
                // <file> = <resource>
                if (file.equals(resource)) {
                    within = true;
                    logger.log(BasicLevel.DEBUG, "XML case. found '" 
                               + resource + "' in '" + file + "'");
                }
            } else if ((f.getPath().endsWith(".war")) 
                       || (f.getPath().endsWith(".jar"))) {
                // test JarFile
                // resource need to be of the form :
                // <resource> = jar:<file>!/META-INF/webservices.xml for example
                int index = resource.toString().indexOf("!/");
                if (index != -1) {
                    if (resource.toString().startsWith(("jar:" + file.toString()))) {
                        within = true;
                        logger.log(BasicLevel.DEBUG, "Archive case. found '" 
                                   + resource + "' in '" + file + "'");
                    }
                }
            }
        }
        return within;
    }

    /**
     * return a 'xml' object containing webservices.xml informations
     *
     * @param is webservices.xml file input stream
     * @param name webservices.xml file name
     *
     * @return a 'xml' object containing webservices.xml informations
     *
     * @throws WSDeploymentDescException when Webservices cannot be created
     *         with the given files.
     */
    public static Webservices loadWebservices(Reader reader, String fileName)
        throws WSDeploymentDescException {

        Webservices ws = new Webservices();

        // Create if wsDigester is null
        if (wsDigester == null) {
            try {
                // Create and initialize the digester
                wsDigester = new JDigester(wsRuleSet,
                                           parsingWithValidation,
                                           true,
                                           null,
                                           new WsSchemas());
            } catch (DeploymentDescException  e) {
                throw new WSDeploymentDescException(e);
            }
        }

        try {
            wsDigester.parse(reader, fileName, ws);
        } catch (DeploymentDescException  e) {
            throw new WSDeploymentDescException(e);
        } finally {
            wsDigester.push(null);
        }

        return ws;
    }

    /**
     * return a 'xml' object containing jonas-webservices.xml informations
     *
     * @param is jonas-webservices.xml file input stream
     * @param name jonas-webservices.xml file name
     *
     * @return a 'xml' object containing jonas-webservices.xml informations
     *
     * @throws WSDeploymentDescException when JonasWebservices cannot be
     *         created with the given files.
     */
    public static JonasWebservices loadJonasWebservices(Reader reader, String fileName)
        throws WSDeploymentDescException {

        JonasWebservices jws = new JonasWebservices();

        // Create if wsDigester is null
        if (jwsDigester == null) {
            try {
                // Create and initialize the digester
                jwsDigester = new JDigester(jwsRuleSet,
                                            parsingWithValidation,
                                            true,
                                            null,
                                            new JonasWsSchemas());
            } catch (DeploymentDescException  e) {
                throw new WSDeploymentDescException(e);
            }
        }

        try {
            jwsDigester.parse(reader, fileName, jws);
        } catch (DeploymentDescException  e) {
            throw new WSDeploymentDescException(e);
        } finally {
            jwsDigester.push(null);
        }

        return jws;

    }

    /**
     * Get the size of the cache (number of entries in the cache). Used only
     * for debugging.
     *
     * @return the size of the cache (number of entries in the cache).
     */
    public int getCacheSize() {
        return urlWsddBindings.size();

    }

    /**
     * Clear the cache.
     */
    public void clearCache() {
        urlWsddBindings = new HashMap();
    }

    /**
     * Remove the DD cache for the specified ClassLoader
     */
    public void removeCache(ClassLoader cl) {
        List urls = (List) classLoader2URLs.remove(cl);
        if (urls != null) {
            for (Iterator i = urls.iterator(); i.hasNext();) {
                URL url = (URL) i.next();
                urlWsddBindings.remove(url);
            }
        }
    }

    /**
     * Return a string representation of the cache. (Used only for debugging).
     *
     * @return a string representation of the cache.
     */
    public String toString() {
        return "" + getCacheSize();

    }

    /**
     * DOCUMENT ME!
     *
     * @param v DOCUMENT ME!
     */
    public static void setParsingWithValidation(boolean v) {
        parsingWithValidation = v;

    }

    /**
     * DOCUMENT ME!
     *
     * @return DOCUMENT ME!
     */
    public static boolean getParsingWithValidation() {
        return (parsingWithValidation);

    }

    /**
     * Return a StatelessSessionDesc found in filename with the name link. Used
     * by WsGen only. (No JOnAS running).
     *
     * @param filename the ejb-jar file to load.
     * @param link the ejb-name to retrieve.
     * @param cl the classLoader to use.
     *
     * @return he StatelessSessionDesc associated with this port component
     *
     * @throws WSDeploymentDescException if ejb-link not found or something
     *         went wrong when parsing ejb deployment desc.
     */
    private static SessionStatelessDesc getBeanDesc(String      filename,
                                                    String      link,
                                                    ClassLoader cl)
        throws WSDeploymentDescException {

        BeanDesc bd = null;

        try {
            DeploymentDesc dd =
                EjbDeploymentDescManager.getDeploymentDesc(filename, cl);
            bd = dd.getBeanDesc(link);

        } catch (DeploymentDescException dde) {
            throw new WSDeploymentDescException("DeploymentDescException when searching '"
                + link + "' in '" + filename + "'", dde);
        }

        if (bd == null) {
            throw new WSDeploymentDescException("Unable to find the ejb-link '"
                + link + "' in '" + filename + "'");
        }

        if (!(bd instanceof SessionStatelessDesc)) {
            throw new WSDeploymentDescException("ejb-link '" + link
                                                + "' must be a Stateless Session Bean not a '" 
                                                + bd.getClass().getName() + "'");
        }

        return (SessionStatelessDesc) bd;

    }

    /**
     * Return a StatelessSessionDesc found in filename with the name link. Used
     * when JOnAS is running.
     *
     * @param filename the ejb-jar file to load.
     * @param link the ejb-name to retrieve.
     * @param jarCL the loader for classes.
     * @param earCL the EAR classLoader .
     *
     * @return the StatelessSessionDesc associated with this port component
     *
     * @throws WSDeploymentDescException if ejb-link not found or something
     *         went wrong when parsing ejb deployment desc.
     */
    private SessionStatelessDesc getBeanDesc(String      filename,
                                             String      link,
                                             ClassLoader jarCL,
                                             ClassLoader earCL) throws WSDeploymentDescException {
        URL url;

        try {
            url = new File(filename).toURL();

        } catch (MalformedURLException mue) {
            throw new WSDeploymentDescException("Url Error with '" + filename
                                                + "'", mue);
        }

        BeanDesc bd = null;

        try {
            DeploymentDesc dd =
                ejbDDManager.getDeploymentDesc(url, jarCL, earCL);
            bd = dd.getBeanDesc(link);

        } catch (DeploymentDescException dde) {
            throw new WSDeploymentDescException("Unable to load EjbJar '" + url
                                                + "'", dde);
        }

        // Ejb Link not found
        if (bd == null) {
            throw new WSDeploymentDescException("Unable to find the ejb-link '"
                                                + link + "' in '" + filename + "'");
        }

        if (!(bd instanceof SessionStatelessDesc)) {
            throw new WSDeploymentDescException("ejb-link '" + link
                                                + "' must be a Stateless Session Bean");
        }

        return (SessionStatelessDesc) bd;

    }

    /**
     * Return the Web Deployment Desc found in filename with the name link.
     *
     * @param filename the war file to load.
     * @param link the servlet-name to retrieve.
     * @param cl the classLoader to use.
     *
     * @return the Web Deployment Desc associated with this port component
     *
     * @throws WSDeploymentDescException if servlet-link not found or something
     *         went wrong when parsing web deployment desc.
     */
    private static WebContainerDeploymentDesc getWebDesc(String      filename,
                                                         String      link,
                                                         ClassLoader cl)
        throws WSDeploymentDescException {

        WebContainerDeploymentDesc wd = null;

        try {
            wd = WebDeploymentDescManager.getDeploymentDesc(filename, cl);
        } catch (WebContainerDeploymentDescException wcdde) {
            throw new WSDeploymentDescException("Error while reading/parsing "
                                                + filename, wcdde);
        }

        // link not found in this deployment desc
        if (wd.getServletClassname(link) == null) {
            throw new WSDeploymentDescException("Unable to find the servlet-link '" + link + "' in '"
                                                + filename + "'");
        }

        return wd;

    }

    /**
     * Return the Web Deployment Desc found in filename with the name link.
     * Used when JOnAS is up and running.
     *
     * @param filename the war file to load.
     * @param link the servlet-name to retrieve.
     * @param jarCL the classLoader to us.
     * @param earCL the EAR classLoader.
     *
     * @return the Web Deployment Desc associated with this port component
     *
     * @throws WSDeploymentDescException if servlet-link not found or something
     *         went wrong when parsing web deployment desc.
     */
    private WebContainerDeploymentDesc getWebDesc(String      filename,
                                                  String      link,
                                                  ClassLoader jarCL,
                                                  ClassLoader earCL)
        throws WSDeploymentDescException {

        URL url;

        try {
            url = new File(filename).toURL();
        } catch (MalformedURLException mue) {
            throw new WSDeploymentDescException("Url Error with '" + filename
                                                + "'", mue);
        }

        WebContainerDeploymentDesc wd = null;

        try {
            wd = wcDDManager.getDeploymentDesc(url, jarCL, earCL);
        } catch (WebContainerDeploymentDescException wcdde) {
            throw new WSDeploymentDescException("Error while reading/parsing "
                                                + filename, wcdde);
        }

        // link not found in this deployment desc
        if (wd.getServletClassname(link) == null) {
            throw new WSDeploymentDescException("Unable to find the servlet-link '" + link + "' in '"
                                                + filename + "'");
        }

        return wd;

    }
}
