/*
 * JOnAS : Java(TM) OpenSource Application Server
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial Developer :
 * --------------------------------------------------------------------------
 * $Id: WsGen.java,v 1.15 2004/02/16 12:16:41 sauthieg Exp $
 * --------------------------------------------------------------------------
*/

package org.objectweb.jonas_ws.wsgen;

import org.objectweb.jonas_ws.wsgen.archive.Archive;
import org.objectweb.jonas_ws.wsgen.generator.Generator;
import org.objectweb.jonas_ws.wsgen.generator.GeneratorFactory;
import org.objectweb.jonas_ws.wsgen.modifier.ArchiveModifier;
import org.objectweb.jonas_ws.wsgen.modifier.ModifierFactory;
import org.objectweb.jonas_ws.wsgen.utils.TempRepository;

import java.io.File;

import java.util.StringTokenizer;

import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import org.objectweb.jonas.common.Log;

/**
 * Main WsGen Class
 *
 * @author Guillaume Sauthier
 */
public class WsGen {

    /** logger */
    private static final Logger logger = Log.getLogger(Log.JONAS_WSGEN_PREFIX);

    /**
     * Main method.
     *
     * @param args Commend line args
     *
     * @throws WsGenException When a failure occurs
     */
    public static void main(String[] args) throws Exception {
        GeneratorFactory gf = GeneratorFactory.getInstance();

        // store configuration properties
        Generator.Config config = parseInput(args);

        if (config.help || config.error) {
            // display usage message
            usage();
        } else {

            logger.log(BasicLevel.INFO, "Generating '" + config.inputname + "'");
            
            // setup configuration
            gf.setConfiguration(config);
            
            // get the ArchiveModifier
            ArchiveModifier am = ModifierFactory.getModifier(config.inputname);
            
            // modify the given archive
            Archive archive = am.modify();
        }

        // delete all the temporary files created
        TempRepository.getInstance().deleteAll();

    }

    /**
     * Parse Command Line input and returns a Config object
     *
     * @param args Command Lines params
     *
     * @return a Config object
     */
    private static Generator.Config parseInput(String[] args) {
        Generator.Config config = new Generator.Config();

        // Get args
        for(int argn = 0; argn < args.length; argn++) {
            String arg = args[argn];
    
            if(arg.equals("-help") || arg.equals("-?")) {
                config.help = true;

                continue;
            }

            if(arg.equals("-verbose")) {
                config.verbose = true;

                continue;
            }

            if(arg.equals("-debug")) {
                config.debug = true;
                config.verbose = true;

                continue;
            }

            if(arg.equals("-keepgenerated")) {
                config.keepGenerated = true;

                continue;
            }

            if(arg.equals("-noconfig")) {
                config.noConfig = true;

                continue;
            }

            if(arg.equals("-novalidation")) {
                config.parseWithValidation = false;

                continue;
            }

            if(arg.equals("-javac")) {
                config.nameJavac = args[++argn];

                continue;
            }

            if(arg.equals("-javacopts")) {
                argn++;

                if(argn < args.length) {
                    StringTokenizer st = new StringTokenizer(args[argn]);

                    while(st.hasMoreTokens()) {
                        config.javac_opts.add(st.nextToken());
                    }
                } else {
                    config.error = true;
                }

                continue;
            }

            if(arg.equals("-d")) {
                argn++;

                if(argn < args.length) {
                    config.out = new File(args[argn]);
                } else {
                    config.error = true;
                }

                continue;
            }

            config.inputname = args[argn];
        }

        return config;

    }

    /**
     * Display the usage
     */
    public static void usage() {
        StringBuffer msg = new StringBuffer();
        msg.append("Usage: java org.objectweb.jonas_ws.wsgen.WsGen -help \n");
        msg.append("   to print this help message \n");
        msg.append(
            " or java org.objectweb.jonas_ws.wsgen.WsGen <Options> <Input_File> \n");
        msg.append("Options include: \n");
        msg.append(
            "    -d <output_dir>  specify where to place the generated files \n");
        msg.append(
            "    -novalidation parse the XML deployment descriptors without \n");
        msg.append("      validation \n");
        msg.append("    -javac  <opt> specify the java compiler to use \n");
        msg.append(
            "    -javacopts <opt> specify the options to pass to the java compiler \n");
        msg.append(
            "    -keepgenerated   do not delete intermediate generated files \n");
        msg.append(
            "    -noconfig  do not generate configuration files (require \n");
        msg.append("      user provided files) \n");
        msg.append("    -verbose \n");
        msg.append("    -debug \n");
        msg.append(" \n");
        msg.append("    Input_File    the ejb-jar, war or ear filename\n");
        System.out.println(msg.toString());
    }
}
