/*
 * JOnAS : Java(TM) OpenSource Application Server
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial Developer : Guillaume Sauthier
 * --------------------------------------------------------------------------
 * $Id: Application.java,v 1.2 2004/02/16 12:16:42 sauthieg Exp $
 * --------------------------------------------------------------------------
*/

package org.objectweb.jonas_ws.wsgen.archive;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Vector;
import java.util.jar.Attributes;
import java.util.jar.JarFile;
import javax.xml.parsers.ParserConfigurationException;
import org.objectweb.jonas_ear.deployment.api.EarDeploymentDesc;
import org.objectweb.jonas_ear.deployment.xml.Web;
import org.objectweb.jonas_lib.deployment.api.DeploymentDescException;
import org.objectweb.jonas_lib.loader.EjbJarClassLoader;
import org.objectweb.jonas_ws.wsgen.WsGenException;
import org.objectweb.jonas_ws.wsgen.utils.FileUtils;
import org.objectweb.jonas_ws.wsgen.utils.XMLUtils;
import org.objectweb.util.monolog.api.BasicLevel;
import org.w3c.dom.Document;
import org.xml.sax.SAXException;


/**
 * Application is a wrapper around an ear 
 * (packaged as a jar or as an unpacked jar).
 *
 * @author Guillaume Sauthier
 */
public class Application extends J2EEArchive {
    /** list of embded clients archive. */
    protected List clients;

    /** list of embded webapps archive. */
    protected List webapps;

    /** list of embded ejbjars archive. */
    protected List ejbjars;

    /** Application DeploymentDesc */
    private EarDeploymentDesc earDD = null;

    /** name to document map */
    private Map descriptors;

    /** application Descriptor */
    protected Document app;

    /** Common ClassLoader for Application */
    protected URLClassLoader earCL = null;

    /** EJB Jar ClassLoader */
    protected URLClassLoader ejbCL = null;

    /** Common Libs ClassLoader */
    protected URLClassLoader commonCL = null;

    /** libs path File */
    protected List pathFiles;

    /** application archive filename */
    private String appFilename;

    /**
     * Creates a new Application archive.
     *
     * @param archive the file containing the application archive.
     *
     * @throws WsGenException When Init fails
     */
    public Application(Archive archive) throws WsGenException {
        super(archive);
        appFilename = archive.getRootFile().getName();
        logger.log(BasicLevel.DEBUG, "Wrapping '" + appFilename + "' in Application");
        init();
    }

    /**
     * Initialize the Application. Creates modules lists, unpack if not
     * unpacked.
     *
     * @throws WsGenException When application unpack fails or
     *          when Decriptors cannot be parsed or found.
     */
    protected void init() throws WsGenException {

        pathFiles = new Vector();

        ejbjars = new Vector();
        webapps = new Vector();
        clients = new Vector();

        if(isPacked()) {

            try {
                // unpack data
                File unpacked =
                    FileUtils.unpack(new JarFile(archive.getRootFile()));
                archive = new FileArchive(unpacked);
            } catch(IOException ioe) {
                String err =
                    i18n.getMessage("Application.init.unpackException",
                        archive.getRootFile());
                throw new WsGenException(err, ioe);
            }
        }

        // load META-INF/application.xml
        try {
            earDD =
                EarDeploymentDesc.getInstance(getRootFile().getAbsolutePath(),
                    Thread.currentThread().getContextClassLoader());
        } catch(DeploymentDescException dde) {
            String err =
                i18n.getMessage("Application.init.earDDExc",
                    archive.getRootFile());
            throw new WsGenException(dde);
        }

        // Create ear ClassLoader (from MANIFEST/Class-Path general entry)
        earCL = createEARClassLoader();

        // add EjbJars
        String[] ejbs = earDD.getEjbTags();

        for(int i = 0; i < ejbs.length; i++) {
            File ejbFile = new File(getRootFile(), ejbs[i]);
            Archive ejbArch = null;

            if(ejbFile.isDirectory()) {
                // Unpacked Jar
                ejbArch = new FileArchive(ejbFile);
            } else {
                // Packed Jar
                ejbArch = new JarArchive(ejbFile);
            }
            ejbjars.add(new EjbJar(ejbArch));

            // add entries for Class-Path
            addClassPathEntry(ejbArch);
        }

        // add WebApps
        Web[] webs = earDD.getWebTags();

        for(int i = 0; i < webs.length; i++) {
            File webFile = new File(getRootFile(), webs[i].getWebUri());

            if(webFile.isDirectory()) {
                // Unpacked Jar
                webapps.add(new WebApp(new FileArchive(webFile)));
            } else {
                // Packed Jar
                webapps.add(new WebApp(new JarArchive(webFile)));
            }
        }

        // add Clients
        String[] clts = earDD.getClientTags();

        for(int i = 0; i < clts.length; i++) {
            File clientFile = new File(getRootFile(), clts[i]);

            if(clientFile.isDirectory()) {
                // Unpacked Jar
                clients.add(new Client(new FileArchive(clientFile)));
            } else {
                // Packed Jar
                clients.add(new Client(new JarArchive(clientFile)));
            }
        }

        ejbCL = createEJBClassLoader();

        loadDescriptors();
    }

    /**
     * Load Deployment Descriptor of an Application.
     *
     * @throws WsGenException When parsing of application.xml fails
     */
    protected void loadDescriptors() throws WsGenException {
        try {
            app = XMLUtils.newDocument(getApplicationInputStream());
        } catch(SAXException saxe) {
            String err =
                i18n.getMessage("Application.loadDescriptors.parseError");
            throw new WsGenException(err, saxe);
        } catch(ParserConfigurationException pce) {
            String err = i18n.getMessage("Application.loadDescriptors.prepare");
            throw new WsGenException(err, pce);
        } catch(IOException ioe) {
            String err =
                i18n.getMessage("Application.loadDescriptors.parseError");
            throw new WsGenException(err, ioe);
        }

        descriptors = new Hashtable();
        descriptors.put("META-INF/application.xml", app);
    }

    /**
     * Returns the name of the Archive.
     * Overrides J2EEArchive.getName();
     * @see org.objectweb.jonas_ws.wsgen.archive.J2EEArchive#getName()
     *
     * @return the name of the Archive.
     */
    public String getName() {
        return appFilename;
    }

    private URLClassLoader createEJBClassLoader() throws WsGenException {
        URL[] urls = new URL[pathFiles.size()];
        int index = 0;
        for (Iterator i = pathFiles.iterator(); i.hasNext(); index++) {
            try {
                urls[index] = ((File) i.next()).toURL();
            } catch(IOException ioe) {
                    String err = "";
                    throw new WsGenException(err, ioe);
            }
        }

        commonCL = new URLClassLoader(urls, earCL);

        urls = new  URL[ejbjars.size()];
        index = 0;
        for (Iterator i = ejbjars.iterator(); i.hasNext(); index++) {
            try {
                urls[index] = ((EjbJar) i.next()).getRootFile().toURL();
            } catch(IOException ioe) {
                    String err = "";
                    throw new WsGenException(err, ioe);
            }
        }
        
        try {
            return new EjbJarClassLoader(urls, commonCL);
        } catch(IOException ioe) {
            String err = "";
            throw new WsGenException(err, ioe);
        }
        
    }

    private URLClassLoader createEARClassLoader() throws WsGenException {

        // find parent ClassLoader
        ClassLoader parent = Thread.currentThread().getContextClassLoader();

        // get Manifest Attributes if any
        String classpath = getManifest().getMainAttributes().getValue(Attributes.Name.CLASS_PATH);
        URL[] urls = new URL[0];
        if (classpath != null) {
            // Lookup specified files.
            String[] paths = classpath.split(",");
            urls = new URL[paths.length];
            for (int i = 0; i < paths.length; i++) {
                try {
                    URL path = new File(getRootFile(), paths[i]).toURL();
                    urls[i] = path;
                } catch(IOException ioe) {
                    String err = "";
                    throw new WsGenException(err, ioe);
                }
            }
        }
        
        return new URLClassLoader(urls, parent);
    }


    private void addClassPathEntry(Archive a) throws WsGenException {
        // get Manifest Attributes if any
        String classpath = a.getManifest().getMainAttributes().getValue(Attributes.Name.CLASS_PATH);
        
        if (classpath != null) {
            // Lookup specified files.
            String[] paths = classpath.split(",");
            for (int i = 0; i < paths.length; i++) {
                try {
                    File path = new File(a.getRootFile().getParentFile(), paths[i]).getCanonicalFile();
                    if (!pathFiles.contains(path)) {
                        pathFiles.add(path);
                    }
                } catch(IOException ioe) {
                    String err = "";
                    throw new WsGenException(err, ioe);
                }
            }
            
        }
        
    }

    /**
     * Returns the Document of the application.xml file.
     *
     * @return the Document of the application.xml file.
     */
    public Document getApplicationDoc() {
        return app;
    }

    /**
     * Returns the InputStream of the application.xml file.
     *
     * @return the InputStream of the application.xml file.
     *
     * @throws IOException When InputStream of application.xml cannot be returned
     */
    public InputStream getApplicationInputStream() throws IOException {
        InputStream is = null;

        if(isPacked()) {
            is = getInputStream("META-INF/application.xml");
        } else {
            is = getInputStream("META-INF" + File.separator + "application.xml");
        }

        return is;
    }

    /**
     * Add a new EjbJar in the Application.
     *
     * @param ejbjar the added EjbJar
     */
    public void addEjbJar(EjbJar ejbjar) {
        ejbjars.add(ejbjar);
        // add module in application.xml
        XMLUtils.addEjb(app, ejbjar);
    }

    /**
     * Add a new Client in the Application.
     *
     * @param client the added Client
     */
    public void addClient(Client client) {
        clients.add(client);
        // add module in application.xml
        XMLUtils.addClient(app, client);
    }

    /**
     * Add a new WebApp in the Application.
     *
     * @param webapp the added webapp
     * @param context context of the webapp
     */
    public void addWebApp(WebApp webapp, String context) {
        webapps.add(webapp);
        // add module in application.xml
        XMLUtils.addWebApp(app, webapp, context);
        addFile(webapp.getRootFile(), webapp.getName());
    }

    /**
     * Returns the Iterator of EjbJar contained in this Application.
     *
     * @return the Iterator of EjbJar contained in this Application.
     */
    public Iterator getEjbJars() {
        return ejbjars.iterator();
    }

    /**
     * Returns the Iterator of WebApp contained in this Application.
     *
     * @return the Iterator of WebApp contained in this Application.
     */
    public Iterator getWebApps() {
        return webapps.iterator();
    }

    /**
     * Returns the Iterator of WebApp contained in this Application.
     *
     * @return the Iterator of WebApp contained in this Application.
     */
    public Iterator getClients() {
        return clients.iterator();
    }

    /**
     * Returns the ClassLoader of this ear archive.
     *
     * @return the ClassLoader of this ear archive.
     */
    public URLClassLoader getEARClassLoader() {
        return commonCL;
    }

    /**
     * Returns the ClassLoader of the ejbs within this archive.
     *
     * @return the ClassLoader of the ejbs within this archive.
     */
    public URLClassLoader getEJBClassLoader() {
        return ejbCL;
    }

    /**
     * Returns a Map of name to Document for each modified Descriptor of the
     * archive.
     *
     * @return a Map of name to Document
     */
    public Map getDescriptors() {
        return descriptors;
    }

    /**
     * Returns true if filename must be omitted in the archive.
     *
     * @param name filename to be tested
     *
     * @return true if filename must be omitted.
     */
    public boolean omit(String name) {
        if(name.equals("META-INF/application.xml")
                || name.equals("META-INF\\application.xml")) {
            return true;
        } else {
            return false;
        }
    }
}
