/*
 * JOnAS : Java(TM) OpenSource Application Server
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial Developer : Guillaume Sauthier
 * --------------------------------------------------------------------------
 * $Id: Client.java,v 1.3 2004/03/19 14:31:49 sauthieg Exp $
 * --------------------------------------------------------------------------
*/

package org.objectweb.jonas_ws.wsgen.archive;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URLClassLoader;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import javax.xml.parsers.ParserConfigurationException;

import org.objectweb.jonas_client.deployment.api.ClientContainerDeploymentDesc;
import org.objectweb.jonas_lib.deployment.api.DeploymentDescException;
import org.objectweb.jonas_lib.loader.ClientClassLoader;
import org.objectweb.jonas_ws.deployment.api.ServiceRefDesc;
import org.objectweb.jonas_ws.wsgen.WsGenException;
import org.objectweb.jonas_ws.wsgen.utils.XMLUtils;
import org.w3c.dom.Document;
import org.xml.sax.SAXException;


/**
 * Client represnets a J2EE Client Archive.
 *
 * @author Guillaume Sauthier
 */
public class Client extends J2EEArchive implements WsClient {
    /** Container application (can be null) */
    private Application app = null;

    /** Client Deployment Desc */
    private ClientContainerDeploymentDesc clientDD = null;

    /** service-ref list */
    private List sRefs;

    /** jonas client descriptors */
    private Map descriptors;

    /**
     * jonas-client.xml
     */
    private Document jclientDoc;

    /**
     * Create a new Client not contained in Application
     *
     * @param archive the archive file
     *
     * @throws WsGenException When Init fails.
     */
    public Client(Archive archive) throws WsGenException {
        super(archive);
        init();
    }

    /**
     * Create a new Client contained in Application
     *
     * @param archive the archive file
     * @param app the container Application
     *
     * @throws WsGenException When Init fails
     */
    public Client(Archive archive, Application app) throws WsGenException {
        super(archive);
        setApplication(app);
        init();
    }

    /**
     * Initialize the Client module.
     *
     * @throws WsGenException When client classloader cannot be created 
     *      or when client deployment desc cannot be parsed.
     */
    private void init() throws WsGenException {
        // if client in application, clientClassLoader includes ejbClassLoader too
        URLClassLoader clientClassLoader = null;

        try {
            if(app == null) {
                // simple client case
                clientClassLoader =
                    new ClientClassLoader(archive.getRootFile().toURL(),
                        Thread.currentThread().getContextClassLoader());
            } else {
                // embedded client case
                clientClassLoader =
                    new ClientClassLoader(archive.getRootFile().toURL(),
                        app.getEJBClassLoader());
            }
        } catch(IOException ioe) {
            String err =
                i18n.getMessage("Client.init.loader", archive.getRootFile());
            throw new WsGenException(err, ioe);
        }

        try {
            clientDD =
                ClientContainerDeploymentDesc.getInstance(getRootFile().getAbsolutePath(),
                                                          clientClassLoader);
        } catch(DeploymentDescException dde) {
            throw new WsGenException(dde);
        }

        // we want a List of service-ref
        sRefs = new Vector();

        ServiceRefDesc[] refs = clientDD.getServiceRefDesc();

        for(int i = 0; i < refs.length; i++) {
            sRefs.add(refs[i]);
        }

        loadDescriptors();

    }

    /**
     * Load Deployment Descriptor of a Client.
     *
     * @throws WsGenException when jonas-client.xml cannot be parsed
     */
    private void loadDescriptors() throws WsGenException {
        try {
            jclientDoc = XMLUtils.newDocument(getJonasClientInputStream());
        } catch(SAXException saxe) {
            String err = i18n.getMessage("Client.loadDescriptors.parseError");
            throw new WsGenException(err, saxe);
        } catch(ParserConfigurationException pce) {
            String err = i18n.getMessage("Client.loadDescriptors.prepare");
            throw new WsGenException(err, pce);
        } catch(IOException ioe) {
            String err = i18n.getMessage("Client.loadDescriptors.parseError");
            throw new WsGenException(err, ioe);
        }

        descriptors = new Hashtable();
        descriptors.put("META-INF/jonas-client.xml", jclientDoc);
    }

    /**
     * Set the container application.
     *
     * @param app the container application.
     */
    public void setApplication(Application app) {
        this.app = app;
    }

    /**
     * Returns the container application (can be null).
     *
     * @return the container application (can be null).
     */
    public Application getApplication() {
        return app;
    }

    /**
     * Returns the list of service-ref elements contained by a module.
     *
     * @return the list of service-ref elements contained by a module.
     */
    public List getServiceRefDescs() {
        return sRefs;
    }

    /**
     * add *.class from directory in the archive.
     *
     * @param classes directory with classes.
     */
    public void addClasses(File classes) {
        addDirectory(classes);
    }

    /**
     * Returns a Map of name to Document for each modified Descriptor of the
     * archive.
     *
     * @return a Map of name to Document
     */
    public Map getDescriptors() {
        return descriptors;
    }

    /**
     * Returns true if filename must be omitted in the archive.
     *
     * @param name filename to be tested
     *
     * @return true if filename must be omitted.
     */
    public boolean omit(String name) {
        if(name.equals("META-INF/jonas-client.xml")
                || name.equals("META-INF\\jonas-client.xml")) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * Returns the Document of the jonas-client.xml file.
     *
     * @return the Document of the jonas-client.xml file.
     */
    public Document getJonasClientDoc() {
        return jclientDoc;
    }

    /**
     * Returns the InputStream of the jonas-client.xml file.
     *
     * @return the InputStream of the jonas-client.xml file.
     *
     * @throws IOException When InputStream of jonas-client.xml cannot be returned.
     */
    public InputStream getJonasClientInputStream() throws IOException {
        InputStream is = null;

        if(isPacked()) {
            is = getInputStream("META-INF/jonas-client.xml");
        } else {
            is = getInputStream("META-INF" + File.separator
                    + "jonas-client.xml");
        }

        return is;
    }
}
