/**
 * JOnAS : Java(TM) OpenSource Application Server
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial Developer : Guillaume Sauthier
 * --------------------------------------------------------------------------
 * $Id: EjbJar.java,v 1.2 2004/03/24 11:11:08 sauthieg Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ws.wsgen.archive;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import javax.xml.parsers.ParserConfigurationException;

import org.objectweb.jonas_ejb.deployment.api.BeanDesc;
import org.objectweb.jonas_ejb.deployment.api.DeploymentDesc;
import org.objectweb.jonas_ejb.deployment.lib.EjbDeploymentDescManager;
import org.objectweb.jonas_lib.deployment.api.DeploymentDescException;
import org.objectweb.jonas_lib.loader.EjbJarClassLoader;
import org.objectweb.jonas_ws.deployment.api.WSDeploymentDesc;
import org.objectweb.jonas_ws.deployment.lib.WSDeploymentDescManager;
import org.objectweb.jonas_ws.wsgen.WsGenException;
import org.objectweb.jonas_ws.wsgen.utils.XMLUtils;
import org.w3c.dom.Document;
import org.xml.sax.SAXException;

/**
 * EjbJar represents an EjbJar J2EE Archive.
 * 
 * @author Guillaume Sauthier
 */
public class EjbJar extends J2EEArchive implements WsEndpoint {

    /** Application containing the ejbjar */
    private Application app;

    /** WebServices deployment descriptor */
    private WSDeploymentDesc wsDD = null;

    /** EjbJar deployment descriptor */
    private DeploymentDesc ejbDD = null;

    /** inner EJBs List */
    private List ejbs;

    /** descriptors list */
    private Map descriptors;

    /** jonas-ejb-jar document */
    private Document jEjbJar;

    /**
     * Create a new EjbJar not contained in Application
     * 
     * @param archive
     *            the archive file
     * 
     * @throws WsGenException
     *             When init fails
     */
    public EjbJar(Archive archive) throws WsGenException {
        super(archive);
        init();
    }

    /**
     * Create a new EjbJar contained in Application
     * 
     * @param archive
     *            the archive file
     * @param app
     *            container application
     * 
     * @throws WsGenException
     *             When init fails
     */
    public EjbJar(Archive archive, Application app) throws WsGenException {
        super(archive);
        setApplication(app);
        init();
    }

    /**
     * Initialize the EjbJar module.
     * 
     * @throws WsGenException
     *             When ejb classlaoder cannot be created or when Descriptor
     *             cannot be loaded.
     */
    private void init() throws WsGenException {
        // !!!todo!!! create the ejbjar classloader
        URLClassLoader ejbClassLoader = null;

        try {
            if (app == null) {
                // simple ejb case
                ejbClassLoader = new EjbJarClassLoader(new URL[] { archive
                        .getRootFile().toURL()}, Thread.currentThread()
                        .getContextClassLoader());
            } else {
                // embedded ejb case
                ejbClassLoader = app.getEJBClassLoader();
            }
        } catch (IOException ioe) {
            String err = i18n.getMessage("EjbJar.init.loader", archive
                    .getRootFile());
            throw new WsGenException(err, ioe);
        }

        try {
            ejbDD = EjbDeploymentDescManager.getDeploymentDesc(getRootFile()
                    .getAbsolutePath(), ejbClassLoader);
        } catch (DeploymentDescException dde) {
            throw new WsGenException(dde);
        }

        try {
            wsDD = WSDeploymentDescManager.getDeploymentDesc(getRootFile()
                    .getAbsolutePath(), ejbClassLoader);
        } catch (DeploymentDescException dde) {
            throw new WsGenException(dde);
        }

        loadDescriptors();

        // create inner EJBs
        ejbs = new Vector();

        BeanDesc[] bd = ejbDD.getBeanDesc();

        for (int i = 0; i < bd.length; i++) {
            ejbs.add(new Ejb(bd[i], XMLUtils.getBeanElement(jEjbJar
                    .getDocumentElement(), bd[i].getEjbName())));
        }
    }

    /**
     * Load Deployment Descriptor of an EjbJar.
     * 
     * @throws WsGenException
     *             When parsing fails
     */
    private void loadDescriptors() throws WsGenException {
        try {
            jEjbJar = XMLUtils.newDocument(getJonasEjbJarInputStream());
        } catch (SAXException saxe) {
            String err = i18n.getMessage("EjbJar.loadDescriptors.parseError");
            throw new WsGenException(err, saxe);
        } catch (ParserConfigurationException pce) {
            String err = i18n.getMessage("EjbJar.loadDescriptors.prepare");
            throw new WsGenException(err, pce);
        } catch (IOException ioe) {
            String err = i18n.getMessage("EjbJar.loadDescriptors.parseError");
            throw new WsGenException(err, ioe);
        }

        descriptors = new Hashtable();
        descriptors.put("META-INF/jonas-ejb-jar.xml", jEjbJar);
    }

    /**
     * Returns the List of Ejb contained in this EjbJar.
     * 
     * @return the List of Ejb contained in this EjbJar.
     */
    public List getEjbs() {
        return ejbs;
    }

    /**
     * Returns the list of webservice-description elements contained by a
     * module.
     * 
     * @return the list of webservice-description elements contained by a
     *         module.
     */
    public List getServiceDescs() {
        if (wsDD != null) {
            return wsDD.getServiceDescs();
        } else {
            return new Vector();
        }
    }

    /**
     * Add *.class from directory in the archive.
     * 
     * @param classes
     *            directory containing classes files.
     */
    public void addClasses(File classes) {
        addDirectory(classes);
    }

    /**
     * Set the container application of this EjbJar
     * 
     * @param app
     *            container application
     */
    public void setApplication(Application app) {
        this.app = app;
    }

    /**
     * Return the container application of this EjbJar
     * 
     * @return the container application of this EjbJar
     */
    public Application getApplication() {
        return app;
    }

    /**
     * Returns a Map of name to Document for each modified Descriptor of the
     * archive.
     * 
     * @return a Map of name to Document
     */
    public Map getDescriptors() {
        return descriptors;
    }

    /**
     * Returns true if filename must be omitted in the archive.
     * 
     * @param name
     *            filename to be tested
     * 
     * @return true if filename must be omitted.
     */
    public boolean omit(String name) {
        if (name.equals("META-INF/jonas-ejb-jar.xml")
                || name.equals("META-INF\\jonas-ejb-jar.xml")) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * Returns the Document of the jonas-ejb-jar.xml file.
     * 
     * @return the Document of the jonas-ejb-jar.xml file.
     */
    public Document getJonasEjbJarDoc() {
        return jEjbJar;
    }

    /**
     * Returns the InputStream of the jonas-ejb-jar.xml file.
     * 
     * @return the InputStream of the jonas-ejb-jar.xml file.
     * 
     * @throws IOException
     *             When InputStream of jonas-ejb-jar.xml cannot be returned
     */
    public InputStream getJonasEjbJarInputStream() throws IOException {
        InputStream is = null;

        if (isPacked()) {
            is = getInputStream("META-INF/jonas-ejb-jar.xml");
        } else {
            is = getInputStream("META-INF" + File.separator
                    + "jonas-ejb-jar.xml");
        }

        return is;
    }
}
