/*
 * JOnAS : Java(TM) OpenSource Application Server
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial Developer : Guillaume Sauthier
 * --------------------------------------------------------------------------
 * $Id: FileArchive.java,v 1.3 2004/03/19 14:31:49 sauthieg Exp $
 * --------------------------------------------------------------------------
*/

package org.objectweb.jonas_ws.wsgen.archive;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.List;
import java.util.Vector;
import java.util.jar.Manifest;

import org.objectweb.jonas_ws.wsgen.WsGenException;


/**
 * A <code>FileArchive</code> is a wrapper for directory structured as a jar.
 *
 * @author Guillaume Sauthier
 */
public class FileArchive extends AbsArchive {
    /**
     * Creates a new FileArchive object.
     *
     * @param archive directory structured as a jar
     *
     * @throws WsGenException When manifest cannot be loaded
     */
    public FileArchive(File archive) throws WsGenException {
        super(archive);

        try {
            File mf =
                new File(archive, "META-INF" + File.separator + "MANIFEST.MF");

            if(mf.exists()) {
                manifest = new Manifest(new FileInputStream(mf));
            } else {
                manifest = new Manifest();
            }
        } catch(Exception e) {
            String err =
                i18n.getMessage("FileArchive.constr.manifest", getRootFile());
            throw new WsGenException(err, e);
        }
    }

    /**
     * Returns the Manifest of the Archive.
     *
     * @return the Manifest of the Archive.
     */
    public Manifest getManifest() {
        return manifest;
    }

    /**
     * Returns an InputStream corresponding to the given filename.
     *
     * @param filename file name source of the InputStream
     *
     * @return the InputStream corresponding to the given filename.
     *
     * @throws IOException When InputStream corersponding to the given filename cannot be found.
     */
    public InputStream getInputStream(String filename)
        throws IOException {
        File file = (File) files.get(filename);

        if(file == null) {
            // filename not found in added files
            // try root search
            file = new File(root, filename);

            if(!file.exists()) {
                return null;
            }
        }

        // file exists (in added files or in original archive)
        return new FileInputStream(file);
    }

    /**
     * Returns a List of all files contained in this archive. Original files in
     * jar, added Files are all included as String in this Enumeration.
     *
     * @return a List of all files contained in this archive.
     */
    public List getContainedFiles() {
        List list = new Vector(files.keySet());

        // add files of the original archive
        traverse("", root, list);

        return list;
    }

    /**
     * Add all files contained in the given directory into a list.
     *
     * @param dirName current directory name
     * @param base directory where file are listed
     * @param map list of filename
     */
    private static void traverse(String dirName, File base, List map) {
        File[] childs = base.listFiles();

        // directory exists ?
        if(childs != null) {
            for(int i = 0; i < childs.length; i++) {
                if(childs[i].isFile()) {
                    // File
                    map.add(dirName + childs[i].getName());
                } else {
                    // Directory
                    traverse(dirName + childs[i].getName() + File.separator,
                        childs[i], map);
                }
            }
        }
    }

    /**
     * Returns true if archive is packed or false if archive is unpacked.
     *
     * @return true if archive is packed or false if archive is unpacked.
     */
    public boolean isPacked() {
        return false;
    }
}
