/*
 * JOnAS : Java(TM) OpenSource Application Server
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial Developer : Guillaume Sauthier
 * --------------------------------------------------------------------------
 * $Id: Generator.java,v 1.5 2004/03/25 15:38:44 sauthieg Exp $
 * --------------------------------------------------------------------------
*/

package org.objectweb.jonas_ws.wsgen.generator;

import java.io.File;
import java.io.FileFilter;
import java.io.IOException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Enumeration;
import java.util.Vector;

import org.objectweb.common.Cmd;
import org.objectweb.common.Env;
import org.objectweb.jonas.common.Log;
import org.objectweb.jonas_lib.I18n;
import org.objectweb.jonas_ws.wsgen.WsGenException;
import org.objectweb.jonas_ws.wsgen.archive.Archive;
import org.objectweb.jonas_ws.wsgen.utils.TempRepository;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;


/**
 * Generators provide a structure to be extended for specific generation
 * mecanisms. Axis will provide an AxisWsClientGenerator that will be in
 * charge of client side generation  (WSDL2Java tool) and an
 * AxisWsEndpointGenerator that will bother with server side artifact
 * generation.
 *
 * @author Guillaume Sauthier
 */
public abstract class Generator {
    /** i18n */
    protected static I18n i18n = I18n.getInstance(Generator.class);

    /** logger */
    protected static final Logger logger = Log.getLogger(Log.JONAS_WSGEN_PREFIX);

    /** Configuration to be used */
    protected Generator.Config config;

    /** compiled classes directory */
    protected File classes;

    /** generated files directory */
    protected File sources;

    /**
     * Creates a new Generator with the given Config.
     *
     * @param config internal configuration object.
     *
     * @throws WsGenException When sources and target temporary directory cannot be created
     */
    public Generator(Generator.Config config) throws WsGenException {
        this.config = config;

        // creates temporary directories
        TempRepository tr = TempRepository.getInstance();

        try {
            sources = tr.createDir();
            classes = tr.createDir();
        } catch(IOException ioe) {
            String err = i18n.getMessage("Generator.constr.ioe");
            logger.log(BasicLevel.ERROR, err);
            throw new WsGenException(err, ioe);
        }
    }

    /**
     * Generate files
     */
    public abstract void generate() throws WsGenException;

    /**
     * Compile generated java files into classes directory.
     *
     * @throws WsGenException When compilation fails
     */
    public void compile() throws WsGenException {
        try {
            // get all java files contained in the sources dir
            Vector srcs = new Vector();
            addJavaSources(sources, srcs);

            Cmd cmd = new Cmd(config.javaHomeBin + config.nameJavac);
            cmd.addArgument("-classpath");
            cmd.addArgument(config.classpath);
            cmd.addArgument("-d");
            cmd.addArgument(classes.getCanonicalPath());
            cmd.addArguments(config.javac_opts);

            if(srcs.size() != 0) {
                for(Enumeration e = srcs.elements(); e.hasMoreElements();) {
                    String srcName =
                        ((File) e.nextElement()).getCanonicalPath();
                    cmd.addArgument(srcName);
                }

                logger.log(BasicLevel.DEBUG, "Running '" + cmd.toString() + "'");
                if(cmd.run()) {
                    logger.log(BasicLevel.INFO, "WebServices Classes successfully compiled.");
                } else {
                    String err = i18n.getMessage("Generator.compile.error");
                    logger.log(BasicLevel.ERROR, err);
                    throw new WsGenException(err);
                }
            }
        } catch(Exception e) {
            String err = i18n.getMessage("Generator.compile.error");
            logger.log(BasicLevel.ERROR, err);
            throw new WsGenException(err, e);
        }
    }

    /**
     * Recursively add java files into a Vector.
     *
     * @param src base directory
     * @param list Vector to be filled
     */
    private void addJavaSources(File src, Vector list) {
        // add each java sources contained in the directory
        File[] files =
            src.listFiles(new FileFilter() {
                    public boolean accept(File file) {
                        return file.isFile()
                        && file.getName()
                        .endsWith(".java");
                    }
                });

        for(int i = 0; i < files.length; i++) {
            list.add(files[i]);
        }

        // reapply on subdirectories
        files =
            src.listFiles(new FileFilter() {
                    public boolean accept(File file) {
                        return file.isDirectory();
                    }
                });

        for(int i = 0; i < files.length; i++) {
            addJavaSources(files[i], list);
        }
    }

    /**
     * Add generated files into an Archive
     *
     * @param archive the archive destination of generated files.
     */
    public abstract void addFiles(Archive archive) throws WsGenException;

    /**
     * Configuration object storing generation params.
     */
    public static class Config {
        /** packed mode */
        public static final int PACKED = 0;

        /** unpacked mode */
        public static final int UNPACKED = 1;

        /** Javac executable name */
        public String nameJavac = "javac";

        /** Javac options */
        public Vector javac_opts = new Vector();

        /** Bin directory for JDK */
        public String javaHomeBin = null;

        /** no config needed */
        public boolean noConfig = false;

        /** verbose mode */
        public boolean verbose = false;

        /** debug mode */
        public boolean debug = false;

        /** classpath */
        public String classpath = ".";

        /** output directory */
        public File out = new File(".");

        /** keep alrady generated files */
        public boolean keepGenerated = false;

        /** validating parser ? */
        public boolean parseWithValidation = true;

        /** configuration error */
        public boolean error = false;

        /** file input name */
        public String inputname;

        /** help requested ? */
        public boolean help = false;

        /** save mode (PACKED/UNPACKED) */
        public int saveMode = PACKED;

        /**
         * Creates a new Config. Automatically setup javaHomeBin property. And
         * create classpath from classloader.
         */
        public Config() {
            // Setup java_home/bin directory
            javaHomeBin = System.getProperty("java.home", "");

            if(!("".equals(javaHomeBin))) {
                if(Env.isOsMacOsX()) {
                    javaHomeBin =
                        javaHomeBin + File.separator + "bin" + File.separator;
                } else {
                    // JRE Directory !
                    javaHomeBin =
                        javaHomeBin + File.separator + ".." + File.separator
                        + "bin" + File.separator;
                }
            }

            // classpath
            URLClassLoader ucl =
                (URLClassLoader) (
                    Thread.currentThread().getContextClassLoader()
                );
            URL[] urls = ucl.getURLs();

            for(int i = 0; i < urls.length; i++) {
                classpath += (File.pathSeparator + urls[i].getFile());
            }
        }
    }
}
