/*
 * JOnAS : Java(TM) OpenSource Application Server
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial Developer : Guillaume Sauthier
 * --------------------------------------------------------------------------
 * $Id: AxisWsEndpointGenerator.java,v 1.2 2004/02/16 12:16:42 sauthieg Exp $
 * --------------------------------------------------------------------------
*/

package org.objectweb.jonas_ws.wsgen.generator.axis;



import java.io.File;
import java.util.Iterator;
import org.apache.velocity.VelocityContext;
import org.objectweb.jonas_ws.deployment.api.ServiceDesc;
import org.objectweb.jonas_ws.deployment.api.SSBPortComponentDesc;
import org.objectweb.jonas_ws.deployment.api.JaxRpcPortComponentDesc;
import org.objectweb.jonas_ejb.deployment.api.SessionStatelessDesc;
import org.objectweb.jonas_ws.wsgen.WsGenException;
import org.objectweb.jonas_ws.wsgen.archive.Archive;
import org.objectweb.jonas_ws.wsgen.archive.WebApp;
import org.objectweb.jonas_ws.wsgen.ddmodifier.WsEndpointDDModifier;
import org.objectweb.jonas_ws.wsgen.generator.Generator;
import org.objectweb.jonas_ws.wsgen.generator.WsEndpointGenerator;


/**
 * Generate Axis specific config files for Endpoint
 *
 * @author Guillaume Sauthier
 */
public class AxisWsEndpointGenerator extends WsEndpointGenerator {
    /** unique JVelocity instance */
    private static JVelocity jvelocity = null;

    /**
     * Axis Servlet classname
     */
    private static final String AXIS_SERVLET_CLASS =
        "org.apache.axis.transport.http.AxisServlet";

    /**
     * init-param name for declaring server configuration file
     */
    private static final String SERVER_CONFIG = "axis.serverConfigFile";

    /** count generated files */
    private static int count = 0;

    /** generated file */
    private File generated;

    /**
     * Creates a new AxisWsEndpointGenerator
     *
     * @param config Generator Configuration
     * @param serviceDesc WebService Endpoint description
     * @param ddm Web DD Modifier
     *
     * @throws WsGenException DOCUMENT ME!
     */
    public AxisWsEndpointGenerator(Generator.Config config,
        ServiceDesc serviceDesc, WsEndpointDDModifier ddm)
        throws WsGenException {
        super(config, serviceDesc, ddm);

        // init velocity
        if(jvelocity == null) {
            jvelocity = new JVelocity("deploy_endpoint.vm");
        }
    }

    /**
     * Generate server side configuration file
     *
     * @throws WsGenException When generation fails
     */
    public void generate() throws WsGenException {
        String sName = sd.getName();

        // construct VelocityContext
        VelocityContext vc = VContextFactory.getContext(sd);

        // Generate file
        String filename = "server-config-" + (count++) + ".wsdd";
        generated = new File(sources, filename);
        jvelocity.generate(generated, vc);

        // remove existing servlets
        for (Iterator i = sd.getPortComponents().iterator(); i.hasNext(); ) {
            Object obj = i.next();
            if (obj instanceof JaxRpcPortComponentDesc) {
                JaxRpcPortComponentDesc jax = (JaxRpcPortComponentDesc) obj;
                modifier.removeServletMapping(jax.getSibLink());
            }
        }

        // and use our own
        modifier.addServlet(sName, AXIS_SERVLET_CLASS);
        modifier.addServletParam(sName, SERVER_CONFIG, filename);
        modifier.addServletMapping(sName, "/" + sName + "/*");

    }

    /**
     * Add generated files in given Archive
     *
     * @param archive WebApp archive
     *
     * @throws WsGenException When cannot add files in archive
     * @throws IllegalArgumentException When archive where file 
     *          should be added is not a WebApp Archive
     */
    public void addFiles(Archive archive) throws WsGenException {
        // archive must be a WebApp
        if(!(archive instanceof WebApp)) {
            String err =
                i18n.getMessage("AxisWsEndpointGenerator.addFiles.illegal",
                    archive.getRootFile());
            throw new IllegalArgumentException(err);
        }

        WebApp web = (WebApp) archive;
        web.addFileIn("WEB-INF/", generated);

        // add Ejb Ref in descriptor only now because filename is unknown before
        // filename of the webapp is of the following form :
        // <ejb-jar filename>.war
        String webFilename = web.getName();
        String ejbjarFilename = webFilename.substring(0, webFilename.length() - 4) + ".jar";
        for (Iterator i = sd.getPortComponents().iterator(); i.hasNext(); ) {
            Object obj = i.next();
            if (obj instanceof SSBPortComponentDesc) {
                SSBPortComponentDesc pc = (SSBPortComponentDesc) obj;
                SessionStatelessDesc ss = pc.getSessionStatelessDesc();
                if (ss.getLocalClass() != null) {
                    // Bean have a local interface
                    modifier.addEjbLocalRef("ejb/endpoint/" + ss.getEjbName(),
                                            ss.getLocalHomeClass().getName(),
                                            ss.getLocalClass().getName(),
                                            ejbjarFilename + "#" + ss.getEjbName());
                } else {
                    // use remote interface
                    modifier.addEjbRef("ejb/endpoint/" + ss.getEjbName(),
                                       ss.getHomeClass().getName(),
                                       ss.getRemoteClass().getName(),
                                       ejbjarFilename + "#" + ss.getEjbName());
                }
            }
        }

    }
}
