/*
 * JOnAS : Java(TM) OpenSource Application Server
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial Developer : Guillaume Sauthier
 * --------------------------------------------------------------------------
 * $Id: ModifierFactory.java,v 1.2 2004/03/24 11:11:08 sauthieg Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ws.wsgen.modifier;

import java.io.File;
import java.io.IOException;
import java.util.jar.JarFile;

import org.objectweb.jonas_lib.I18n;
import org.objectweb.jonas_ws.wsgen.WsGenException;
import org.objectweb.jonas_ws.wsgen.archive.Application;
import org.objectweb.jonas_ws.wsgen.archive.Archive;
import org.objectweb.jonas_ws.wsgen.archive.Client;
import org.objectweb.jonas_ws.wsgen.archive.EjbJar;
import org.objectweb.jonas_ws.wsgen.archive.FileArchive;
import org.objectweb.jonas_ws.wsgen.archive.J2EEArchive;
import org.objectweb.jonas_ws.wsgen.archive.JarArchive;
import org.objectweb.jonas_ws.wsgen.archive.WebApp;

/**
 * Used to create the right ArchiveModifier from a given filename (ear, jar,
 * war, ...)
 * 
 * @author Guillaume Sauthier
 */
public class ModifierFactory {

    /** i18n */
    private static I18n i18n = I18n.getInstance(ModifierFactory.class);

    /** Archive modes */
    private static final int APPLICATION = 0;

    private static final int EJBJAR = 1;

    private static final int WEBAPP = 2;

    private static final int CLIENT = 3;

    /**
     * Returns an <code>ArchiveModifier</code> according to archive type
     * (application, ejbjar, webapp or client).
     * 
     * @param filename
     *            input filename.
     * 
     * @return an <code>ArchiveModifier</code>.
     * 
     * @exception WsGenException
     *                when archive creation fails.
     */
    public static ArchiveModifier getModifier(String filename)
            throws WsGenException {
        Archive archive = null;
        J2EEArchive j2eeArchive = null;
        ArchiveModifier modifier = null;
        int mode;

        File file = new File(filename);

        if (file.exists()) {
            if (file.isFile()) {
                // should be a jar file
                // test jar file
                JarFile jarfile;

                try {
                    jarfile = new JarFile(file);
                } catch (IOException ioe) {
                    // not a jar file
                    String err = i18n.getMessage(
                            "ModifierFactory.getModifier.notjar", filename);
                    throw new WsGenException(err);
                }

                // Create the inner Archive
                archive = new JarArchive(file);

                // now we are sure that filename is a correct jar
                if (isApplication(jarfile)) {
                    // Application J2EE Archive
                    j2eeArchive = new Application(archive);
                    mode = APPLICATION;

                } else if (isEjbJar(jarfile)) {
                    // EjbJar J2EE Archive
                    j2eeArchive = new EjbJar(archive);
                    mode = EJBJAR;

                } else if (isWebApp(jarfile)) {
                    // WebApp J2EE Archive
                    j2eeArchive = new WebApp(archive);
                    mode = WEBAPP;

                } else if (isClient(jarfile)) {
                    // Client J2EE Archive
                    j2eeArchive = new Client(archive);
                    mode = CLIENT;

                } else {
                    // unsupported archive type
                    String err = i18n
                            .getMessage(
                                    "ModifierFactory.getModifier.unsupported",
                                    filename);
                    throw new WsGenException(err);
                }
            } else {
                // directory unpacked
                // Create the inner Archive
                archive = new FileArchive(file);

                // now we are sure that filename is a correct jar
                if (isApplication(file)) {
                    // Application J2EE Archive
                    j2eeArchive = new Application(archive);
                    mode = APPLICATION;

                } else if (isEjbJar(file)) {
                    // EjbJar J2EE Archive
                    j2eeArchive = new EjbJar(archive);
                    mode = EJBJAR;

                } else if (isWebApp(file)) {
                    // WebApp J2EE Archive
                    j2eeArchive = new WebApp(archive);
                    mode = WEBAPP;

                } else if (isClient(file)) {
                    // Client J2EE Archive
                    j2eeArchive = new Client(archive);
                    mode = CLIENT;

                } else {
                    // unsupported archive type
                    String err = i18n
                            .getMessage(
                                    "ModifierFactory.getModifier.unsupported",
                                    filename);
                    throw new WsGenException(err);
                }
            }
        } else {
            String err = i18n.getMessage(
                    "ModifierFactory.getModifier.notfound", filename);
            throw new WsGenException(err);
        }

        //create the modifier
        switch (mode) {
        case APPLICATION:
            modifier = new ApplicationModifier((Application) j2eeArchive);

            break;

        case EJBJAR:
            modifier = new EjbJarModifier((EjbJar) j2eeArchive);

            break;

        case WEBAPP:
            modifier = new WebAppModifier((WebApp) j2eeArchive);

            break;

        case CLIENT:
            modifier = new ClientModifier((Client) j2eeArchive);

            break;
        }

        return modifier;

    }

    /**
     * returns true if the file is an Application archive.
     * 
     * @param jf
     *            JarFile to explore.
     * 
     * @return true if the file is an Application archive.
     */
    private static boolean isApplication(JarFile jf) {
        return jf.getEntry("META-INF/application.xml") != null;
    }

    /**
     * returns true if the file is an EjbJar archive.
     * 
     * @param jf
     *            JarFile to explore.
     * 
     * @return true if the file is an EjbJar archive.
     */
    private static boolean isEjbJar(JarFile jf) {
        return jf.getEntry("META-INF/ejb-jar.xml") != null;
    }

    /**
     * returns true if the file is a WebApp archive.
     * 
     * @param jf
     *            JarFile to explore.
     * 
     * @return true if the file is a WebApp archive.
     */
    private static boolean isWebApp(JarFile jf) {
        return jf.getEntry("WEB-INF/web.xml") != null;
    }

    /**
     * returns true if the file is a Client archive.
     * 
     * @param jf
     *            JarFile to explore.
     * 
     * @return true if the file is a Client archive.
     */
    private static boolean isClient(JarFile jf) {
        return jf.getEntry("META-INF/application-client.xml") != null;
    }

    /**
     * returns true if the file is an Application archive.
     * 
     * @param f
     *            Directory to explore.
     * 
     * @return true if the file is an Application archive.
     */
    private static boolean isApplication(File f) {
        return new File(f, "META-INF" + File.separator + "application.xml")
                .exists();
    }

    /**
     * returns true if the file is an EjbJar archive.
     * 
     * @param f
     *            Directory to explore.
     * 
     * @return true if the file is an EjbJar archive.
     */
    private static boolean isEjbJar(File f) {
        return new File(f, "META-INF" + File.separator + "ejb-jar.xml")
                .exists();
    }

    /**
     * returns true if the file is a WebApp archive.
     * 
     * @param f
     *            Directory to explore.
     * 
     * @return true if the file is a WebApp archive.
     */
    private static boolean isWebApp(File f) {
        return new File(f, "WEB-INF" + File.separator + "web.xml").exists();
    }

    /**
     * returns true if the file is a Client archive.
     * 
     * @param f
     *            Directory to explore.
     * 
     * @return true if the file is a Client archive.
     */
    private static boolean isClient(File f) {
        return new File(f, "META-INF" + File.separator
                + "application-client.xml").exists();
    }
}
